{$IFNDEF FMX}
unit RSGraphics;
{$ENDIF}
//=== File Prolog ============================================================
//    This code was developed by RiverSoftAVG.
//
//--- Notes ------------------------------------------------------------------
//
//--- Development History  ---------------------------------------------------
//
//  12/2015 T. Grubb
//          - Added TPathPointKindHelper (FMX) to remove deprecated warnings
//            introduced in XE6
//          - Added TCornerTypeHelper (FMX) to remove deprecated warnings
//            introduced in XE6
//          - Added TRSPathData.FormatStr property to allow formatting the floating
//            point values (generally to make the GetPathString return a smaller
//            string)
//          - Added TRSPathData.LongFormatStr constant
//          - Added TRSPathData.ShortFormatStr constant
//          - Modified TRSPathData.GetPathString method to use FormatStr
//          - Modified TRSPathData.GetPathString method to use local method
//            instead of inherited GetPathString in FMX
//  11/2015 T. Grubb
//          - Added RSDistance inline function
//          - Added RSCenterPoint( const Pt1, Pt2: TRSPoint; Midpoint: Single = 0.5 ) overloaded function
//          - Added TRSPathData.SplitDataByLength method
//          - Fixed bug in RSPointToString function to use USFormatSettings
//            for FloatToStr calls
//          - Fixed bug in Arc function to use USFormatSettings
//            for FloatToStr calls
//          - Fixed bug in TRSPathData.GetPathString method to use USFormatSettings
//            for FloatToStr calls
//          - Fixed bug in TRSPathData.AddArcSvg method where arcs that begin
//            and end at same point would not get added (now adds ellipse)
//          - (Re)Fixed TRSPathData.SetPathString method to CONSTRAIN value to 0 or 1 when
//            parsing large-arc-flag or sweep-arc-flag
//          - Fixed bug in IdentToCanvasColor function which used LongInt as
//            Color parameter instead of Integer
//  09/2015 T. Grubb
//          - Added RSMatrixEquals function
//          - Added TPathClickLocation type
//          - Added TRSPathData.PointInPathOutline(TRSPoint;Single): Boolean
//            method to detect if point is on outline of a path
//          - Added TRSPathData.PointInPathOutline(Single;Single;Single): Boolean
//            method to detect if point is on outline of a path
//          - Added TRSPathData.AddPathOutline method to create
//            a crude outline around a path (useful for detecting a point that
//            is on the stroke of a path (which FMX does not detect)
//          - Added TRSPathData.AddRectangleForSegment protected method to create
//            a rectangle around a line segment
//          - Added PointInPath(TRSPoint;TRSPoints): TPathClickLocation
//            overloaded function
//          - Added PointInPath(TPoint;TPoints): TPathClickLocation
//            overloaded function
//  06/2015 T. Grubb
//          - Fixed TRSPathData.SetPathString method to only accept 0 or 1 when
//            parsing large-arc-flag or sweep-arc-flag
//  05/2015 T. Grubb
//          - Fixed divide by zero exception in FitRect( TRSMatrix, TRSRect, TRSRect )
//            function
//          - Modified TRSPathData.SetPathString method to add a ClosePath when the path
//            ends in a Move.  FMX seems to have a problem with a path that ends
//            with a move and fills the space with black
//          - Fixed bug in TRSPathData.SetPathString method, which would not set control point
//            to be coincident with the current point after every non-Q/T segment,
//            causing T segment creation to fail if T not preceded by either Q or T
//  01/2015 T. Grubb
//          - Changed TVclCanvasHelper.PenWidth property from Single to Integer
//  12/2014 T. Grubb
//          - Added overloaded RSInflateRect( TRect, Integer, Integer ) function
//          - Added overloaded RSIntersectRect( TRect, TRect ) function
//          - Added overloaded RSIntersectRect( TRect, TRect, var TRect ) function
//          - Added overloaded RSIntersectRect( TRSRect, TRSRect ) function
//          - Fixed bug in TRSPathData.DrawPath, which would draw consecutive
//            curves in a non-continuous manner.  Changed drawing code for VCL
//            to use Windows.BeginPath/PolyBezierTo/etc statements
//          - Added DrawDonutSlice function
//          - Added Arc (donut-like) path function
//          - Added overloaded DrawPieSlice function (with OuterRadius parameter)
//          - Fixed bug in DrawPieSlice function (FMX) which ignored offset parameter
//          - Fixed bug in DrawArrow function where arrow line would not draw all
//            the way to its base
//  09/2014 T. Grubb
//          - Fixed *relative* Quadratic Bezier Curve support for TRSPathData (q & t)
//          - Fixed TRSPathData.SmoothCurveTo method where LastPoint was a MoveTo
//            would create a bad curve
//          - Fixed TRSPathData.SmoothCurveToRel method where LastPoint was a MoveTo
//            would create a bad curve
//  07/2014 T. Grubb
//          - Added Quadratic Bezier Curve support to TRSPathData (Q/q and T/t),
//            SetPathString method converts Quadratic Bezier curves to Cubic
//            Bezier curves
//  04/2014 T. Grubb
//          - Updated code for XE6
//          - Fixed TRSPathData.GetNum method to accept numbers like '2.964.947'
//            as 2 valid separate numbers (2.964, 0.947), which apparently
//            Photoshop can output as SVG path numbers
//  03/2014 T. Grubb
//          - Added range checks for string in TRSPathData.SetPathString method
//          - Added range checks for string in TRSPathData.GetNum method
//          - Added EqualsRect(R1,R2: TRect) function
//  02/2014 T. Grubb
//          - Fixed TRSPathData.GetNum method to skip spaces AND commas
//          - Fixed bug in TRSPathData.SmoothCurveTo method when Count of points
//            is less than or equal to 2
//          - Fixed bug in TRSPathData.SmoothCurveToRel method when Count of points
//            is less than or equal to 2
//  01/2014 T. Grubb
//          - Modified TRSPathData.FillPath to fix a bug in TCanvas.FillPath
//            (FMX, at least up to XE5) where if the path does not end with a Close point,
//            it can cause visual errors like blacking out the canvas
//          - Added Grow paremeter to FitRect functions, which (when true) forces
//            the rectangle to grow to the full size of the target rect
//  01/2014 T. Grubb
//          - Added RSIntersectLine function
//          - Added overloaded RSIntersectRect function
//          - Fixed bug in TRSPathData.PointInPath method
//  12/2013 T. Grubb
//          - Added TRSPathData.Clone method
//  11/2013 T. Grubb
//          - Fixed StrToFloat to use US Locale information in TRSPathData (Path
//            definitions use standard decimal format, not locale-specific)
//          - Fixed StrToFloat to use US Locale information in StringToRSPoint function
//          - Added USFormatSettings variable
//  09/2013 T. Grubb
//          - Updated code for XE5
//          - Fixed bug in TPathData.SetPathString which would not properly parse
//            consecutive arcs (A/a) if the A/a token did not exist between
//            arc points
//          - Fixed bug in SetGradientColors function which would not properly set
//            up the gradient colors
//          - Fixed bug in TRSPathData.GetNum method which would not accept
//            uppercase E notation in floating point number
//          - Added TRSPosition class (maps to FMX TPosition in XE2+)
//          - Added TRSTransform class (maps to FMX TTransform in XE2+)
//          - Added overloaded RSRect( const TopLeft, BottomRight: TRSPoint ) method
//          - Added overloaded EllipseGradient method
//          - Added TRSPolygons type
//          - Added TPointPolygons type
//          - Added TCanvasPolygons type
//          - Added TRSPathData.AddPolygon( const Poly: TPoints ) method
//          - Added TRSPathData.AddPolygon( const Poly: TRSPoints ) method
//          - Added TRSPathData.AddPolygons( const Poly: TPointPolygons ) method
//          - Added TRSPathData.AddPolygons( const Poly: TRSPolygons ) method
//          - Added FitRect( aMatrix: TRSMatrix; const aRect, BoundsRect: TRSRect ): TRSRect method
//          - Added TRSPathData.FillMode property (not used in FMX)
//          - Modified TRSPathData.FitToRect method to create a transformation
//            matrix and then call ApplyMatrix
//          - Renamed RSGraphics.IntersectRect function to RSIntersectRect function
//            to avoid collisions with Delphi function
//  08/2013 T. Grubb
//          - Changed TRSPathData to descend from TPathData in FMX (speeds up draw)
//          - Added opacity parameter to TRSPathData.Draw methods
//          - Fixed bug in TRSPathData.DrawPath where polygon/polyline would
//            not be drawn without a closepath
//          - Added TRSPathData.AssignTo method to assign to TPathData in FMX
//          - Changed TRSPathData.DrawPath method to use Canvas.DrawPath in FMX
//  07/2013 T. Grubb
//          - Added TRSPathData object, similar to FMX TPathData
//          - Added TRSCorner type
//          - Added TRSCorners type
//          - Added TRSCornerType type
//          - Added TRSPathPointKind type
//          - Added TRSPathPoint type
//          - Added RSVectorAdd function
//          - Added RSVectorSubtract function
//          - Added RSMatrixTransform(TRSMatrix, TRSVector) function
//          - Added overloaded CanvasRect( TRSRect ): TCanvasRect function
//          - Added overloaded CanvasRect( TRect ): TCanvasRect function
//          - Added overloaded RSRectHeight( TRect ) function
//          - Added overloaed RSRectWidth( TRect ) function
//          - Added CenterRect( var R: TRect; const Bounds: TRect ) function 
//  07/2013 T. Grubb
//          - Fixed bug in DrawArrow which draw the caption whether aoDrawCaption
//            was in Options or not
//          - Changed CharInSet expressions to use RSCharInSet to avoid deprecation
//            warnings from XE4+
//          - Added TRSVector record type and functions
//          - Added TRSVectorArray type
//          - Added TRSMatrix record type and functions
//          - Added TRSPolygon type
//          - Added TRSCubicBezier type
//          - Added RSRectHeight function
//          - Added RSRectWidth function
//          - Added TRSPoint Class Operator methods (Pre-Delphi XE2)
//          - Added TRSRect.Fit method for Pre-Delphi XE2
//          - Added RSPoint( const aVector: TRSVector ) function
//          - Added RSVector( const X, Y: Single; const W: Single = 1.0 ) function
//          - Added RSVector( const aPoint: TRSPoint; const W: Single = 1.0 ) function
//          - Added RSVector( const aVector: TVector ) function for XE2+
//          - Added RSMatrix( const aMatrix: TMatrix ) function for XE2+
//          - Added Vector( const aVector: TRSVector ) function for XE2+
//          - Added Matrix( const aMatrix: TRSMatrix ) function for XE2+
//          - Added RSVectorAngleCosine(const v1, v2: TRSVector) function
//          - Added RSVectorDotProduct(const v1, v2: TRSVector) function
//          - Added RSVectorCrossProductZ(const v1, v2: TRSVector) function
//          - Added RSVectorAngle(const V, N: TRSVector) function
//          - Added RSVectorLength( const V: TRSVector ) function
//          - Added RSVectorNorm(const V: TRSVector) function
//          - Added RSMatrixMultiply(const M1, M2: TRSMatrix) function
//          - Added RSMatrixDeterminant(const M: TRSMatrix) function
//          - Added AdjointRSMatrix(var M function
//          - Added ScaleRSMatrix(var M: TRSMatrix; const factor function
//          - Added InvertRSMatrix(var M function
//          - Added CreateRotationRSMatrix(const Angle: Single) function
//          - Added CreateScaleRSMatrix(const ScaleX, ScaleY: Single) function
//          - Added CreateTranslateRSMatrix(const DX, DY: Single) function
//          - Added CenterRect( var R: TRSRect; const Bounds: TRSRect ) function
//          - Added OffsetRSRect(var R: TRSRect; DX, DY: Single) function
//          - Added RSRectHeight( const ARect: TRSRect ) function
//          - Added RSRectWidth( const ARect: TRSRect ) function
//          - Added RSInflateRect( var R: TRSRect; const DX, DY: Single ) procedure
//          - Added WrapCanvasText function for NEXTGEN compiler
//          - Added RSClosePolygon constant
//          - Added RSIdentityMatrix constant
//  11/2012 T. Grubb
//          - Added TExtCanvasPixel type for increased precision in CanvasXXX functions
//          - Changed CanvasDiv, CanvasTrunc, and CanvasRound to use higher-precision
//            data types (doubles and Int64)
//          - Added ProgressCircle function, which was removed until updated for VCL and FMX
//          - Added ProgressPie function, which was removed until updated for VCL and FMX
//          - Added ProgressRect function, which was removed until updated for VCL and FMX
//          - Added IndexToColor function, which was removed until updated for VCL and FMX
//  09/2012 T. Grubb
//          - Redefined clxColor constants to use TAlphaColors.White, etc instead
//            of claWhite which XE3 deleted (sigh)
//  04/2012 T. Grubb
//          - Added more clxColor constants
//          - Added TRSColorScheme type
//          - Added TColorList constants for the different color schemes
//          - Added PointInPath functions
//          - Added ScaleRect( TRect, Single, Single ) function
//          - Added ScalePoints functions
//          - Added overloaded CenterPoint( TRSPoints/TPoints ) functions
//          - Added Filled parameter for DrawBezierPolygon function
//          - Added Opacity parameter for DrawBezierPolygon function (FMX Only)
//  03/2012 T. Grubb
//          - Added support for FireMonkey
//          - Fixed bug in DrawShape for ss3DCircle where an ellipse was drawn
//            instead
//          - Fixed bug in DrawShape where the shape would be flipped if the
//            rectangle was not correctly specified (e.g., Left > Right)
//          - Fixed bug in DrawPieSlice where the slice would be flipped if the
//            rectangle was not correctly specified (e.g., Left > Right)
//          - Fixed Divide by Zero bug in OffsetPoint function
//          - Added RSPointsToPoints and PointsToRSPoints functions
//          - Added Opacity parameter for DrawShape function (FMX Only)
//          - Added Opacity parameter for DirectionGradient function (FMX Only)
//            Note: LineSize is ignored in FMX for diagonal gradients
//          - Added Opacity parameter for LineGradient function (FMX Only)
//          - Added Opacity parameter for EllipseGradient function (FMX Only)
//          - Added Opacity parameter for RectGradient function (FMX Only)
//          - Added Opacity parameter for DrawArrow functions (FMX Only)
//          - Added Opacity parameter for DrawPieSlice functions (FMX Only)
//          - Added SetGradientColors function (FMX Only)
//          - Added DrawGradient function (FMX Only)
//          - Added TVclCanvasHelper, class helper for TCanvas (VCL), which
//            adds functions that make the transition to FMX and back easier
//          - Added TCanvasRect type that changes for FMX(TRectF)/Vcl(TRect)
//          - Added TCanvasPoint type that changes for FMX(TPointF)/Vcl(TPoint)
//          - Added TCanvasPoints type that changes for FMX(Array of TPointF)/Vcl(Array of TPoint)
//          - Added TCanvasSize type that changes for FMX(TSizeF)/Vcl(TSize)
//          - Added TCanvasColor type that changes for FMX(TAlphaColor)/Vcl(TColor)
//          - Added TColors type that changes for FMX(array of TAlphaColor)/Vcl(array of TColor)
//          - Added TCanvasPixel type that changes for FMX(Single)/Vcl(Integer)
//          - Added clxXXX colors constants that change for FMX/Vcl
//          - Added CanvasRect function that changes for FMX/Vcl
//          - Added CanvasPoint function that changes for FMX/Vcl
//          - Added CanvasMax function that changes for FMX/Vcl
//          - Added CanvasMin function that changes for FMX/Vcl
//          - Added CanvasRound function that changes for FMX/Vcl
//          - Added CanvasTrunc function that changes for FMX/Vcl
//          - Added CanvasDiv function that changes for FMX/Vcl
//          - Added CanvasCenterPoint function that changes for FMX/Vcl
//  01/2009 T. Grubb
//          - Turned off range checks around ColorToRGB function to avoid range
//            check errors
//          - Fixed DrawShape for beveled rectangles
//          - Added ss3DRectangle, ss3DEllipse,
//            ss3DDiamond, ss3DLeftTriangle, ss3DRightTriangle,
//            ss3DUpTriangle, ss3DDownTriangle, ss3DRoundRect,
//            ss3DSquare, ss3DCircle, ss3DRoundSquare to TRSShapeStyle
//          - Added Frame3DExt function for creating the 3D frame for an rectangle
//          - Added Frame3DEllipse function for creating the 3D frame for an ellipse
//          - Added TRSShapeStyles type
//          - Added ShapeStyles3D constant
//          - Added InCenterPoint function to get InCenter Point of a triangle
//          - Changed DrawShape functions to return the non-beveled part of the
//            drawing rectangle
//          - Added CenterRect function, which creates a rectangle centered around a point
//          - Fixed IntersectRect function to detect intersection when intersection
//            is single point that equals
//  11/2008 T. Grubb
//          - Changed set expressions to use CharInSet
//  11/2008 T. Grubb
//          - Fixed small warning generated in D2009 in RotateFont function
//  06/2007 T. Grubb
//          - Added DrawTextLabel function
//          - Added GetMultilineTextHeight function
//          - Added WrapCanvasText functions
//  04/2006 T. Grubb
//          - Added BezierPolygonPoints function
//          - Added DrawBezierPolygon function
//  10/2006 T. Grubb
//          - Added InterpolateColor function
//  12/2005 T. Grubb
//          - Added overloaded DrawArrow function that accepts a series of points
//          - Added TRSPoints
//          - Added TConnectionPoint
//          - Added GetPointOnRect and GetRSPointOnRSRect functions
//  10/2005 T. Grubb
//          - Added FixRect function
//          - Added PtInRSRect functions
//          - Added IntersectRect function for TRSRect
//  09/2005 T. Grubb
//          - Fixed DirectionGradient to work for diagonal gradients (number of
//            colors greater than 2 and LineSize are ignored) in Win32 ONLY
//          - Added Arrow Width to DrawArrow routine
//  06/2005 T. Grubb
//          - Changed TRSPoint and TRSRect to be packed records
//          - Changed TRSPoint and TRSRect to be safe types in .NET
//  03/2005 T. Grubb
//          - Added RSCenterPoint functions
//          - Added DrawPieSlice procedure
//          - Added DrawArrow procedure and TArrowOption types
//          - Changed Gradient functions to accept an open array of TColor,
//            instead of TColors
//          - Moved TObjectSize from Game.pas to RSGraphics.pas
//          - Created hierarchy of TRSCustomRect, TRSObjectRect, and TObjectSize
//          - Added TRSRect, TRSPoint and functions to manipulate them
//          - Added overloaded DrawShape procedures and TRSShapeStyle type
//          - Fixed LineGradient to handle "reversed" TRects (left > Right, etc)
//          - Fixed RotateFont to change font name if font is not a true type
//            (and wouldn't thus rotate)
//          - Added GetRotatedTextExtent function (Delphi's Canvas.TextExtent
//            does not work for rotated fonts)
//  05/2004 J. Hosler
//          - Corrected LineGradient, EllipseGradient and RectGradient so that
//            they properly draw the gradient if the CanvasRect is not anchored
//            at (0,0).
//          - Added Intensity functions
//  04/2004 J. Hosler
//          - Initial Version
//
//      File Contents:
//          DrawShape -
//          RotateFont -
//          LineGradient -
//          EllipseGradient -
//          RectGradient -
//          ProgressCircle -
//          ProgressPie -
//          ProgressRect -
//          IndexToColor -
//          DoubleToByte -
//          ColorToRGB -
//          RGBToColor -
//          PaletteEntryToRGBQuad -
//          PaletteEntriesToRGBQuads -
//          RGBQuadToPaletteEntry -
//          RGBQuadsToPaletteEntries -
//          TRSPathData         VCL/FMX version of FMX TPathData, Represents a series
//                              of connected curves and lines
//
//
//--- Warning ----------------------------------------------------------------
//    This software is property of RiverSoftAVG. Unauthorized use or
//    duplication of this software is strictly prohibited. Authorized users
//    are subject to the following restrictions:
//      * RiverSoftAVG is not responsible for
//        any consequence of the use of this software.
//      * The origin of this software must not be misrepresented either by
//        explicit claim or by omission.
//      * Altered versions of this software must be plainly marked as such.
//      * This notice may not be removed or altered.
//
//       1998-2015, Thomas G. Grubb & Jeffrey C. Hosler
//
//=== End File Prolog ========================================================


interface

{$INCLUDE RSDefines.inc}

uses
  Classes, SysUtils, RSConstants,
{$IFDEF DELPHI16PLUS}
  UITypes,
{$ENDIF}
{$IFDEF DELPHI20PLUS}
  System.Math.Vectors,
{$ENDIF}
{$IFDEF FMX}
  FMX.Types, FMX.RS.CanvasHelper,
  {$IFDEF DELPHI19PLUS}FMX.Graphics,{$ENDIF}
{$ELSE}
  Windows, Graphics,
{$ENDIF}
  Types;

type
  TGradientOrientation=(diHorizontal, diVertical);
  TGradientDirection=( gdDownRight, gdDown, gdDownLeft, gdRight, gdLeft,
                       gdUpRight, gdUp, gdUpLeft );
  TGradientDirDeltas = Array[TGradientDirection] of Integer;
  /// <summary>
  ///   Defines a geometric shape
  /// </summary>
  TRSShapeStyle = (
    /// <summary>
    ///   Rectangle
    /// </summary>
    ssRectangle,
    /// <summary>
    ///   Ellipse
    /// </summary>
    ssEllipse,
    /// <summary>
    ///   Image
    /// </summary>
    ssImage,
    /// <summary>
    ///   Diamon
    /// </summary>
    ssDiamond,
    /// <summary>
    ///   Left Triangle
    /// </summary>
    ssLeftTriangle,
    /// <summary>
    ///   Right Triangle
    /// </summary>
    ssRightTriangle,
    /// <summary>
    ///   Up Triangle
    /// </summary>
    ssUpTriangle,
    /// <summary>
    ///   Down Triangle
    /// </summary>
    ssDownTriangle,
    /// <summary>
    ///   Rounded Rectangle
    /// </summary>
    ssRoundRect,
    /// <summary>
    ///   Square
    /// </summary>
    ssSquare,
    /// <summary>
    ///   Circle
    /// </summary>
    ssCircle,
    /// <summary>
    ///   Rounded Square
    /// </summary>
    ssRoundSquare,
    /// <summary>
    ///   Rectangle with a bevel
    /// </summary>
    ss3DRectangle,
    /// <summary>
    ///   Ellipse with a bevel
    /// </summary>
    ss3DEllipse,
    /// <summary>
    ///   Image with a bevel
    /// </summary>
    ss3DImage,
    /// <summary>
    ///   Diamond with a bevel
    /// </summary>
    ss3DDiamond,
    /// <summary>
    ///   Left Triangle with a bevel
    /// </summary>
    ss3DLeftTriangle,
    /// <summary>
    ///   Right Triangle with a bevel
    /// </summary>
    ss3DRightTriangle,
    /// <summary>
    ///   Up Triangle with a bevel
    /// </summary>
    ss3DUpTriangle,
    /// <summary>
    ///   Down Triangle with a bevel
    /// </summary>
    ss3DDownTriangle,
    /// <summary>
    ///   Rounded Rectangle with a bevel
    /// </summary>
    ss3DRoundRect,
    /// <summary>
    ///   Square with a bevel
    /// </summary>
    ss3DSquare,
    /// <summary>
    ///   Circle with a bevel
    /// </summary>
    ss3DCircle,
    /// <summary>
    ///   Rounded Square with a bevel
    /// </summary>
    ss3DRoundSquare );
  TRSShapeStyles = set of TRSShapeStyle;

const
  GradDirXDeltas: TGradientDirDeltas = ( 1, 0, -1, 1, -1, 1, 0, -1 );
  GradDirYDeltas: TGradientDirDeltas = ( 1, 1, 1, 0, 0, -1, -1, -1 );
  ShapeStyles3D: TRSShapeStyles = [ ss3DRectangle, ss3DEllipse,
                    ss3DDiamond, ss3DLeftTriangle, ss3DRightTriangle,
                    ss3DUpTriangle, ss3DDownTriangle, ss3DRoundRect,
                    ss3DSquare, ss3DCircle, ss3DRoundSquare ];

type
{$IFDEF FMX}
  // "native" types used by the framework
  TCanvasColor = TAlphaColor;
  TTileGraphic = TBitmap;
  TRSPen = TRSStroke;
  TRSPicture = TBitmap;
  TCanvasPixel = Single;
  TExtCanvasPixel = Double;
  TCanvasRect = TRectF;
  TCanvasPoint = TPointF;
  TCanvasPoints = Array of TPointF;
  TCanvasSize = TSizeF;
  TColors = array of TAlphaColor;

const
  clxBtnHighlight = TAlphaColors.Lightgray;
  clxBtnShadow = TAlphaColors.Black;
  clxActiveBorder = TColorRec.cACTIVEBORDER;
  clxInactiveBorder = TColorRec.cINACTIVEBORDER;
  clxBtnFace = TColorRec.cBTNFACE;
  clxDefaultColor = TAlphaColors.Null;
  clxAliceblue = TAlphaColors.Aliceblue;
  clxAntiquewhite = TAlphaColors.Antiquewhite;
  clxAqua = TAlphaColors.Aqua;
  clxAquamarine = TAlphaColors.Aquamarine;
  clxAzure = TAlphaColors.Azure;
  clxBeige = TAlphaColors.Beige;
  clxBisque = TAlphaColors.Bisque;
  clxBlack = TAlphaColors.Black;
  clxBlanchedalmond = TAlphaColors.Blanchedalmond;
  clxBlue = TAlphaColors.Blue;
  clxBlueviolet = TAlphaColors.Blueviolet;
  clxBrown = TAlphaColors.Brown;
  clxBurlywood = TAlphaColors.Burlywood;
  clxCadetblue = TAlphaColors.Cadetblue;
  clxChartreuse = TAlphaColors.Chartreuse;
  clxChocolate = TAlphaColors.Chocolate;
  clxCoral = TAlphaColors.Coral;
  clxCornflowerblue = TAlphaColors.Cornflowerblue;
  clxCornsilk = TAlphaColors.Cornsilk;
  clxCrimson = TAlphaColors.Crimson;
  clxCyan = TAlphaColors.Cyan;
  clxDarkBlue = TAlphaColors.Darkblue;
  clxDarkCyan = TAlphaColors.Darkcyan;
  clxDarkGoldenrod = TAlphaColors.Darkgoldenrod;
  clxDarkGray = TAlphaColors.Darkgray;
  clxDarkGreen = TAlphaColors.Darkgreen;
  clxDarkKhaki = TAlphaColors.Darkkhaki;
  clxDarkMagenta = TAlphaColors.Darkmagenta;
  clxDarkOliveGreen = TAlphaColors.Darkolivegreen;
  clxDarkOrange = TAlphaColors.Darkorange;
  clxDarkOrchid = TAlphaColors.Darkorchid;
  clxDarkRed = TAlphaColors.Darkred;
  clxDarkSalmon = TAlphaColors.Darksalmon;
  clxDarkSeagreen = TAlphaColors.Darkseagreen;
  clxDarkSlateBlue = TAlphaColors.Darkslateblue;
  clxDarkSlateGray = TAlphaColors.Darkslategray;
  clxDarkTurquoise = TAlphaColors.Darkturquoise;
  clxDarkViolet = TAlphaColors.Darkviolet;
  clxDeepPink = TAlphaColors.Deeppink;
  clxDeepSkyBlue = TAlphaColors.Deepskyblue;
  clxDimGray = TAlphaColors.Dimgray;
  clxDodgerBlue = TAlphaColors.Dodgerblue;
  clxFireBrick = TAlphaColors.Firebrick;
  clxFloralWhite = TAlphaColors.Floralwhite;
  clxForestGreen = TAlphaColors.Forestgreen;
  clxFuchsia = TAlphaColors.Fuchsia;
  clxGainsboro = TAlphaColors.Gainsboro;
  clxGhostWhite = TAlphaColors.Ghostwhite;
  clxGold = TAlphaColors.Gold;
  clxGoldenrod = TAlphaColors.Goldenrod;
  clxGray = TAlphaColors.Gray;
  clxGreen = TAlphaColors.Green;
  clxGreenYellow = TAlphaColors.Greenyellow;
  clxHoneydew = TAlphaColors.Honeydew;
  clxHotPink = TAlphaColors.Hotpink;
  clxIndianRed = TAlphaColors.Indianred;
  clxIndigo = TAlphaColors.Indigo;
  clxIvory = TAlphaColors.Ivory;
  clxKhaki = TAlphaColors.Khaki;
  clxLavender = TAlphaColors.Lavender;
  clxLavenderBlush = TAlphaColors.Lavenderblush;
  clxLawnGreen = TAlphaColors.Lawngreen;
  clxLemonChiffon = TAlphaColors.Lemonchiffon;
  clxLightBlue = TAlphaColors.Lightblue;
  clxLightCoral = TAlphaColors.Lightcoral;
  clxLightCyan = TAlphaColors.Lightcyan;
  clxLightGoldenrodYellow = TAlphaColors.Lightgoldenrodyellow;
  clxLightGray = TAlphaColors.Lightgray;
  clxLightGreen = TAlphaColors.Lightgreen;
  clxLightPink = TAlphaColors.Lightpink;
  clxLightSalmon = TAlphaColors.Lightsalmon;
  clxLightSeaGreen = TAlphaColors.Lightseagreen;
  clxLightSkyBlue = TAlphaColors.Lightskyblue;
  clxLightSlateGray = TAlphaColors.Lightslategray;
  clxLightSteelBlue = TAlphaColors.Lightsteelblue;
  clxLightYellow = TAlphaColors.Lightyellow;
  clxLime = TAlphaColors.Lime;
  clxLimegreen = TAlphaColors.Limegreen;
  clxLinen = TAlphaColors.Linen;
  clxMagenta = TAlphaColors.Magenta;
  clxMaroon = TAlphaColors.Maroon;
  clxMediumAquamarine = TAlphaColors.Mediumaquamarine;
  clxMediumBlue = TAlphaColors.Mediumblue;
  clxMediumOrchid = TAlphaColors.Mediumorchid;
  clxMediumPurple = TAlphaColors.Mediumpurple;
  clxMediumSeagreen = TAlphaColors.Mediumseagreen;
  clxMediumSlateBlue = TAlphaColors.Mediumslateblue;
  clxMediumSpringGreen = TAlphaColors.Mediumspringgreen;
  clxMediumTurquoise = TAlphaColors.Mediumturquoise;
  clxMediumVioletRed = TAlphaColors.Mediumvioletred;
  clxMidnightBlue = TAlphaColors.Midnightblue;
  clxMintCream = TAlphaColors.Mintcream;
  clxMistyRose = TAlphaColors.Mistyrose;
  clxMoccasin = TAlphaColors.Moccasin;
  clxNavajoWhite = TAlphaColors.Navajowhite;
  clxNavy = TAlphaColors.Navy;
  clxOldLace = TAlphaColors.Oldlace;
  clxOlive = TAlphaColors.Olive;
  clxOliveDrab = TAlphaColors.Olivedrab;
  clxOrange = TAlphaColors.Orange;
  clxOrangeRed = TAlphaColors.Orangered;
  clxOrchid = TAlphaColors.Orchid;
  clxPaleGoldenrod = TAlphaColors.Palegoldenrod;
  clxPaleGreen = TAlphaColors.Palegreen;
  clxPaleTurquoise = TAlphaColors.Paleturquoise;
  clxPaleVioletRed = TAlphaColors.Palevioletred;
  clxPapayaWhip = TAlphaColors.Papayawhip;
  clxPeachPuff = TAlphaColors.Peachpuff;
  clxPeru = TAlphaColors.Peru;
  clxPink = TAlphaColors.Pink;
  clxPlum = TAlphaColors.Plum;
  clxPowderBlue = TAlphaColors.Powderblue;
  clxPurple = TAlphaColors.Purple;
  clxRed = TAlphaColors.Red;
  clxRosyBrown = TAlphaColors.Rosybrown;
  clxRoyalBlue = TAlphaColors.Royalblue;
  clxSaddleBrown = TAlphaColors.Saddlebrown;
  clxSalmon = TAlphaColors.Salmon;
  clxSandyBrown = TAlphaColors.Sandybrown;
  clxSeagreen = TAlphaColors.Seagreen;
  clxSeashell = TAlphaColors.Seashell;
  clxSienna = TAlphaColors.Sienna;
  clxSilver = TAlphaColors.Silver;
  clxSkyBlue = TAlphaColors.Skyblue;
  clxSlateBlue = TAlphaColors.Slateblue;
  clxSlateGray = TAlphaColors.Slategray;
  clxSnow = TAlphaColors.Snow;
  clxSpringGreen = TAlphaColors.Springgreen;
  clxSteelBlue = TAlphaColors.Steelblue;
  clxTan = TAlphaColors.Tan;
  clxTeal = TAlphaColors.Teal;
  clxThistle = TAlphaColors.Thistle;
  clxTomato = TAlphaColors.Tomato;
  clxTurquoise = TAlphaColors.Turquoise;
  clxViolet = TAlphaColors.Violet;
  clxWheat = TAlphaColors.Wheat;
  clxWhite = TAlphaColors.White;
  clxWhiteSmoke = TAlphaColors.Whitesmoke;
  clxYellow = TAlphaColors.Yellow;
  clxYellowGreen = TAlphaColors.Yellowgreen;

{$ELSE}
  TCanvasColor = TColor;
  TTileGraphic = TGraphic;
  TRSPen = TPen;
  TRSPicture = TPicture;
  TCanvasPixel = Integer;
  TExtCanvasPixel = Int64;
  TCanvasRect = TRect;
  TCanvasPoint = TPoint;
  TCanvasPoints = Array of TPoint;
  TCanvasSize = TSize;
  TColors = array of TColor;

const
  clxBtnHighlight = clBtnHighlight;
  clxBtnShadow = clBtnShadow;
  clxActiveBorder = clActiveBorder;
  clxInactiveBorder = clInactiveBorder;
  clxBtnFace = clBtnFace;
  clxDefaultColor = -1;
  clxAliceBlue = clWebAliceblue;
  clxAntiqueWhite = clWebAntiquewhite;
  clxAqua = clWebAqua;
  clxAquamarine = clWebAquamarine;
  clxAzure = clWebAzure;
  clxBeige = clWebBeige;
  clxBisque = clWebBisque;
  clxBlack = clBlack;
  clxBlanchedAlmond = clWebBlanchedalmond;
  clxBlue = clWebBlue;
  clxBlueViolet = clWebBlueviolet;
  clxBrown = clWebBrown;
  clxBurlywood = clWebBurlywood;
  clxCadetBlue = clWebCadetblue;
  clxChartreuse = clWebChartreuse;
  clxChocolate = clWebChocolate;
  clxCoral = clWebCoral;
  clxCornflowerBlue = clWebCornflowerblue;
  clxCornsilk = clWebCornsilk;
  clxCrimson = clWebCrimson;
  clxCyan = clWebCyan;
  clxDarkBlue = clWebDarkblue;
  clxDarkCyan = clWebDarkcyan;
  clxDarkGoldenrod = clWebDarkgoldenrod;
  clxDarkGray = clWebDarkgray;
  clxDarkGreen = clWebDarkgreen;
  clxDarkKhaki = clWebDarkkhaki;
  clxDarkMagenta = clWebDarkmagenta;
  clxDarkOliveGreen = clWebDarkolivegreen;
  clxDarkOrange = clWebDarkorange;
  clxDarkOrchid = clWebDarkorchid;
  clxDarkRed = clWebDarkred;
  clxDarkSalmon = clWebDarksalmon;
  clxDarkSeagreen = clWebDarkseagreen;
  clxDarkSlateBlue = clWebDarkslateblue;
  clxDarkSlateGray = clWebDarkslategray;
  clxDarkTurquoise = clWebDarkturquoise;
  clxDarkViolet = clWebDarkviolet;
  clxDeepPink = clWebDeeppink;
  clxDeepSkyBlue = clWebDeepskyblue;
  clxDimGray = clWebDimgray;
  clxDodgerBlue = clWebDodgerblue;
  clxFireBrick = clWebFirebrick;
  clxFloralWhite = clWebFloralwhite;
  clxForestGreen = clWebForestgreen;
  clxFuchsia = clWebFuchsia;
  clxGainsboro = clWebGainsboro;
  clxGhostWhite = clWebGhostwhite;
  clxGold = clWebGold;
  clxGoldenrod = clWebGoldenrod;
  clxGray = clWebGray;
  clxGreen = clWebGreen;
  clxGreenYellow = clWebGreenyellow;
  clxHoneydew = clWebHoneydew;
  clxHotPink = clWebHotpink;
  clxIndianRed = clWebIndianred;
  clxIndigo = clWebIndigo;
  clxIvory = clWebIvory;
  clxKhaki = clWebKhaki;
  clxLavender = clWebLavender;
  clxLavenderBlush = clWebLavenderblush;
  clxLawnGreen = clWebLawngreen;
  clxLemonChiffon = clWebLemonchiffon;
  clxLightBlue = clWebLightblue;
  clxLightCoral = clWebLightcoral;
  clxLightCyan = clWebLightcyan;
  clxLightGoldenrodyellow = clWebLightgoldenrodyellow;
  clxLightGray = clWebLightgrey;
  clxLightGreen = clWebLightgreen;
  clxLightPink = clWebLightpink;
  clxLightSalmon = clWebLightsalmon;
  clxLightSeagreen = clWebLightseagreen;
  clxLightSkyBlue = clWebLightskyblue;
  clxLightSlateGray = clWebLightslategray;
  clxLightSteelBlue = clWebLightsteelblue;
  clxLightYellow = clWebLightyellow;
  clxLime = clWebLime;
  clxLimeGreen = clWebLimegreen;
  clxLinen = clWebLinen;
  clxMagenta = clWebMagenta;
  clxMaroon = clWebMaroon;
  clxMediumAquamarine = clWebMediumaquamarine;
  clxMediumBlue = clWebMediumblue;
  clxMediumOrchid = clWebMediumorchid;
  clxMediumPurple = clWebMediumpurple;
  clxMediumSeagreen = clWebMediumseagreen;
  clxMediumSlateBlue = clWebMediumslateblue;
  clxMediumSpringGreen = clWebMediumspringgreen;
  clxMediumTurquoise = clWebMediumturquoise;
  clxMediumVioletRed = clWebMediumvioletred;
  clxMidnightBlue = clWebMidnightblue;
  clxMintCream = clWebMintcream;
  clxMistyRose = clWebMistyrose;
  clxMoccasin = clWebMoccasin;
  clxNavajoWhite = clWebNavajowhite;
  clxNavy = clWebNavy;
  clxOldLace = clWebOldlace;
  clxOlive = clWebOlive;
  clxOliveDrab = clWebOlivedrab;
  clxOrange = clWebOrange;
  clxOrangeRed = clWebOrangered;
  clxOrchid = clWebOrchid;
  clxPaleGoldenrod = clWebPalegoldenrod;
  clxPaleGreen = clWebPalegreen;
  clxPaleTurquoise = clWebPaleturquoise;
  clxPaleVioletRed = clWebPalevioletred;
  clxPapayaWhip = clWebPapayawhip;
  clxPeachPuff = clWebPeachpuff;
  clxPeru = clWebPeru;
  clxPink = clWebPink;
  clxPlum = clWebPlum;
  clxPowderBlue = clWebPowderblue;
  clxPurple = clWebPurple;
  clxRed = clWebRed;
  clxRosyBrown = clWebRosybrown;
  clxRoyalBlue = clWebRoyalblue;
  clxSaddleBrown = clWebSaddlebrown;
  clxSalmon = clWebSalmon;
  clxSandyBrown = clWebSandybrown;
  clxSeagreen = clWebSeagreen;
  clxSeashell = clWebSeashell;
  clxSienna = clWebSienna;
  clxSilver = clWebSilver;
  clxSkyBlue = clWebSkyblue;
  clxSlateBlue = clWebSlateblue;
  clxSlateGray = clWebSlategray;
  clxSnow = clWebSnow;
  clxSpringGreen = clWebSpringgreen;
  clxSteelBlue = clWebSteelblue;
  clxTan = clWebTan;
  clxTeal = clWebTeal;
  clxThistle = clWebThistle;
  clxTomato = clWebTomato;
  clxTurquoise = clWebTurquoise;
  clxViolet = clWebViolet;
  clxWheat = clWebWheat;
  clxWhite = clWebWhite;
  clxWhiteSmoke = clWebWhitesmoke;
  clxYellow = clWebYellow;
  clxYellowGreen = clWebYellowgreen;

{$ENDIF}

type
  TRSGfxFloat = Single;
  TRSColorScheme = (csCustom, csDefaultColors, csPastels, csBlackandWhite, csGrays,
    csSouthwest, csReds, csGreens, csBlues, csViolets, csDystopia, csAntique);
  TColorList = Array[0..9] of TCanvasColor;
  ERSGfxException = class(ERSException);
  TConnectionPoint = ( cpTop, cpTopRight, cpRight, cpBottomRight,
                       cpBottom, cpBottomLeft, cpLeft, cpTopLeft );
  /// <summary>
  ///   Defines the horizontal and vertical alignment for a text output inside
  ///   an area
  /// </summary>
  T2DTextAlignment = (
    /// <summary>
    ///   Align text to the top vertically and align left horizontally <br />
    /// </summary>
    taTopLeft,
    /// <summary>
    ///   Center the text horizontally and put it on the tp[ vertically <br />
    /// </summary>
    taTopCenter,
    /// <summary>
    ///   Align text to the top vertically and align right horizontally
    /// </summary>
    taTopRight,
    /// <summary>
    ///   Center the text vertically and align left
    ///   horizontally <br />
    /// </summary>
    taCenterLeft,
    /// <summary>
    ///   Center the text horizontally and vertically <br />
    /// </summary>
    taCentered,
    /// <summary>
    ///   Center the text vertically and align right horizontally
    /// </summary>
    taCenterRight,
    /// <summary>
    ///   Align text to the bottom vertically and align left
    ///   horizontally
    /// </summary>
    taBottomLeft,
    /// <summary>
    ///   Center the text horizontally and put it on the bottom vertically
    /// </summary>
    taBottomCenter,
    /// <summary>
    ///   Align text to the bottom vertically and align right horizontally <br />
    /// </summary>
    taBottomRight );
  /// <summary>
  ///   Defines the options for where and how the arrow head (the triangle
  ///   portion) of an arrow should be drawn. <br />
  /// </summary>
  /// <seealso cref="TArrowOptions" />
  /// <seealso cref="DrawArrow" />
  TArrowOption = (
    /// <summary>
    ///   When an arrow has a direction (e.g., a beginning and an end), this
    ///   option directs that the arrow head should be drawn at the start of
    ///   the line. <br />
    /// </summary>
    aoArrowAtStart,
    /// <summary>
    ///   When an arrow has a direction (e.g., a beginning and an end), this
    ///   option directs that the arrow head should be drawn at the end of the
    ///   line. <br />
    /// </summary>
    aoArrowAtEnd,
    /// <summary>
    ///   This option controls whether the arrow head should be drawn as an
    ///   outline or filled in. <br />
    /// </summary>
    aoFillHead,
    /// <summary>
    ///   This option controls whether the arrow head should be drawn closed,
    ///   e.g., should a line be draw across the back of the arrow head to make
    ///   a closed shape. <br />
    /// </summary>
    aoClosedHead,
    /// <summary>
    ///   This options controls whether the arrow should have a caption drawn
    ///   next to it. This option is not used by the DrawArrow function.
    /// </summary>
    aoDrawCaption);
  /// <summary>
  ///   Defines the set of options for where and how the arrow head (the
  ///   triangle portion) of an arrow should be drawn.
  /// </summary>
  /// <seealso cref="DrawArrow" />
  /// <seealso cref="TArrowOption" />
  TArrowOptions = set of TArrowOption;
  TRSShapeType = (sRect, sCircle, sIsoscelesTriangle, sRightTriangle);
  TProgressDirection = (pdForward, pdBackward);
  TProgressDisplayType = (dtAllProgress, dtCurrentProgress);
  TProgressOrientation = (poHorizontal, poVertical);
  TRSShapeOrientation = (soUp, soDown, soLeft, soRight);
  TPathClickLocation = (clOutsidePath, clInsidePath, clOnPath);

{$IFDEF DELPHI16PLUS}
  TRSSize = TSizeF;
{$ELSE}
  TRSSize = packed record
    cx: Single;
    cy: Single;
  end;
{$ENDIF}

{$IFDEF DELPHI16PLUS}
  TRSPoint = TPointF;
{$ELSE}
  TRSPoint = packed record
     X: TRSGfxFloat;
     Y: TRSGfxFloat;
     //operator overloads
     class operator Equal(const Lhs, Rhs: TRSPoint): Boolean;
     class operator NotEqual(const Lhs, Rhs: TRSPoint): Boolean;
     class operator Add(const Lhs, Rhs: TRSPoint): TRSPoint;
     class operator Subtract(const Lhs, Rhs: TRSPoint): TRSPoint;
     class operator Multiply(const Lhs, Rhs: TRSPoint): TRSPoint;

{$IFDEF CLR}
  public
    class function Create(const AX, AY: TRSGfxFloat): TRSPoint; static;
    class function Empty: TRSPoint; static;
    function ToString: string; override;

    function Left: TRSGfxFloat;
    function Top: TRSGfxFloat;

    function IsEmpty: Boolean;
    function IsInvalid: Boolean;
    function Equals(const AX, AY: TRSGfxFloat): Boolean; overload;
    function Equals(const Value: TRSPoint): Boolean; overload;
{$ENDIF}
  end;
{$ENDIF}

{$IFDEF DELPHI16PLUS}
  TRSRect = TRectF;
{$ELSE}
  TRSRect = packed record
{$IFNDEF CLR}
    // Current rect enters into the boundaries of the rectangle with keeping
    // aspect ratio and return ratio's value

     function Fit(BoundsRect: TRSRect): Single;
     case Integer of
          0: (Left, Top, Right, Bottom: TRSGfxFloat);
          1: (TopLeft, BottomRight: TRSPoint);

{$ELSE}
    Left, Top, Right, Bottom: TRSGfxFloat;
  strict private
    class function IfThen(const ATest: Boolean; const ATrue, AFalse: TRSGfxFloat): TRSGfxFloat; static;
  public
    class function Create(const ALeft, ATop, ARight, ABottom: TRSGfxFloat): TRSRect; overload; static;
    class function Create(const ATopLeft, ABottomRight: TPoint): TRSRect; overload; static;
    class function Create(const ATopLeft, ABottomRight: TRSPoint): TRSRect; overload; static;
    class function Create(const ATopLeft: TPoint; const ASize: TSize): TRSRect; overload; static;
    class function Create(const ATopLeft: TRSPoint; const ASize: TSize): TRSRect; overload; static;
    class function Bounds(const ALeft, ATop, AWidth, AHeight: TRSGfxFloat): TRSRect; static;
    class function Empty: TRSRect; static;
    function ToString: string; override;

    function FixRect: TRSRect;
    function get_TopLeft: TRSPoint;
    procedure set_TopLeft(const Value: TRSPoint);
    function get_BottomRight: TRSPoint;
    procedure set_BottomRight(const Value: TRSPoint);


    property TopLeft: TRSPoint read get_TopLeft write set_TopLeft;
    property BottomRight: TRSPoint read get_BottomRight write set_BottomRight;
    function GetWidth: TRSGfxFloat;
    function GetHeight: TRSGfxFloat;
    function GetSize: TSize;
    function GetCenterPoint: TRSPoint;

    function IsEmpty: Boolean;
    function Equals(const Value: TRSRect): Boolean; overload;
    function Equals(const Value: TSize): Boolean; overload;
    function Contains(const X, Y: TRSGfxFloat): Boolean; overload;
    function Contains(const Value: TPoint): Boolean; overload;
    function Contains(const Value: TRSPoint): Boolean; overload;
    function Contains(const Value: TRSRect): Boolean; overload;
    function Intersects(const Other: TRSRect): Boolean; overload;

    function GetIntersect(const Other: TRSRect): TRSRect; overload;
    function GetIntersect(const Other: TRSRect; out AResult: TRSRect): Boolean; overload;

    function GetUnion(const Other: TRSRect): TRSRect; overload;
    function GetUnion(const Other: TRSRect; out AResult: TRSRect): Boolean; overload;

    function GetOffset(const X, Y: TRSGfxFloat): TRSRect; overload;
    function GetOffset(const Value: TRSPoint): TRSRect; overload;

    function GetInflate(const DeltaWidth, DeltaHeight: TRSGfxFloat): TRSRect; overload;
    function GetInflate(const Delta: TRSPoint): TRSRect; overload;
{$ENDIF}
  end;
{$ENDIF}
  TRSRectEvent = procedure (Sender: TObject; aRect: TRSRect) of object;
  TRSGetRectEvent = procedure (Sender: TObject; var aRect: TRSRect) of object;

{ Rect and Point Stuff }
{$IFDEF FMX}
  TRSPoints = TCanvasPoints;
  TPoints = array of TPoint;
  TPointPolygons = Array of TPoints;
  TRSPolygon = TRSPoints;
  TRSPolygons = Array of TRSPoints;
  TCanvasPolygons = TRSPolygons;
  TRSCubicBezier = TCubicBezier;
  TRSCorner = TCorner;
  TRSCorners = TCorners;
  TRSCornerType = TCornerType;
  {$IFDEF DELPHI20PLUS}
  TCornerTypeHelper = record helper for TCornerType
  const
    ctRound = TCornerType.Round;
    ctBevel = TCornerType.Bevel;
    ctInnerRound = TCornerType.InnerRound;
    ctInnerLine = TCornerType.InnerLine;
  end;
  {$ENDIF}
  TRSSide = TSide;
  TRSSides = TSides;
  TRSPathPointKind = TPathPointKind;
  {$IFDEF DELPHI20PLUS}
  TPathPointKindHelper = record helper for TPathPointKind
  const
    ppMoveTo = TPathPointKind.MoveTo;
    ppLineTo = TPathPointKind.LineTo;
    ppCurveTo = TPathPointKind.CurveTo;
    ppClose = TPathPointKind.Close;
  end;
  {$ENDIF}
  TRSPathPoint = TPathPoint;
{$ELSE}
  TRSPoints = Array of TRSPoint;
  TPoints = TCanvasPoints;
  TPointPolygons = Array of TPoints;
  TCanvasPolygons = TPointPolygons;
  TRSPolygon = TRSPoints;
  TRSPolygons = Array of TRSPoints;
  TRSCubicBezier = array [0..3] of TRSPoint;
  TRSCorner = (crTopLeft, crTopRight, crBottomLeft, crBottomRight);
  TRSPathPointKind = (ppMoveTo, ppLineTo, ppCurveTo, ppClose);
  TRSCornerType = (ctRound, ctBevel, ctInnerRound, ctInnerLine);
  TRSSide = (sdTop, sdLeft, sdBottom, sdRight);
  TRSCorners = set of TRSCorner;
  TRSSides = set of TRSSide;
  TRSPathPoint = packed record
    Kind: TRSPathPointKind;
    Point: TRSPoint;
  end;
{$ENDIF}

  TRSCustomRect = class(TPersistent)
  { Purpose: To define a helper class to contain and perform calculations on   }
  {          a TRect                                                           }
  private
    { Private declarations }
    FBoundsRect: TRect;
    FOnChange: TNotifyEvent;
  protected
    { Protected declarations }
    procedure Change; dynamic;
    function GetSize(const Index: Integer): Integer; virtual;
    procedure SetSize(const Index, Value: Integer); virtual;
    procedure SetBoundsRect(const Value: TRect); virtual;
  public
    { Public declarations }
    procedure Assign(Source: TPersistent); override;
    procedure MoveTo( X, Y: Integer ); virtual;
    procedure Translate( DX, DY: Integer ); virtual;
    procedure Scale( DX, DY: Integer ); virtual;
    property BoundsRect: TRect read FBoundsRect write SetBoundsRect;
    property OnChange: TNotifyEvent read FOnChange write FOnChange;
    property Left: Integer index 0 read GetSize write SetSize default 0;
    property Top: Integer index 1 read GetSize write SetSize default 0;
    property Right: Integer index 2 read GetSize write SetSize default 0;
    property Bottom: Integer index 3 read GetSize write SetSize default 0;
    property Height: Integer index 4 read GetSize write SetSize default 0;
    property Width: Integer index 5 read GetSize write SetSize default 0;
  published
    { Published declarations }
  end; { TRSCustomRect }

  TRSObjectRect = class(TRSCustomRect)
  private
    { Private declarations }
  protected
    { Protected declarations }
  public
    { Public declarations }
  published
    { Published declarations }
    property Left;
    property Top;
    property Right;
    property Bottom;
  end; { TRSObjectRect }

  TObjectSize = class(TRSCustomRect)
  private
    { Private declarations }
  protected
    { Protected declarations }
  public
    { Public declarations }
  published
    { Published declarations }
    property Left;
    property Top;
    property Right;
    property Bottom;
    property Height;
    property Width;
  end;

  TRSCustomRSRect = class(TPersistent)
  { Purpose: To define a helper class to contain and perform calculations on   }
  {          a TRSRect                                                           }
  private
    { Private declarations }
    FBoundsRect: TRSRect;
    FOnChange: TNotifyEvent;
  protected
    { Protected declarations }
    procedure Change; dynamic;
    function GetSize(const Index: Integer): Single; virtual;
    procedure SetSize(const Index: Integer; const Value: Single); virtual;
    procedure SetBoundsRect(const Value: TRSRect); virtual;
  public
    { Public declarations }
    procedure Assign(Source: TPersistent); override;
    procedure MoveTo( X, Y: Single ); virtual;
    procedure Translate( DX, DY: Single ); virtual;
    procedure Scale( DX, DY: Single ); virtual;
    property BoundsRect: TRSRect read FBoundsRect write SetBoundsRect;
    property OnChange: TNotifyEvent read FOnChange write FOnChange;
    property Left: Single index 0 read GetSize write SetSize;
    property Top: Single index 1 read GetSize write SetSize;
    property Right: Single index 2 read GetSize write SetSize;
    property Bottom: Single index 3 read GetSize write SetSize;
    property Height: Single index 4 read GetSize write SetSize;
    property Width: Single index 5 read GetSize write SetSize;
  published
    { Published declarations }
  end; { TRSCustomRSRect }

  TRSObjectRSRect = class(TRSCustomRSRect)
  private
    { Private declarations }
  protected
    { Protected declarations }
  public
    { Public declarations }
  published
    { Published declarations }
    property Left;
    property Top;
    property Right;
    property Bottom;
  end; { TRSObjectRect }

{$IFNDEF FMX}
  TVclCanvasHelper = class helper for TCanvas
  { Purpose: To provide FMX-specific method signatures to Vcl TCanvas
  }
  private
    function GetPenWidth: Integer;
    procedure SetPenWidth(const Value: Integer);
  public
    procedure DrawLine( P1, P2: TRSPoint ); overload;
    procedure DrawLine( P1, P2: TPoint ); overload;
    procedure Polygon(const Points: TRSPoints); overload;
    procedure Polyline(const Points: TRSPoints); overload;
    property PenWidth: Integer read GetPenWidth write SetPenWidth;
  end;
{$ENDIF}

  TRSVectorArray = array [0..2] of Single;
{$IFDEF DELPHI16PLUS}
  TRSVector = TVector;
{$ELSE}
  TRSVector = record
  private
    // sum of squares coords
    function Norm: Single;

    // 1 div square root, when square root is not null
    function RSqrt(const AValue: Single): Single;

    function Add(const AVector: TRSVector): TRSVector;
    function Subtract(const AVector: TRSVector): TRSVector;
  public
    constructor Create(const AX, AY, AW : Single); overload;
    constructor Create(const APoint : TRSPoint); overload;

    function Normalize: TRSVector;

    // proportional change vector's length
    function Scale(const AFactor: Single): TRSVector;

    // calc vector's length
    function Length: Single;

    // multiplying the vector's coordinates
    function DotProduct(const AVector: TRSVector): Single;

    function Reflect(const AVector: TRSVector): TRSVector;

    // get TRSPoint value
    function ToPointF: TRSPoint;
    function ToRSPoint: TRSPoint;
    class operator Add(const AFirstVector, ASecondVector : TRSVector) : TRSVector;
    class operator Subtract(const AFirstVector, ASecondVector : TRSVector) : TRSVector;
    class operator Equal(const AFirstVector, ASecondVector : TRSVector) : Boolean;
    class operator NotEqual(const AFirstVector, ASecondVector : TRSVector): Boolean;
    class operator Implicit(AVector : TRSVector) : TRSPoint;
    class operator Explicit(AVector : TRSVector) : TRSPoint;
    class operator Implicit(APointF : TRSPoint) : TRSVector;
    class operator Explicit(APointF : TRSPoint) : TRSVector;
    class operator Implicit(ASizeF : TRSSize) : TRSVector;
    class operator Explicit(ASizeF : TRSSize) : TRSVector;
    case Integer of
      0: (V: TRSVectorArray;);
      1: (X: Single;
          Y: Single;
          W: Single;);
  end;
{$ENDIF}

  TRSMatrixArray = array [0..2] of TRSVector;

{$IFDEF DELPHI16PLUS}
  TRSMatrix = TMatrix;
{$ELSE}
  TRSMatrix = record
  private
  public
    function Transform(const AVector: TRSVector): TRSVector;
    case Integer of
      0: (M: TRSMatrixArray;);
      1: (m11, m12, m13: Single;
          m21, m22, m23: Single;
          m31, m32, m33: Single);
  end;
{$ENDIF}

{$IF Defined(DELPHI16PLUS) and Defined(FMX)}
  TRSPosition = TPosition;
{$ELSE}
  TRSPosition = class(TPersistent)
  private
    FOnChange: TNotifyEvent;
    FY: Single;
    FX: Single;
    FDefaultValue: TRSPoint;
    procedure SetPoint(const Value: TRSPoint);
    procedure SetX(const Value: Single);
    procedure SetY(const Value: Single);
    function GetPoint: TRSPoint;
    function GetVector: TRSVector;
    procedure SetVector(const Value: TRSVector);
    function IsXStored: Boolean;
    function IsYStored: Boolean;
  protected
    procedure DefineProperties(Filer: TFiler); override;
    procedure ReadPoint(Reader: TReader);
    procedure WritePoint(Writer: TWriter);
    procedure Change; dynamic;
  public
    constructor Create(const ADefaultValue: TRSPoint); virtual;
    procedure Assign(Source: TPersistent); override;
    procedure SetPointNoChange(const P: TRSPoint);
    function Empty: Boolean;
    procedure Reflect(const Normal: TRSVector);
    property Point: TRSPoint read GetPoint write SetPoint;
    property Vector: TRSVector read GetVector write SetVector;
    property DefaultValue: TRSPoint read FDefaultValue write FDefaultValue;
    property OnChange: TNotifyEvent read FOnChange write FOnChange;
  published
    property X: Single read FX write SetX stored IsXStored nodefault;
    property Y: Single read FY write SetY stored IsYStored nodefault;
  end;
{$IFEND}

{$IF Defined(DELPHI16PLUS) and Defined(FMX)}
  TRSTransform = TTransform;
{$ELSE}
  TRSTransform = class(TPersistent)
  private
    FMatrix: TRSMatrix;
    FRotationAngle: Single;
    FPosition: TRSPosition;
    FScale: TRSPosition;
    FSkew: TRSPosition;
    FRotationCenter: TRSPosition;
    FOnChanged: TNotifyEvent;
    FOnGetShapeRect: TRSGetRectEvent;
    procedure SetRotationAngle(const Value: Single);
    procedure SetScale(const Value: TRSPosition);
    procedure SetSkew(const Value: TRSPosition);
    procedure SetPosition(const Value: TRSPosition);
  protected
    procedure MatrixChanged(Sender: TObject);
    procedure Change; dynamic;
  public
    constructor Create; virtual;
    destructor Destroy; override;
    procedure Assign(Source: TPersistent); override;
    property Matrix: TRSMatrix read FMatrix;
    property OnChanged: TNotifyEvent read FOnChanged write FOnChanged;
    property OnGetShapeRect: TRSGetRectEvent read FOnGetShapeRect write FOnGetShapeRect;
  published
    property Position: TRSPosition read FPosition write SetPosition;
    property Scale: TRSPosition read FScale write SetScale;
    property Skew: TRSPosition read FSkew write SetSkew;
    property RotationAngle: Single read FRotationAngle write SetRotationAngle;
    property RotationCenter: TRSPosition read FRotationCenter write FRotationCenter;
  end;
{$IFEND}

  TRSFillMode = (fmEvenOdd, fmNonZero);
  TRSPathData = class({$IFDEF FMX}TPathData{$ELSE}TPersistent{$ENDIF})
  { Purpose: To reproduce TPathData for Vcl and FMX
    Also refactor the class to provide smarter Change events }
  public const
    LongFormatStr = '0.######################################';
    ShortFormatStr = '0.###';
  public type
    TDistances = Array of TSingleDynArray;
  private
    { Private declarations }
    FOnChanged: TNotifyEvent;
    FUpdateCount: Integer;
    FFillMode: TRSFillMode;
    FLength: Double;
    FPolygons: TRSPolygons;
    FDistances: TDistances;
    FFormatStr: String;
{$IFNDEF FMX}
    FStartPoint: TRSPoint;
    FPathData: array of TRSPathPoint;
    FRecalcBounds: Boolean;
    FBounds: TRSRect;
{$ENDIF}
    function GetDistances: TDistances;
    procedure AddArcSvgPart(const Center, Radius: TRSPoint; StartAngle, SweepAngle: Single);
    function GetPathString: {$IF Defined(FMX) and not Defined(DELPHI17PLUS)}AnsiString{$ELSE}String{$IFEND};
    procedure SetPathString(const Value: {$IF Defined(FMX) and not Defined(DELPHI17PLUS)}AnsiString{$ELSE}String{$IFEND});
    function GetTok(const S: string; var Pos: Integer): string;
    function GetPointFromStr(const S: string; var Pos: Integer): TRSPoint;
    function GetDigit(const S: string; var Pos: Integer): string;
    function GetNum(const S: string; var Pos: Integer): string;
    function IsFormatStrStored: Boolean;
    procedure SetFillMode(const Value: TRSFillMode);
    procedure SetFormatStr(const Value: String);
{$IFNDEF FMX}
    function GetCount: Integer; inline;
    function GetPoint(AIndex: Integer): TRSPathPoint; inline;
  protected
    { Protected declarations }
    procedure Change; dynamic;
    procedure DefineProperties(Filer: TFiler); override;
    procedure ReadPath(Stream: TStream);
    procedure WritePath(Stream: TStream);
    procedure DrawPath( const Canvas: TCanvas; Filled: Boolean; Opacity: Single = 1 ); overload; virtual;
{$ELSE !FMX}
  protected
    { Protected declarations }
    procedure AssignTo(Dest: TPersistent); override;
    procedure PathChange(Sender: TObject);
{$ENDIF}
    procedure AddRectangleForSegment( p1, p2: TRSPoint; OffSet: Single );
    procedure AddArcSvg(const P1, Radius: TRSPoint; Angle: Single; const LargeFlag, SweepFlag: Boolean; const P2: TRSPoint);
    function CalculateDistances: Double;
    property UpdateCount: Integer read FUpdateCount write FUpdateCount;
  public
    { Public declarations }
    function Clone: TRSPathData;
    constructor Create; {$IFDEF FMX}override;{$ELSE}virtual;{$ENDIF}
    procedure AddPathOutline( const P: TRSPathData; StrokeWidth: Single );
    procedure AddPolygon( const Poly: TPoints ); overload;
    procedure AddPolygon( const Poly: TRSPoints ); overload;
    procedure AddPolygons( const Poly: TPointPolygons ); overload;
    procedure AddPolygons( const Poly: TRSPolygons ); overload;
    procedure Append(Source: TRSPathData); overload; virtual;
    procedure Append(const PathString: String); overload;
    procedure BeginUpdate;
    procedure EndUpdate;
    function ToPolygons: TRSPolygons;
    property OnChanged: TNotifyEvent read FOnChanged write FOnChanged;
    procedure DrawPath( const Canvas: TCanvas; Opacity: Single = 1 ); {$IFDEF FMX}inline;{$ELSE}overload;{$ENDIF}
    procedure FillPath( const Canvas: TCanvas; Opacity: Single = 1 ); {$IFDEF FMX}inline;{$ELSE}overload;{$ENDIF}
    function PointInPath( const P: TRSPoint ): Boolean; overload;
    function PointInPath( const X, Y: Single ): Boolean; overload;
    function SplitDataByLength( const Len: Integer = 80; const Delimiter: String = NewLineString ): String;
    function PointInPathOutline( const P: TRSPoint; StrokeWidth: Single ): Boolean; overload;
    function PointInPathOutline( const X, Y: Single; StrokeWidth: Single ): Boolean; overload;
    procedure SmoothCurveTo(const ControlPoint2, EndPoint: TRSPoint);
    procedure SmoothCurveToRel(const ControlPoint2, EndPoint: TRSPoint);
    function GetLength: Double;
    property Distances: TDistances read GetDistances;
    property Polygons: TRSPolygons read ToPolygons;
{$IFNDEF FMX}
    procedure Assign(Source: TPersistent); override;
    { creation }
    function LastPoint: TRSPoint;
    procedure MoveTo(const P: TRSPoint);
    procedure MoveToRel(const P: TRSPoint);
    procedure LineTo(const P: TRSPoint);
    procedure LineToRel(const P: TRSPoint);
    procedure HLineTo(const X: Single);
    procedure HLineToRel(const X: Single);
    procedure VLineTo(const Y: Single);
    procedure VLineToRel(const Y: Single);
    procedure CurveTo(const ControlPoint1, ControlPoint2, EndPoint: TRSPoint);
    procedure CurveToRel(const ControlPoint1, ControlPoint2, EndPoint: TRSPoint);
    procedure QuadCurveTo(const ControlPoint, EndPoint: TRSPoint);
    procedure ClosePath;
    { shapes }
    procedure AddEllipse(const ARect: TRSRect);
    procedure AddRectangle(const ARect: TRSRect; const XRadius, YRadius: Single; const ACorners: TRSCorners;
      const ACornerType: TRSCornerType = TRSCornerType.ctRound);
    procedure AddArc(const Center, Radius: TRSPoint; StartAngle, SweepAngle: Single);
    { modification }
    procedure Clear;
    procedure Flatten(const Flatness: Single = 0.25);
    procedure Scale(const scaleX, scaleY: Single);
    procedure Translate(const dX, dY: Single);
    procedure FitToRect(const ARect: TRSRect);
    procedure ApplyMatrix(const M: TRSMatrix);
    { params }
    function GetBounds: TRSRect;
    { convert }
    function FlattenToPolygon(var Polygon: TRSPolygon; const Flatness: Single = 0.25): TRSPoint;
    function IsEmpty: Boolean;
    { access }
    property Count: Integer read GetCount;
    property Points[AIndex: Integer]: TRSPathPoint read GetPoint; default;
  published
    { Published declarations }
{$ENDIF}
    property FillMode: TRSFillMode read FFillMode write SetFillMode; // not used by FMX
    property FormatStr: String read FFormatStr write SetFormatStr stored IsFormatStrStored;
    property Data: {$IF Defined(FMX) and not Defined(DELPHI17PLUS)}AnsiString{$ELSE}String{$IFEND} read GetPathString write SetPathString stored False;
  end;
  TRSPathDataClass = class of TRSPathData;

{ Vector and Matrix routines }
function RSPoint( const aVector: TRSVector ): TRSPoint; overload;
function RSVector( const X, Y: Single; const W: Single = 1.0 ): TRSVector; overload;
function RSVector( const aPoint: TRSPoint; const W: Single = 1.0 ): TRSVector; overload;
{$IFDEF DELPHI16PLUS}
function RSVector( const aVector: TVector ): TRSVector; overload;
function RSMatrix( const aMatrix: TMatrix ): TRSMatrix;
function Vector( const aVector: TRSVector ): TVector;
function Matrix( const aMatrix: TRSMatrix ): TMatrix;
{$ENDIF}
function RSVectorAdd(const AFirstVector, ASecondVector : TRSVector) : TRSVector; inline;
function RSVectorSubtract(const AFirstVector, ASecondVector : TRSVector) : TRSVector; inline;
function RSVectorAngleCosine(const v1, v2: TRSVector): Single;
function RSVectorDotProduct(const v1, v2: TRSVector): Single; inline;
function RSVectorCrossProductZ(const v1, v2: TRSVector): Single; inline;
function RSVectorAngle(const V, N: TRSVector): Single;
function RSVectorLength( const V: TRSVector ): Single; inline;
function RSVectorNorm(const V: TRSVector): Single; inline;
function RSVectorMultiply(const AVector: TRSVector; const AFactor: Single): TRSVector; overload; inline;
function RSVectorMultiply(const AFactor: Single; const AVector: TRSVector): TRSVector; overload; inline;
function RSVectorDivide(const AVector: TRSVector; const AFactor: Single): TRSVector;
function RSVectorReflect(const V1, V2: TRSVector): TRSVector; inline;
function RSVectorNormalize(const V: TRSVector): TRSVector;

function RSMatrixEquals( const M1, M2: TRSMatrix; const Epsilon: Single = 1E-5 ): Boolean;
function RSMatrixTransform(const AMatrix: TRSMatrix; const aRect: TRSRect): TRSRect; overload;
function RSMatrixTransform(const AMatrix: TRSMatrix; const AVector: TRSVector): TRSVector; overload;
function RSMatrixMultiply(const M1, M2: TRSMatrix): TRSMatrix; overload;
function RSMatrixMultiply(const APoint: TRSPoint; const AMatrix: TRSMatrix): TRSPoint; overload;
function RSMatrixMultiply(const AVector: TRSVector; const AMatrix: TRSMatrix): TRSVector; overload;
function RSMatrixDeterminant(const M: TRSMatrix): Single;
procedure AdjointRSMatrix(var M: TRSMatrix);
procedure ScaleRSMatrix(var M: TRSMatrix; const factor: Single);
procedure InvertRSMatrix(var M: TRSMatrix);
function CreateRotationRSMatrix(const Angle: Single): TRSMatrix; overload;
function CreateRotationRSMatrix(const Point: TRSPoint; const Angle: Single): TRSMatrix; overload;
function CreateScaleRSMatrix(const ScaleX, ScaleY: Single): TRSMatrix;
function CreateTranslateRSMatrix(const DX, DY: Single): TRSMatrix;
function RSMatrixEulerAngles( const M: TRSMatrix ): TRSVector;
function RSMatrixRotationAngle( const M: TRSMatrix ): Single;
function RSMatrixTranslationVector( const M: TRSMatrix ): TRSPoint;
function RSMatrixScaleVector( const M: TRSMatrix ): TRSPoint;
function IsIdentityMatrix( const M: TRSMatrix ): Boolean;

function CenterRect( const Pt: TPoint; Width, Height: Integer ): TRect; overload;
function CenterRect( const X, Y: Integer; Width, Height: Integer ): TRect; overload;
function CenterRect( const Pt: TRSPoint; Width, Height: TRSGfxFloat ): TRSRect; overload;
function CenterRect( const X, Y: TRSGfxFloat; Width, Height: TRSGfxFloat ): TRSRect; overload;
function CenterRect( var R: TRect; const Bounds: TRect ): TRect; overload;
function CenterRect( var R: TRSRect; const Bounds: TRSRect ): TRSRect; overload;
function OffsetRSRect(var R: TRSRect; DX, DY: Single): Boolean; overload;
function RSIntersectLine( const V1, V2, V3, V4: TRSPoint; var IntersectionPt: TRSPoint ): Boolean;
function RSIntersectRect( const R1, R2: TRect ): Boolean; overload;
function RSIntersectRect( const R1, R2: TRSRect ): Boolean; overload;
function RSIntersectRect( const R1, R2: TRect; var R: TRect ): Boolean; overload;
function RSIntersectRect( const R1, R2: TRSRect; var R: TRSRect ): Boolean; overload;
function RSIntersectRect( const V1, V2: TRSPoint; const R: TRSRect; var Pts: TRSPoints ): Boolean; overload;
function PtInRSRect( const Pt: TPoint; const ARect: TRSRect ): Boolean; overload;
function PtInRSRect( const Pt: TRSPoint; const ARect: TRSRect ): Boolean; overload;
function GetPointOnRect( const ARect: TRect; ConnectionPoint: TConnectionPoint): TPoint;
function GetRSPointOnRSRect( const ARect: TRSRect; ConnectionPoint: TConnectionPoint): TRSPoint;
function RSCenterPoint( const ARect: TRSRect ): TRSPoint; overload;
function RSCenterPoint( const ARect: TRect ): TRSPoint; overload;
function PointToRSPoint( const Pt: TPoint ): TRSPoint;
function PointsToRSPoints( const Pt: TPoints ): TRSPoints;
function RSPointToString(R: TRSPoint; FormatStr: String = '%f'): string;
function StringToRSPoint(S: string): TRSPoint;
function RSPointToPoint( const Pt: TRSPoint ): TPoint;
function RSPointToCanvasPoint( const Pt: TRSPoint ): TCanvasPoint;
function RSPointsToPoints( const Pt: TRSPoints ): TPoints;
function RSPoint( const X, Y: TRSGfxFloat ): TRSPoint; overload; inline;
function RSPoint( const Pt: TPoint ): TRSPoint; overload; inline;
function PointsEqual( const Pt1, Pt2: TRSPoint ): Boolean; overload;
function EqualsRect( const R1, R2: TRSRect ): Boolean; overload;
function EqualsRect( const R1, R2: TRect ): Boolean; overload;
function RSRectHeight( const ARect: TRect ): Integer; inline; overload;
function RSRectHeight( const ARect: TRSRect ): TRSGfxFloat; inline; overload;
function RSRectWidth( const ARect: TRSRect ): TRSGfxFloat; inline; overload;
function RSRectWidth( const ARect: TRect ): Integer; inline; overload;
procedure RSInflateRect( var R: TRect; const DX, DY: Integer ); overload;
procedure RSInflateRect( var R: TRSRect; const DX, DY: Single ); overload;
/// Scale the Points from weighted center, Scale=1 equals no scaling
function ScalePoints( const Pt: TRSPoints; ScaleDX, ScaleDY: Single ): TRSPoints; overload;
/// Scale the Points from weighted center, Scale=1 equals no scaling
function ScalePoints( const Pt: TPoints; ScaleDX, ScaleDY: Single ): TPoints; overload;
/// Scale the Rect by the number of pixels
function ScaleRect( ARect: TRect; DX, DY: Single ): TRect; overload;
/// Scale the Rect by the number of pixels
function ScaleRect( ARect: TRSRect; DX, DY: Single ): TRSRect; overload;
function RSRect( const ALeft, ATop, ARight, ABottom: TRSGfxFloat ): TRSRect; overload;
function RSRect( const TopLeft, BottomRight: TRSPoint ): TRSRect; overload;
function RSRect( const ARect: TRect ): TRSRect; overload; inline;
function RSRect( const ARect: TRSRect ): TRSRect; overload; inline;
function RectToRSRect( const ARect: TRect ): TRSRect;
function RSRectToRect( const ARect: TRSRect ): TRect;
function FixRect( const ARect: TRect ): TRect; overload;
function FixRect( const ARect: TRSRect ): TRSRect; overload;
function FitRect( aMatrix: TRSMatrix; const aRect, TargetRect: TRSRect ): TRSRect; overload;
function FitRect( var aRect: TRect; const boundsRect: TRect; Grow: Boolean = False ): Single; overload;
function FitRect( var aRect: TRSRect; const boundsRect: TRSRect; Grow: Boolean = False ): Single; overload;
function InCenterPoint( const A, B, C: TPoint ): TPoint; overload;
function InCenterPoint( const A, B, C: TRSPoint ): TRSPoint; overload;
function OffsetPoint( var Pt: TPoint; const X, Y: Integer ): TPoint; overload;
function OffsetPoint( var Pt: TPoint; LineEndPt: TPoint; Dist: TRSGfxFloat ): TPoint; overload;
function OffsetPoint( var Pt: TRSPoint; const X, Y: Single ): TRSPoint; overload;
function OffsetPoint( var Pt: TRSPoint; LineEndPt: TRSPoint; Dist: TRSGfxFloat ): TRSPoint; overload;
function PointInPath(const Points: TPoints; X, Y: Integer): Boolean; overload;
function PointInPath(const Points: TRSPoints; X, Y: Single): Boolean; overload;
function PointInPath(const Pt: TRSPoint; const Points: TRSPoints): TPathClickLocation; overload;
function PointInPath(const Pt: TPoint; const Points: TPoints): TPathClickLocation; overload;
function RSCenterPoint( const Pt: TRSPoints ): TRSPoint; overload;
function RSCenterPoint( const Pt: TPoints ): TRSPoint; overload;
function RSDistance( const Pt1, Pt2: TRSPoint ): Single; inline;
function RSCenterPoint( const Pt1, Pt2: TRSPoint; Midpoint: Single = 0.5 ): TRSPoint; overload;

{$IFDEF DELPHI5ORLESS}
function CenterPoint( const ARect: TRect ): TPoint;
function PointsEqual( const Pt1, Pt2: TPoint ): Boolean; overload;
{$ENDIF}

{ Colors }
function DoubleToByte(D: Double): Byte; { For color precision in gradients }

{ Gradients }
{$IFDEF FMX}
function SetGradientColors( const Gradient: TGradient;
  const Colors: array of TCanvasColor ): TGradient;
procedure DrawGradient( const Canvas: TCanvas; const ARect: TCanvasRect;
  const Colors: array of TCanvasColor; Style: TGradientStyle;
  StartPt, StopPt: TPointF; FillAsRect: Boolean = True; const Opacity: Single = 1 );
{$ENDIF}
procedure DirectionGradient( const Canvas: TCanvas; const CanvasRect: TCanvasRect;
   const Colors: array of TCanvasColor; Direction: TGradientDirection = gdDownRight;
   LineSize: Integer = 1; const Opacity: Single = 1);
procedure LineGradient(const Canvas: TCanvas; const ARect: TCanvasRect;
  const Colors: array of TCanvasColor; LineSize: Integer;
  const Orientation: TGradientOrientation; const Opacity: Single = 1);
procedure EllipseGradient(const Canvas: TCanvas; const ARect: TCanvasRect;
  FocalX, FocalY: Single;
  const Colors: array of TCanvasColor; LineSize: Single;
  const Opacity: Single = 1); overload;
procedure EllipseGradient(const Canvas: TCanvas; const ARect: TCanvasRect;
  const Colors: array of TCanvasColor; LineSize: Integer;
  const Opacity: Single = 1); overload;
procedure RectGradient(const Canvas: TCanvas; const ARect: TCanvasRect;
  const Colors: array of TCanvasColor; LineSize: Integer;
  const Opacity: Single = 1);
function IndexToColor(const Index: Integer; const Count: Integer; const Colors: TColors): TCanvasColor;
procedure ProgressCircle(const Canvas: TCanvas; const CanvasR: TCanvasRect;
   const OuterPadding: Integer; const Progress: Integer;
   const Min, Max: Integer; const Colors: TColors; const AdjustPenColor: Boolean;
   const SubDivisions: Integer; const SubShape: TRSShapeType;
   const SubShapeOrientation: TRSShapeOrientation;
   const SubShapeBevel: Boolean; const SubShapeBevelSize: Integer;
   const Spacing: Integer;
   const Direction: TProgressDirection; const DisplayType: TProgressDisplayType);
procedure ProgressPie(const Canvas: TCanvas; const CanvasR: TCanvasRect;
   const OuterPadding: Integer; const Progress: Integer;
   const Min, Max: Integer; const Colors: TColors; const AdjustPenColor: Boolean;
   const SubDivisions: Integer; const Spacing: Integer;
   const Direction: TProgressDirection; const DisplayType: TProgressDisplayType);
procedure ProgressRect(const Canvas: TCanvas;  const CanvasR: TCanvasRect;
   const OuterPadding: Integer; const Progress: Integer;
   const Min, Max: Integer; const Colors: TColors; const AdjustPenColor: Boolean;
   const SubDivisions: Integer; const SubShape: TRSShapeType;
   const SubShapeOrientation: TRSShapeOrientation;
   const SubShapeBevel: Boolean; const SubShapeBevelSize: Integer;
   const Spacing: Integer;
   const Orientation: TProgressOrientation;
   const Direction: TProgressDirection; const DisplayType: TProgressDisplayType);


{ Text }
function GetMultilineTextHeight(const Canvas: TCanvas; Text: string): TRSGfxFloat;
// get size and origin of rotated string
function GetRotatedTextExtent( const Canvas: TCanvas; Angle: Integer; const Str: String; var Origin: TPoint ): TSize; overload;
function GetRotatedTextExtent( const Canvas: TCanvas; Angle: Integer; const Str: String; var Origin: TRSPoint ): TRSSize; overload;
function WrapCanvasText(const Canvas: TCanvas; Text: string; MaxWidth: TRSGfxFloat): String; overload;
{$IFDEF NEXTGEN}
function WrapCanvasText( const Canvas: TCanvas; Text: string;
                         const BreakStr: string; BreakChars: Array of Char;
                         MaxWidth: TRSGfxFloat ): String; overload;
{$ELSE}
function WrapCanvasText( const Canvas: TCanvas; Text: string;
                         const BreakStr: string; BreakChars: TSysCharSet;
                         MaxWidth: TRSGfxFloat ): String; overload;
{$ENDIF NEXTGEN}

{ Shapes }
function Arc( const Path: TRSPathData; StartAngle, EndAngle,
  InnerRadius, OuterRadius: Double; ArcOffset: Double = 0 ): TRSPathData; overload;
function Arc( StartAngle, EndAngle, InnerRadius, OuterRadius: Double; ArcOffset: Double = 0 ): String; overload;
procedure DrawBezierPolygon( const Canvas: TCanvas; Vc: TPoints;
                         Smooth: Boolean = False;
                         Closed: Boolean = False;
                         Filled: Boolean = False;
                         PointStep: Integer = 1;
                         Opacity: Single = 1 ); overload;
procedure DrawBezierPolygon( const Canvas: TCanvas; Vc: TRSPoints;
                         Smooth: Boolean = False;
                         Closed: Boolean = False;
                         Filled: Boolean = False;
                         PointStep: Integer = 1;
                         Opacity: Single = 1 ); overload;
function BezierPolygonPoints( Vc: TPoints;
                         Smooth: Boolean = False;
                         Closed: Boolean = False; PointStep: Integer = 1 ): TPoints; overload;
function BezierPolygonPoints( Vc: TRSPoints;
                         Smooth: Boolean = False;
                         Closed: Boolean = False; PointStep: Integer = 1 ): TRSPoints; overload;
procedure DrawArrow(const Canvas:  TCanvas;
                    const x1, y1, x2, y2: TCanvasPixel;
                    const HeadLength: TCanvasPixel;
                    Caption: String = '';
                    Options: TArrowOptions = [aoArrowAtEnd];
                    Width: TCanvasPixel = 1;
                    Opacity: Single = 1
                    {$IFDEF FMX}; FontColor: TCanvasColor = clxBlack{$ENDIF}); overload;
procedure DrawArrow(const Canvas:  TCanvas;
                    const Points: TPoints;
                    const HeadLength: TCanvasPixel;
                    Caption: String = '';
                    Options: TArrowOptions = [aoArrowAtEnd];
                    Width: TCanvasPixel = 1;
                    CaptionSpan: TCanvasPixel = -1;
                    Opacity: Single = 1
                    {$IFDEF FMX}; FontColor: TCanvasColor = clxBlack{$ENDIF}); overload;
procedure DrawArrow(const Canvas:  TCanvas;
                    const Points: TRSPoints;
                    const HeadLength: TCanvasPixel;
                    Caption: String = '';
                    Options: TArrowOptions = [aoArrowAtEnd];
                    Width: TCanvasPixel = 1;
                    CaptionSpan: TCanvasPixel = -1;
                    Opacity: Single = 1
                    {$IFDEF FMX}; FontColor: TCanvasColor = clxBlack{$ENDIF}); overload;
procedure DrawDonutSlice( const Canvas: TCanvas; ARect: TCanvasRect;
                        StartDegrees, StopDegrees: Double; Offset: Integer = 0;
                        InnerRadius: Double = 0; OuterRadius: Double = 1;
                        Opacity: Single = 1 );
procedure DrawPieSlice( const Canvas: TCanvas; ARect: TCanvasRect;
                        StartDegrees, StopDegrees: Double;
                        OuterRadius: Double;
                        Offset: Integer = 0;
                        Opacity: Single = 1 ); overload;
procedure DrawPieSlice( const Canvas: TCanvas; ARect: TCanvasRect;
                        StartDegrees, StopDegrees: Double; Offset: Integer = 0;
                        Opacity: Single = 1 ); overload;
function DrawShape(const Canvas: TCanvas; const CanvasRect: TCanvasRect;
   const Shape: TRSShapeType; const ShapeOrientation: TRSShapeOrientation;
   const Bevel: Boolean; const BevelSize: TCanvasPixel): TCanvasRect; overload;
function DrawShape( const Canvas: TCanvas; ARect: TCanvasRect;
                     Image: TBitmap;
                     Style: TRSShapeStyle;
                     StretchImage: Boolean = True;
                     BevelHighlight: TCanvasColor = clxBtnHighlight;
                     BevelShadow: TCanvasColor = clxBtnShadow;
                     BevelSize: TCanvasPixel = 3;
                     Opacity: Single = 1 ): TCanvasRect; overload;
procedure Frame3DEllipse(Canvas: TCanvas; var Rect: TCanvasRect;
  TopColor, BottomColor: TCanvasColor; Width: TCanvasPixel);
procedure Frame3DExt(Canvas: TCanvas; var Rect: TCanvasRect;
  TopColor, BottomColor: TCanvasColor; Width: TCanvasPixel;
  Rounded: Boolean = False; X: TCanvasPixel = 0; Y: TCanvasPixel = 0 );

{ Canvas functions }
{ These functions make it easier to switch between Vcl and FMX }
function CanvasRect( aRect: TRSRect ): TCanvasRect; inline; overload;
function CanvasRect( aRect: TRect ): TCanvasRect; inline; overload;
function CanvasRect( Left, Top, Right, Bottom: Integer ): TCanvasRect; inline; overload;
function CanvasRect( Left, Top, Right, Bottom: Single ): TCanvasRect; inline; overload;
function CanvasPoint( X, Y: Integer ): TCanvasPoint; inline; overload;
function CanvasPoint( X, Y: Single ): TCanvasPoint; inline; overload;
function CanvasRound( X: Double ): TExtCanvasPixel; inline;
function CanvasTrunc( X: Double ): TExtCanvasPixel; inline;
function CanvasDiv( X, Y: Int64 ): TExtCanvasPixel; inline; overload;
function CanvasDiv( X, Y: Double ): TExtCanvasPixel; inline; overload;
function CanvasMax( const Data: array of TCanvasPixel ): TCanvasPixel;
function CanvasMin( const Data: array of TCanvasPixel ): TCanvasPixel;
function CanvasCenterPoint( const Rect: TCanvasRect ): TCanvasPoint; inline;
procedure CopyCanvasState( const CanvasTo, CanvasFrom: TCanvas );
function GetCanvasPointOnRect( const ARect: TCanvasRect; ConnectionPoint: TConnectionPoint): TCanvasPoint;
function CanvasColorToString(Color: TCanvasColor): string;
function StringToCanvasColor(const S: string): TCanvasColor;
function CanvasColorToIdent(Color: Longint; var Ident: string): Boolean;
function IdentToCanvasColor(const Ident: string; var Color: Integer): Boolean;

const
  RSClosePolygon: TRSPoint = (X: $FFFF; Y: $FFFF);
  DefaultColorList: TColorList = (clxRed, clxGreen, clxYellow, clxBlue, clxPurple,
      clxMaroon, clxLime, clxTeal, clxFuchsia, clxAqua);
  PastelsColorList: TColorList = ( clxLightBlue, clxLightCoral,
      clxLightGoldenrodyellow, clxLightGreen, clxLightPink, clxLightSalmon, clxLightCyan,
      clxLightSkyBlue, clxLightSteelBlue, clxLightYellow);
  BWColorList: TColorList = (clxBlack, clxWhite, clxBlack, clxWhite, clxBlack,
      clxWhite, clxBlack, clxWhite, clxBlack, clxWhite);
  GraysColorList: TColorList = (clxBlack, clxDarkGray, clxGray, clxLightGray, clxDimGray,
      clxDarkSlateGray, clxLightSlateGray, clxSlateGray, clxWhite, clxGray);
  SouthwestColorList: TColorList = (clxYellow, clxDarkOrange, clxRed, clxIndianRed, clxGold,
      clxChocolate, clxLightSalmon, clxFirebrick, clxSienna, clxRosyBrown);
  RedsColorList: TColorList = (clxDarkorange, clxRed, clxDarkRed, clxMaroon,
      clxIndianRed, clxSalmon, clxCoral, clxTomato, clxOrangeRed, clxFirebrick );
  GreensColorList: TColorList = (clxPaleGoldenrod, clxOlive, clxGreenYellow,
      clxLightGreen, clxSeagreen, clxKhaki, clxGreen, clxLawnGreen, clxMediumAquamarine,
      clxDarkGoldenrod );
  BluesColorList: TColorList = (clxBlue, clxLightCyan, clxLightSkyblue,
      clxDarkBlue, clxMediumTurquoise, clxCyan,
      clxPowderBlue, clxRoyalBlue, clxMidnightBlue, clxTeal);
  VioletsColorList: TColorList = (clxDarkViolet, clxDarkOrchid, clxFuchsia,
      clxMediumVioletRed, clxBlueViolet, clxMediumPurple, clxDeepPink,
      clxViolet, clxPlum, clxHotPink);
  DystopiaColorList: TColorList = (clxLightSteelBlue, clxLightSlateGray, clxLightGray,
      clxSteelBlue, clxSlateGray, clxSilver, clxDarkSlateBlue,
      clxDarkSlateGray, clxGainsboro, clxDarkGray);
  AntiqueColorList: TColorList = (clxTeal, clxSilver, clxSeashell, clxLavender,
      clxKhaki, clxPurple, clxRoyalBlue, clxDarkGoldenrod,
      clxPaleGoldenrod, clxFireBrick);

//  TRSColorScheme = (csCustom, csDefaultColors, csBlackandWhite, csGrays,
//    csReds, csGreens, csBlues, csAntique);
const
  RSIdentityMatrix: TRSMatrix = (m11: 1.0; m12: 0.0; m13: 0.0;
                                  m21: 0.0; m22: 1.0; m23: 0.0;
                                  m31: 0.0; m32: 0.0; m33: 1.0);
  CanvasColorPrefix = {$IFDEF FMX}'cla'{$ELSE}'cl'{$ENDIF};
  AllCorners = [Low(TRSCorner)..High(TRSCorner)];
  AllSides = [Low(TRSSide)..High(TRSSide)];

{$IFNDEF DELPHI16PLUS}
  cPI: Single = 3.141592654;
  CurveKappa = 0.5522847498;
  Epsilon: Single = 1E-40;
{$ENDIF}
var
  USFormatSettings: TFormatSettings; // used for correct string to float convertion

implementation

uses
{$IFNDEF FMX}
  RSVclGraphics,
{$ENDIF}
{$IFDEF DELPHI16PLUS}
  UIConsts,
{$ENDIF}
  Generics.Collections, Math, StrUtils, HelperFunctions;

function CanvasRect( aRect: TRSRect ): TCanvasRect;
begin
  result := {$IFDEF FMX}aRect;{$ELSE}RSRectToRect(aRect);{$ENDIF}
end;

function CanvasRect( aRect: TRect ): TCanvasRect; 
begin
  result := {$IFDEF FMX}RectToRSRect(aRect);{$ELSE}aRect;{$ENDIF}
end;

function CanvasRect( Left, Top, Right, Bottom: Integer ): TCanvasRect;
begin
  result := {$IFDEF FMX}RectF(Left, Top, Right, Bottom);{$ELSE}Rect(Left, Top, Right, Bottom);{$ENDIF}
end;

function CanvasRect( Left, Top, Right, Bottom: Single ): TCanvasRect;
begin
  result := {$IFDEF FMX}RectF(Left, Top, Right, Bottom);{$ELSE}Rect(Round(Left), Round(Top), Round(Right), Round(Bottom));{$ENDIF}
end;

function CanvasPoint( X, Y: Integer): TCanvasPoint;
begin
  result := {$IFDEF FMX}PointF(X,Y);{$ELSE}Point(X,Y);{$ENDIF}
end;

function CanvasPoint( X, Y: Single): TCanvasPoint;
begin
  result := {$IFDEF FMX}PointF(X,Y);{$ELSE}Point(Round(X),Round(Y));{$ENDIF}
end;

function CanvasRound( X: Double ): TExtCanvasPixel;
begin
  result := {$IFDEF FMX}X;{$ELSE}Round(X);{$ENDIF}
end;

function CanvasTrunc( X: Double ): TExtCanvasPixel;
begin
  result := {$IFDEF FMX}X;{$ELSE}Trunc(X);{$ENDIF}
end;

function CanvasDiv( X, Y: Int64 ): TExtCanvasPixel;
begin
  result := {$IFDEF FMX}X/Y;{$ELSE}X div Y;{$ENDIF}
end;

function CanvasDiv( X, Y: Double ): TExtCanvasPixel;
begin
  result := {$IFDEF FMX}X/Y;{$ELSE}Round(X / Y);{$ENDIF}
end;

function CanvasMax( const Data: array of TCanvasPixel ): TCanvasPixel;
begin
  result := {$IFDEF FMX}MaxValue(Data);{$ELSE}MaxIntValue(Data);{$ENDIF}
end;

function CanvasMin( const Data: array of TCanvasPixel ): TCanvasPixel;
begin
  result := {$IFDEF FMX}MinValue(Data);{$ELSE}MinIntValue(Data);{$ENDIF}
end;

function CanvasCenterPoint( const Rect: TCanvasRect ): TCanvasPoint;
begin
  result := {$IFDEF FMX}RSCenterPoint(Rect);{$ELSE}CenterPoint(Rect);{$ENDIF}
end;

function CanvasColorToString(Color: TCanvasColor): string;
begin
  result := {$IFDEF FMX}AlphaColorToString(Color);{$ELSE}ColorToString(Color);{$ENDIF}
end;

function StringToCanvasColor(const S: string): TCanvasColor;
begin
  result := {$IFDEF FMX}StringToAlphaColor(S);{$ELSE}StringToColor(S);{$ENDIF}
end;

function CanvasColorToIdent(Color: Longint; var Ident: string): Boolean;
begin
  result := {$IFDEF FMX}AlphaColorToIdent(Color, Ident);{$ELSE}ColorToIdent(Color, Ident);{$ENDIF}
end;

function IdentToCanvasColor(const Ident: string; var Color: Integer): Boolean;
begin
  result := {$IFDEF FMX}IdentToAlphaColor(Ident, Color);{$ELSE}IdentToColor(Ident, Color);{$ENDIF}
end;

function PointToRSPoint( const Pt: TPoint ): TRSPoint;
begin
     result.X := Pt.X;
     result.Y := Pt.Y;
end;

function PointsToRSPoints( const Pt: TPoints ): TRSPoints;
var
  i: Integer;
begin
  SetLength(Result, Length(Pt));
  for i := 0 to Length(Pt)-1 do
    result[i] := RSPoint(Pt[i].X, Pt[i].Y);
end;

function GetToken(var S: string; const Separators: string; const Stop: string = ''): string;
var
  I: Integer;
  Breaks: string;
begin
  Breaks := Separators + Stop;
  for I := 0 to Length(S) - 1 do
  begin
    if Pos(S[i], Breaks) > 0 then
      Break;
  end;

  Result := Copy(S, 0, I);
  S := Copy(S,I + 1,MaxInt);
end;

function RSPointToString(R: TRSPoint; FormatStr: String): string;
begin
  Result := '(' + FormatFloat(FormatStr, R.X, USFormatSettings) + ',' + FormatFloat(FormatStr, R.Y, USFormatSettings) + ')';
end;

function StringToRSPoint(S: string): TRSPoint;
begin
  try
    GetToken(S, ',()');
    Result.X := StrToFloat(GetToken(S, ',()'), USFormatSettings);
    Result.Y := StrToFloat(GetToken(S, ',()'), USFormatSettings);
  except
    Result := RSPoint(0, 0);
  end;
end;

function RSPointToPoint( const Pt: TRSPoint ): TPoint;
begin
     result.X := round(Pt.X);
     result.Y := round(Pt.Y);
end;

function RSPointToCanvasPoint( const Pt: TRSPoint ): TCanvasPoint;
begin
     result.X := CanvasRound(Pt.X);
     result.Y := CanvasRound(Pt.Y);
end;

function RSPointsToPoints( const Pt: TRSPoints ): TPoints;
var
  i: Integer;
begin
  SetLength(Result, Length(Pt));
  for i := 0 to Length(Pt)-1 do
    result[i] := Point(Round(Pt[i].X), Round(Pt[i].Y));
end;

function RSPoint( const X, Y: TRSGfxFloat ): TRSPoint;
begin
     result.X := X;
     result.Y := Y;
end;

function RSPoint( const Pt: TPoint ): TRSPoint;
begin
     result.X := Pt.X;
     result.Y := Pt.Y;
end;

function PointsEqual( const Pt1, Pt2: TRSPoint ): Boolean; overload;
begin
     result := SameValue( Pt1.X, Pt2.X ) and
               SameValue( Pt1.Y, Pt2.Y );
end;

function EqualsRect( const R1, R2: TRSRect ): Boolean;
begin
     result := SameValue( R1.Left, R2.Left ) and
               SameValue( R1.Top, R2.Top ) and
               SameValue( R1.Right, R2.Right ) and
               SameValue( R1.Bottom, R2.Bottom );
end;

function EqualsRect( const R1, R2: TRect ): Boolean;
begin
     result := ( R1.Left = R2.Left ) and
               ( R1.Top = R2.Top ) and
               ( R1.Right = R2.Right ) and
               ( R1.Bottom = R2.Bottom );
end;

function RSRectHeight( const ARect: TRect ): Integer;
begin
  result := ARect.Bottom-ARect.Top;
end;

function RSRectHeight( const ARect: TRSRect ): TRSGfxFloat;
begin
  result := ARect.Bottom-ARect.Top;
end;

function RSRectWidth( const ARect: TRect ): Integer;
begin
  result := ARect.Right-ARect.Left;
end;

function RSRectWidth( const ARect: TRSRect ): TRSGfxFloat;
begin
  result := ARect.Right-ARect.Left;
end;

procedure RSInflateRect( var R: TRect; const DX, DY: Integer );
begin
  R.Left := R.Left - DX;
  R.Right := R.Right + DX;
  R.Top := R.Top - DY;
  R.Bottom := R.Bottom + DY;
end;

procedure RSInflateRect( var R: TRSRect; const DX, DY: Single );
begin
  R.Left := R.Left - DX;
  R.Right := R.Right + DX;
  R.Top := R.Top - DY;
  R.Bottom := R.Bottom + DY;
end;

function RSDistance( const Pt1, Pt2: TRSPoint ): Single;
begin
  result := Sqrt(Sqr(Pt1.X-Pt2.X) + Sqr(Pt1.Y-Pt2.Y));
end;

function RSCenterPoint( const Pt1, Pt2: TRSPoint; Midpoint: Single = 0.5 ): TRSPoint;
begin
  if pt2.X = pt1.X then
    result.X := pt2.X
  else
    result.X := pt1.X + ((pt2.X - pt1.X)*Midpoint);
  if pt2.Y = pt1.Y then
    result.Y := pt2.Y
  else
    result.Y := pt1.Y + ((pt2.Y - pt1.Y)*Midpoint);
end;

function RSCenterPoint( const Pt: TRSPoints ): TRSPoint;
var
  i: Integer;
begin
  result := RSPoint(0,0);
  if Length(Pt) = 0 then Exit;

  for i := 0 to Length(Pt) - 1 do
  begin
    result.X := result.X + Pt[i].X;
    result.Y := result.Y + Pt[i].Y;
  end;
  result.X := result.X / Length(Pt);
  result.Y := result.Y / Length(Pt);
end;

function RSCenterPoint( const Pt: TPoints ): TRSPoint;
var
  i: Integer;
begin
  result := RSPoint(0,0);
  if Length(Pt) = 0 then Exit;

  for i := 0 to Length(Pt) - 1 do
  begin
    result.X := result.X + Pt[i].X;
    result.Y := result.Y + Pt[i].Y;
  end;
  result.X := result.X / Length(Pt);
  result.Y := result.Y / Length(Pt);
end;

function ScalePoints( const Pt: TRSPoints; ScaleDX, ScaleDY: Single ): TRSPoints;
var
  Center: TRSPoint;
  i: Integer;
begin
  Center := RSCenterPoint( Pt );
  SetLength(result, Length(Pt));
  for i := 0 to Length(Pt)-1 do
  begin
    result[i].X := (Pt[i].X-Center.X)*ScaleDX+Center.X;
    result[i].Y := (Pt[i].Y-Center.Y)*ScaleDY+Center.Y;
  end;
end;

function ScalePoints( const Pt: TPoints; ScaleDX, ScaleDY: Single ): TPoints;
var
  Center: TRSPoint;
  i: Integer;
begin
  Center := RSCenterPoint( Pt );
  SetLength(result, Length(Pt));
  for i := 0 to Length(Pt)-1 do
  begin
    result[i].X := Round((Pt[i].X-Center.X)*ScaleDX+Center.X);
    result[i].Y := Round((Pt[i].Y-Center.Y)*ScaleDY+Center.Y);
  end;
end;


function ScaleRect( ARect: TRect; DX, DY: Single ): TRect;
var
  ADX, ADY: Integer;
begin
     // scales the rectangle centered
     ADX := Round(DX / 2);
     ADY := Round(DY / 2);
     ARect.Left := ARect.Left - ADX;
     ARect.Top := ARect.Top - ADY;
     ARect.Right := ARect.Right + ADX;
     ARect.Bottom := ARect.Bottom + ADY;
     result := ARect;
end;

function ScaleRect( ARect: TRSRect; DX, DY: Single ): TRSRect;
begin
     // scales the rectangle centered
     DX := DX / 2;
     DY := DY / 2;
     ARect.Left := ARect.Left - DX;
     ARect.Top := ARect.Top - DY;
     ARect.Right := ARect.Right + DX;
     ARect.Bottom := ARect.Bottom + DY;
     result := ARect;
end;

function RSRect( const ALeft, ATop, ARight, ABottom: TRSGfxFloat ): TRSRect;
begin
     Result.Left := ALeft;
     Result.Top := ATop;
     Result.Right := ARight;
     Result.Bottom := ABottom;
end;

function RSRect( const TopLeft, BottomRight: TRSPoint ): TRSRect;
begin
  result.TopLeft := TopLeft;
  result.BottomRight := BottomRight;
end;

function RectToRSRect( const ARect: TRect ): TRSRect;
begin
     Result.Left := ARect.Left;
     Result.Top := ARect.Top;
     Result.Right := ARect.Right;
     Result.Bottom := ARect.Bottom;
end;

function RSRect( const ARect: TRect ): TRSRect;
begin
  result := RectToRSRect(aRect);
end;

function RSRect( const ARect: TRSRect ): TRSRect;
begin
  result := aRect;
end;

function RSRectToRect( const ARect: TRSRect ): TRect;
begin
     Result.Left := Round(ARect.Left);
     Result.Top := Round(ARect.Top);
     Result.Right := Round(ARect.Right);
     Result.Bottom := Round(ARect.Bottom);
end;

function RSCenterPoint( const ARect: TRSRect ): TRSPoint;
begin
     with ARect do
     begin
          Result.X := (Right - Left) / 2 + Left;
          Result.Y := (Bottom - Top) / 2 + Top;
     end;
end;

function CenterRect( const Pt: TPoint; Width, Height: Integer ): TRect;
begin
     // create rectangle centered around point
     Width := Width div 2;
     Height := Height div 2;
     result := Rect( Pt.X - Width, Pt.Y - Height, Pt.X + Width, Pt.Y + Height );
end;

function CenterRect( const X, Y: Integer; Width, Height: Integer ): TRect;
begin
     // create rectangle centered around point
     Width := Width div 2;
     Height := Height div 2;
     result := Rect( X - Width, Y - Height, X + Width, Y + Height );
end;

function CenterRect( const Pt: TRSPoint; Width, Height: TRSGfxFloat ): TRSRect;
begin
     // create rectangle centered around point
     Width := Width / 2;
     Height := Height / 2;
     result := RSRect( Pt.X - Width, Pt.Y - Height, Pt.X + Width, Pt.Y + Height );
end;

function CenterRect( const X, Y: TRSGfxFloat; Width, Height: TRSGfxFloat ): TRSRect;
begin
     // create rectangle centered around point
     Width := Width / 2;
     Height := Height / 2;
     result := RSRect( X - Width, Y - Height, X + Width, Y + Height );
end;

function CenterRect( var R: TRect; const Bounds: TRect ): TRect;
var
  R2: TRSRect;
begin
  R2 := RectToRSRect(R);
  CenterRect( R2, RectToRSRect(Bounds) );
  R := RSRectToRect(R2);
  result := R;
end;

function CenterRect( var R: TRSRect; const Bounds: TRSRect ): TRSRect;
begin
  OffsetRSRect(R, -R.Left, -R.Top);
  OffsetRSRect(R, Round((RSRectWidth(Bounds) - RSRectWidth(R)) / 2), Round((RSRectHeight(Bounds) - RSRectHeight(R)) / 2));
  OffsetRSRect(R, Bounds.Left, Bounds.Top);
  Result := R;
end;

function OffsetRSRect(var R: TRSRect; DX, DY: Single): Boolean;
begin
  if @R <> nil then // Test to increase compatiblity with Windows
  begin
    R.Left := R.Left + DX;
    R.Right := R.Right + DX;
    R.Top := R.Top + DY;
    R.Bottom := R.Bottom + DY;
    Result := True;
  end
  else
    Result := False;
end;

function RSIntersectRect( const R1, R2: TRect ): Boolean;
var
  R: TRect;
begin
  result := RSIntersectRect( R1, R2, R );
end;

function RSIntersectRect( const R1, R2: TRSRect ): Boolean;
var
  R: TRSRect;
begin
  result := RSIntersectRect( R1, R2, R );
end;

function RSIntersectRect( const R1, R2: TRect; var R: TRect ): Boolean;
begin
     if R1.Left > R2.Left then
        R.Left := R1.Left
     else
         R.Left := R2.Left;
     if R1.Right < R2.Right then
        R.Right := R1.Right
     else
         R.Right := R2.Right;
     if R1.Top > R2.Top then
        R.Top := R1.Top
     else
         R.Top := R2.Top;
     if R1.Bottom < R2.Bottom then
        R.Bottom := R1.Bottom
     else
         R.Bottom := R2.Bottom;

     result := not ((R.Left > R.Right) or
                    (R.Top > R.Bottom));
end;

function RSIntersectRect( const R1, R2: TRSRect; var R: TRSRect ): Boolean;
begin
     if R1.Left > R2.Left then
        R.Left := R1.Left
     else
         R.Left := R2.Left;
     if R1.Right < R2.Right then
        R.Right := R1.Right
     else
         R.Right := R2.Right;
     if R1.Top > R2.Top then
        R.Top := R1.Top
     else
         R.Top := R2.Top;
     if R1.Bottom < R2.Bottom then
        R.Bottom := R1.Bottom
     else
         R.Bottom := R2.Bottom;

     result := not ((R.Left > R.Right) or
                    (R.Top > R.Bottom));
end;

function RSIntersectLine( const V1, V2, V3, V4: TRSPoint; var IntersectionPt: TRSPoint ): Boolean;
var
  a1, b1, c1: Single;
  r1, r2: Single;
  a2, b2, c2: Single;
  r3, r4: Single;
  denom: Single;
  offset: Single;
  num: Single;
begin
  result := False;
  // returns the intersection of two line segments: v1, v2 are first line segment
  // v3, v4 are other line segment
  a1 := v2.Y - v1.Y;
  b1 := v1.X - v2.X;
  c1 := v2.X * v1.Y - v1.X * v2.Y;

  r3 := a1 * v3.X + b1 * v3.Y + c1;
  r4 := a1 * v4.X + b1 * v4.Y + c1;

  // Check signs of r3 and r4.  If both point 3 and point 4 lie on
  // same side of line 1, the line segments do not intersect.
  if (r3 <> 0) and (r4 <> 0) and (Sign(r3) = Sign(r4)) then
    Exit;

  a2 := v4.Y - v3.Y;
  b2 := v3.X - v4.X;
  c2 := v4.X * v3.Y - v3.X * v4.Y;

  r1 := a2 * v1.X + b2 * v1.Y + c2;
  r2 := a2 * v2.X + b2 * v2.Y + c2;
  // Check signs of r1 and r2.  If both point 1 and point 2 lie on
  // same side of line 2, the line segments do not intersect.
  if (r1 <> 0) and (r2 <> 0) and (Sign(r1) = Sign(r2)) then
    Exit;

  denom := a1 * b2 - a2 * b1;
  if Denom = 0 then Exit; // Colinear

  // Line segments intersect: compute intersection point
  result := True;
  Offset := abs(denom) / 2;

  // The denom/2 is to get rounding instead of truncating.  It
  // is added or subtracted to the numerator, depending upon the
  // sign of the numerator
  num := b1 * c2 - b2 * c1;
  if num < 0 then
    IntersectionPt.X := (num - offset) / denom
  else
    IntersectionPt.X := (num + offset) / denom;

  num :=  a2 * c1 - a1 * c2;
  if num < 0 then
    IntersectionPt.Y := (num - offset) / denom
  else
    IntersectionPt.Y := (num + offset) / denom;
end;

function RSIntersectRect( const V1, V2: TRSPoint; const R: TRSRect; var Pts: TRSPoints ): Boolean;
  procedure AddPoint( const Pt: TRSPoint );
  begin
    result := True;
    SetLength(Pts, Length(Pts)+1);
    Pts[Length(Pts)-1] := Pt;
  end;
var
  IntersectionPt: TRSPoint;
begin
  // returns true if line segment inside the rect or intersects a side of the
  // rect.
  SetLength(Pts, 0);
  // return true if inside the rect
  result := (V1.X >= R.Left) and (V1.X <= R.Right) and
            (V1.Y >= R.Top) and (V1.Y <= R.Bottom) and
            (V2.X >= R.Left) and (V2.X <= R.Right) and
            (V2.Y >= R.Top) and (V2.Y <= R.Bottom);
  // find intersection of a line segment with a rect.
  if RSIntersectLine(V1, V2, R.TopLeft, RSPoint(R.Right, R.Top), IntersectionPt) then
    AddPoint( IntersectionPt );
  if RSIntersectLine(V1, V2, R.TopLeft, RSPoint(R.Left, R.Bottom), IntersectionPt) then
    AddPoint( IntersectionPt );
  if RSIntersectLine(V1, V2, RSPoint(R.Right, R.Top), R.BottomRight, IntersectionPt) then
    AddPoint( IntersectionPt );
  if RSIntersectLine(V1, V2, RSPoint(R.Left, R.Bottom), R.BottomRight, IntersectionPt) then
    AddPoint( IntersectionPt );
end;

function PtInRSRect( const Pt: TPoint; const ARect: TRSRect ): Boolean;
begin
     result := (Pt.X >= ARect.Left) and (Pt.X < ARect.Right) and
               (Pt.Y >= ARect.Top) and (Pt.Y < ARect.Bottom);
end;

function PtInRSRect( const Pt: TRSPoint; const ARect: TRSRect ): Boolean;
begin
     result := (Pt.X >= ARect.Left) and (Pt.X < ARect.Right) and
               (Pt.Y >= ARect.Top) and (Pt.Y < ARect.Bottom);
end;

function GetPointOnRect( const ARect: TRect;
  ConnectionPoint: TConnectionPoint): TPoint;
begin
     case ConnectionPoint of
          cpTop:          result := Point( ARect.Left+ ((ARect.Right - ARect.Left) div 2),
                                           ARect.Top );
          cpTopRight:     result := Point( ARect.Right, ARect.Top );
          cpRight:        result := Point( ARect.Right,
                                           ARect.Top + ((ARect.Bottom - ARect.Top) div 2) );
          cpBottomRight:  result := ARect.BottomRight;
          cpBottom:       result := Point( ARect.Left+ ((ARect.Right - ARect.Left) div 2),
                                           ARect.Bottom );
          cpBottomLeft:   result := Point( ARect.Left, ARect.Bottom );
          cpLeft:         result := Point( ARect.Left,
                                           ARect.Top + ((ARect.Bottom - ARect.Top) div 2) );
          cpTopLeft:      result := ARect.TopLeft;
     end;
end;

function GetCanvasPointOnRect( const ARect: TCanvasRect;
  ConnectionPoint: TConnectionPoint): TCanvasPoint;
begin
{$IFDEF FMX}
  result := GetRSPointOnRSRect( ARect, ConnectionPoint );
{$ELSE}
  result := GetPointOnRect( ARect, ConnectionPoint );
{$ENDIF}
end;


function GetRSPointOnRSRect( const ARect: TRSRect;
  ConnectionPoint: TConnectionPoint): TRSPoint;
begin
     case ConnectionPoint of
          cpTop:          result := RSPoint( ARect.Left+ ((ARect.Right - ARect.Left) / 2),
                                           ARect.Top );
          cpTopRight:     result := RSPoint( ARect.Right, ARect.Top );
          cpRight:        result := RSPoint( ARect.Right,
                                           ARect.Top + ((ARect.Bottom - ARect.Top) / 2) );
          cpBottomRight:  result := ARect.BottomRight;
          cpBottom:       result := RSPoint( ARect.Left+ ((ARect.Right - ARect.Left) / 2),
                                           ARect.Bottom );
          cpBottomLeft:   result := RSPoint( ARect.Left, ARect.Bottom );
          cpLeft:         result := RSPoint( ARect.Left,
                                           ARect.Top + ((ARect.Bottom - ARect.Top) / 2) );
          cpTopLeft:      result := ARect.TopLeft;
     end;
end;

function RSCenterPoint( const ARect: TRect ): TRSPoint; overload;
begin
     with ARect do
     begin
          Result.X := (Right - Left) / 2 + Left;
          Result.Y := (Bottom - Top) / 2 + Top;
     end;
end;

function FitRect( aMatrix: TRSMatrix; const aRect, TargetRect: TRSRect ): TRSRect;
begin
  // modify matrix with translates and scales to fit aRect into TargetRect
  aMatrix := RSMatrixMultiply( CreateTranslateRSMatrix(TargetRect.Left, TargetRect.Top), aMatrix );
  if (RSRectWidth(aRect) * RSRectHeight(aRect)) <> 0 then
    aMatrix := RSMatrixMultiply( CreateScaleRSMatrix( RSRectWidth(TargetRect)/RSRectWidth(aRect),
                                                      RSRectHeight(TargetRect)/RSRectHeight(aRect) ),
                                 aMatrix );
  result := RSMatrixTransform(aMatrix, aRect);
end;

function FitRect( var aRect: TRect; const boundsRect: TRect; Grow: Boolean ): Single;
var
  ratio: Single;
begin
  Result := 1;
  if RSRectWidth(BoundsRect) * RSRectHeight(BoundsRect) = 0 then
    Exit;

  if (RSRectWidth(aRect) / RSRectWidth(BoundsRect)) > (RSRectHeight(aRect) / RSRectHeight(BoundsRect)) then
    ratio := RSRectWidth(aRect) / RSRectWidth(BoundsRect)
  else
    ratio := RSRectHeight(aRect) / RSRectHeight(BoundsRect);

  if (ratio < 1) and (not Grow) then
  begin
    aRect := Rect(0, 0, RSRectWidth(aRect), RSRectHeight(aRect));
  end
  else
  begin
    aRect := Rect(0, 0, Round(RSRectWidth(aRect) / ratio), Round(RSRectHeight(aRect) / ratio));
  end;

  Result := ratio;
  aRect := CenterRect(aRect, BoundsRect);
end;

function FitRect( var aRect: TRSRect; const boundsRect: TRSRect; Grow: Boolean ): Single;
var
  ratio: Single;
begin
  Result := 1;
  if RSRectWidth(BoundsRect) * RSRectHeight(BoundsRect) = 0 then
    Exit;

  if (RSRectWidth(aRect) / RSRectWidth(BoundsRect)) > (RSRectHeight(aRect) / RSRectHeight(BoundsRect)) then
    ratio := RSRectWidth(aRect) / RSRectWidth(BoundsRect)
  else
    ratio := RSRectHeight(aRect) / RSRectHeight(BoundsRect);

  if (ratio < 1) and (not Grow) then
  begin
    aRect := RSRect(0, 0, RSRectWidth(aRect), RSRectHeight(aRect));
  end
  else
  begin
    aRect := RSRect(0, 0, RSRectWidth(aRect) / ratio, RSRectHeight(aRect) / ratio);
  end;

  Result := ratio;
  aRect := CenterRect(aRect, BoundsRect);
end;

function FixRect( const ARect: TRect ): TRect; overload;
begin
     result := ARect;
     if ARect.Bottom < ARect.Top then
     begin
          result.Top := ARect.Bottom;
          result.Bottom := ARect.Top;
     end;
     if ARect.Right < ARect.Left then
     begin
          result.Left := ARect.Right;
          result.Right := ARect.Left;
     end;
end;

function FixRect( const ARect: TRSRect ): TRSRect; overload;
begin
     result := ARect;
     if ARect.Bottom < ARect.Top then
     begin
          result.Top := ARect.Bottom;
          result.Bottom := ARect.Top;
     end;
     if ARect.Right < ARect.Left then
     begin
          result.Left := ARect.Right;
          result.Right := ARect.Left;
     end;
end;

function InCenterPoint( const A, B, C: TPoint ): TPoint;
var
   AB, AC, BC: TRSGfxFloat;
begin
     // InCenterPoint returns the InCenter of a triangle
     AB := Sqrt(Sqr(A.Y-B.Y) + Sqr(A.X-B.X));
     AC := Sqrt(Sqr(A.Y-C.Y) + Sqr(A.X-C.X));
     BC := Sqrt(Sqr(B.Y-C.Y) + Sqr(B.X-C.X));
     result := Point(
            Round((BC * A.X + AC * B.X + AB * C.X) / (AB + AC + BC)),
            Round((BC * A.Y + AC * B.Y + AB * C.Y) / (AB + AC + BC)) );
end;

function InCenterPoint( const A, B, C: TRSPoint ): TRSPoint;
var
   AB, AC, BC: TRSGfxFloat;
begin
     // InCenterPoint returns the InCenter of a triangle
     AB := Sqrt(Sqr(A.Y-B.Y) + Sqr(A.X-B.X));
     AC := Sqrt(Sqr(A.Y-C.Y) + Sqr(A.X-C.X));
     BC := Sqrt(Sqr(B.Y-C.Y) + Sqr(B.X-C.X));
     result := RSPoint(
            ((BC * A.X + AC * B.X + AB * C.X) / (AB + AC + BC)),
            ((BC * A.Y + AC * B.Y + AB * C.Y) / (AB + AC + BC)) );
end;

function OffsetPoint( var Pt: TPoint; const X, Y: Integer ): TPoint; overload;
begin
     result := Pt;
     Pt.X := Pt.X + X;
     Pt.Y := Pt.Y + Y;
end;

function OffsetPoint( var Pt: TPoint; LineEndPt: TPoint; Dist: TRSGfxFloat ): TPoint; overload;
var
   Len: TRSGfxFloat;
begin
     // Move the Point in the DIST direction of the end point
     result := Pt;
     Len := Sqrt(Sqr(LineEndPt.Y-Pt.Y) + Sqr(LineEndPt.X-Pt.X));
     if Len = 0 then Exit;
     Pt.Y := Pt.Y + Round((LineEndPt.Y-Pt.Y)*(Dist/Len));
     Pt.X := Pt.X + Round((LineEndPt.X-Pt.X)*(Dist/Len));
end;

function OffsetPoint( var Pt: TRSPoint; const X, Y: Single ): TRSPoint; overload;
begin
     result := Pt;
     Pt.X := Pt.X + X;
     Pt.Y := Pt.Y + Y;
end;

function OffsetPoint( var Pt: TRSPoint; LineEndPt: TRSPoint; Dist: TRSGfxFloat ): TRSPoint; overload;
var
   Len: TRSGfxFloat;
begin
     // Move the Point in the DIST direction of the end point
     result := Pt;
     Len := Sqrt(Sqr(LineEndPt.Y-Pt.Y) + Sqr(LineEndPt.X-Pt.X));
     if Len = 0 then Exit;
     Pt.Y := Pt.Y + ((LineEndPt.Y-Pt.Y)*(Dist/Len));
     Pt.X := Pt.X + ((LineEndPt.X-Pt.X)*(Dist/Len));
end;

function PointInPath(const Points: TPoints; X, Y: Integer): Boolean;
var
  Count, K, J: Integer;
begin
  Result := False;
  Count := Length(Points);
  J := Count-1;
  for K := 0 to Count-1 do
  begin
    if ((Points[K].Y <=Y) and (Y < Points[J].Y)) or
      ((Points[J].Y <=Y) and (Y < Points[K].Y)) then
    begin
    if (x < (Points[j].X - Points[K].X) *
       (y - Points[K].Y) /
       (Points[j].Y - Points[K].Y) + Points[K].X) then
        Result := not Result;
    end;
    J := K;
  end;
end;

function PointInPath(const Points: TRSPoints; X, Y: Single): Boolean;
var
  Count, K, J: Integer;
begin
  Result := False;
  Count := Length(Points);
  J := Count-1;
  for K := 0 to Count-1 do
  begin
    if ((Points[K].Y <=Y) and (Y < Points[J].Y)) or
      ((Points[J].Y <=Y) and (Y < Points[K].Y)) then
    begin
    if (x < (Points[j].X - Points[K].X) *
       (y - Points[K].Y) /
       (Points[j].Y - Points[K].Y) + Points[K].X) then
        Result := not Result;
    end;
    J := K;
  end;
end;

function PointInPath(const Pt: TPoint; const Points: TPoints): TPathClickLocation;
var
  Count: Integer;
  pCurrent, pNext: TPoint;
  i: Integer;
  Crossing: Integer;
  d, d2, d3: Double;
begin
  // determines if Pt is inside path, outside path, or exactly on path
  // Epsilon decides if something is ON path
  Result := clOutsidePath;
  Count := Length(Points);
  // if Path is a line or point then exit
  if Count < 3 then Exit;

  Crossing := 0;
  pCurrent := Points[0];
  for i := 1 to Count do
  begin
    // get next line segment (pCurrent to pNext)
    if i < Count then
      pNext := Points[i]
    else
      pNext := Points[0];

    if (pNext.Y = pt.Y) then
    begin
      if (pNext.X = pt.X) or
         ((pCurrent.Y = pt.Y) and
          ((pNext.X > pt.X) = (pCurrent.X < pt.X))) then
      begin
        result := clOnPath;
        Exit;
      end;
    end;

    if ((pCurrent.Y < pt.Y) <> (pNext.Y < pt.Y)) then
    begin
      if (pCurrent.X >= pt.X) then
      begin
        if (pNext.X > pt.X) then
          Crossing := 1 - Crossing
        else
        begin
          d2 := (pCurrent.X - pt.X);
          d3 := (pNext.X - pt.X);
          d := d2 * (pNext.Y - pt.Y) - d3 * (pCurrent.Y - pt.Y);
          if (d = 0) then
          begin
            result := clOnPath;
            Exit;
          end;
          if ((d > 0) = (pNext.Y > pCurrent.Y)) then
            Crossing := 1 - Crossing;
        end;
      end else
      begin
        if (pNext.X > pt.X) then
        begin
          d2 := (pCurrent.X - pt.X);
          d3 := (pNext.X - pt.X);
          d := d2 * (pNext.Y - pt.Y) - d3 * (pCurrent.Y - pt.Y);
          if (d = 0) then
          begin
            result := clOnPath;
            Exit;
          end;
          if ((d > 0) = (pNext.Y > pCurrent.Y)) then
            Crossing := 1 - Crossing;
        end;
      end;
    end;
    pCurrent := pNext;
  end;
  case Crossing of
    -1: result := clOnPath;
    0:  result := clOutsidePath;
    1:  result := clInsidePath;
  end;
end;

function PointInPath(const Pt: TRSPoint; const Points: TRSPoints): TPathClickLocation;
var
  Count: Integer;
  pCurrent, pNext: TRSPoint;
  i: Integer;
  Crossing: Integer;
  d, d2, d3: Double;
begin
  // Note that function does not work too well for detecting clOnPath :-(

  // determines if Pt is inside path, outside path, or exactly on path
  // Epsilon decides if something is ON path
  Result := clOutsidePath;
  Count := Length(Points);
  // if Path is a line or point then exit
  if Count < 3 then Exit;

  Crossing := 0;
  pCurrent := Points[0];
  for i := 1 to Count do
  begin
    // get next line segment (pCurrent to pNext)
    if i < Count then
      pNext := Points[i]
    else
      pNext := Points[0];

    if (pNext.Y = pt.Y) then
    begin
      if (pNext.X = pt.X) or
         ((pCurrent.Y = pt.Y) and
          ((pNext.X > pt.X) = (pCurrent.X < pt.X))) then
      begin
        result := clOnPath;
        Exit;
      end;
    end;

    if ((pCurrent.Y < pt.Y) <> (pNext.Y < pt.Y)) then
    begin
      if (pCurrent.X >= pt.X) then
      begin
        if (pNext.X > pt.X) then
          Crossing := 1 - Crossing
        else
        begin
          d2 := (pCurrent.X - pt.X);
          d3 := (pNext.X - pt.X);
          d := d2 * (pNext.Y - pt.Y) - d3 * (pCurrent.Y - pt.Y);
          if (d = 0) then
          begin
            result := clOnPath;
            Exit;
          end;
          if ((d > 0) = (pNext.Y > pCurrent.Y)) then
            Crossing := 1 - Crossing;
        end;
      end else
      begin
        if (pNext.X > pt.X) then
        begin
          d2 := (pCurrent.X - pt.X);
          d3 := (pNext.X - pt.X);
          d := d2 * (pNext.Y - pt.Y) - d3 * (pCurrent.Y - pt.Y);
          if (d = 0) then
          begin
            result := clOnPath;
            Exit;
          end;
          if ((d > 0) = (pNext.Y > pCurrent.Y)) then
            Crossing := 1 - Crossing;
        end;
      end;
    end;
    pCurrent := pNext;
  end;
  case Crossing of
    -1: result := clOnPath;
    0:  result := clOutsidePath;
    1:  result := clInsidePath;
  end;
end;

{$IFDEF DELPHI5ORLESS}
function CenterPoint( const ARect: TRect ): TPoint;
begin
     with ARect do
     begin
          Result.X := (Right - Left) div 2 + Left;
          Result.Y := (Bottom - Top) div 2 + Top;
     end;
end;
function PointsEqual( const Pt1, Pt2: TPoint ): Boolean;
begin
     result := SameValue( Pt1.X, Pt2.X ) and
               SameValue( Pt1.Y, Pt2.Y );
end;
{$ENDIF}

{$IFNDEF DELPHI16PLUS}
function TRSRect.Fit(BoundsRect: TRSRect): Single;
var
  ratio: Single;
begin
  Result := 1;
  if RSRectWidth(BoundsRect) * RSRectHeight(BoundsRect) = 0 then
    Exit;

  if (RSRectWidth(Self) / RSRectWidth(BoundsRect)) > (RSRectHeight(Self) / RSRectHeight(BoundsRect)) then
    ratio := RSRectWidth(Self) / RSRectWidth(BoundsRect)
  else
    ratio := RSRectHeight(Self) / RSRectHeight(BoundsRect);

  if ratio < 1 then
  begin
    Self := RSRect(0, 0, RSRectWidth(Self), RSRectHeight(Self));
  end
  else
  begin
    Self := RSRect(0, 0, RSRectWidth(Self) / ratio, RSRectHeight(Self) / ratio);
  end;

  Result := ratio;
  CenterRect(Self, BoundsRect);
end;
{$ENDIF}

{$IFDEF CLR}
class function TRSRect.Create(const ALeft, ATop, ARight, ABottom: TRSGfxFloat): TRSRect;
begin
  Result.Left := ALeft;
  Result.Top := ATop;
  Result.Right := ARight;
  Result.Bottom := ABottom;
end;

class function TRSRect.Create(const ATopLeft, ABottomRight: TPoint): TRSRect;
begin
  Result := TRSRect.Create(ATopLeft.X, ATopLeft.Y,
                         ABottomRight.X, ABottomRight.Y);
end;

class function TRSRect.Create(const ATopLeft, ABottomRight: TRSPoint): TRSRect;
begin
  Result := TRSRect.Create(ATopLeft.X, ATopLeft.Y,
                         ABottomRight.X, ABottomRight.Y);
end;

class function TRSRect.Create(const ATopLeft: TPoint; const ASize: TSize): TRSRect;
begin
  Result := TRSRect.Create(ATopLeft.X, ATopLeft.Y,
                         ATopLeft.X + ASize.Width, ATopLeft.Y + ASize.Height);
end;

class function TRSRect.Create(const ATopLeft: TRSPoint; const ASize: TSize): TRSRect;
begin
  Result := TRSRect.Create(ATopLeft.X, ATopLeft.Y,
                         ATopLeft.X + ASize.Width, ATopLeft.Y + ASize.Height);
end;

class function TRSRect.Bounds(const ALeft, ATop, AWidth, AHeight: TRSGfxFloat): TRSRect;
begin
  Result := TRSRect.Create(ALeft, ATop, ALeft + AWidth, ATop + AHeight);
end;

class function TRSRect.Empty: TRSRect;
begin
  Result := TRSRect.Create(0, 0, 0, 0);
end;

function TRSRect.ToString: string;
begin
  Result := '{X=' + Left.ToString + // DO NOT LOCALIZE
            ',Y=' + Top.ToString + // DO NOT LOCALIZE
            ',Width=' + GetWidth.ToString + // DO NOT LOCALIZE
            ',Height=' + GetHeight.ToString + '}'; // DO NOT LOCALIZE
end;

function TRSRect.get_TopLeft: TRSPoint;
begin
  Result := TRSPoint.Create(Left, Top);
end;

procedure TRSRect.set_TopLeft(const Value: TRSPoint);
begin
  Left := Value.X;
  Top := Value.Y;
end;

function TRSRect.get_BottomRight: TRSPoint;
begin
  Result := TRSPoint.Create(Right, Bottom);
end;

procedure TRSRect.set_BottomRight(const Value: TRSPoint);
begin
  Right := Value.X;
  Bottom := Value.Y;
end;

function TRSRect.GetWidth: TRSGfxFloat;
begin
  Result := Right - Left;
end;

function TRSRect.GetHeight: TRSGfxFloat;
begin
  Result := Bottom - Top;
end;

function TRSRect.GetSize: TSize;
begin
  Result := TSize.Create(Round(GetWidth), Round(GetHeight));
end;

function TRSRect.GetCenterPoint: TRSPoint;
begin
  Result := TRSPoint.Create(GetWidth / 2 + Left,
                          GetHeight / 2 + Top);
end;

function TRSRect.FixRect: TRSRect;
begin
     result := TRSRect.Create(Left, Top, Right, Bottom);
     if Bottom < Top then
     begin
          result.Top := Bottom;
          result.Bottom := Top;
     end;
     if Right < Left then
     begin
          result.Left := Right;
          result.Right := Left;
     end;
end;

function TRSRect.IsEmpty: Boolean;
begin
  Result := (Right <= Left) or
            (Bottom <= Top);
end;

function TRSRect.Equals(const Value: TRSRect): Boolean;
begin
  Result := (Left = Value.Left) and
            (Top = Value.Top) and
            (Right = Value.Right) and
            (Bottom = Value.Bottom);
end;

function TRSRect.Equals(const Value: TSize): Boolean;
begin
  Result := Value.Equals(GetSize);
end;

function TRSRect.Contains(const X, Y: TRSGfxFloat): Boolean;
begin
  Result := (X >= Left) and
            (Y >= Top) and
            (X < Right) and
            (Y < Bottom);
end;

function TRSRect.Contains(const Value: TPoint): Boolean;
begin
  Result := Contains(Value.X, Value.Y);
end;

function TRSRect.Contains(const Value: TRSPoint): Boolean;
begin
  Result := Contains(Value.X, Value.Y);
end;

function TRSRect.Contains(const Value: TRSRect): Boolean;
begin
  Result := (Value.Left >= Left) and
            (Value.Top >= Top) and
            (Value.Right < Right) and
            (Value.Bottom < Bottom);
end;

// This is here because Types can't know about Math
class function TRSRect.IfThen(const ATest: Boolean; const ATrue, AFalse: TRSGfxFloat): TRSGfxFloat;
begin
  if ATest then
    Result := ATrue
  else
    Result := AFalse;
end;

function TRSRect.Intersects(const Other: TRSRect): Boolean;
begin
  Result := ((Left < Other.Right) and
             (Right > Other.Left)) and
            ((Top < Other.Bottom) and
             (Bottom > Other.Top));
end;

function TRSRect.GetIntersect(const Other: TRSRect): TRSRect;
begin
  Result := TRSRect.Create(IfThen(Other.Left > Left, Other.Left, Left),
                         IfThen(Other.Top > Top, Other.Top, Top),
                         IfThen(Other.Right < Right, Other.Right, Right),
                         IfThen(Other.Bottom < Bottom, Other.Bottom, Bottom));
end;

function TRSRect.GetIntersect(const Other: TRSRect; out AResult: TRSRect): Boolean;
begin
  AResult := GetIntersect(Other);
  Result := not AResult.IsEmpty;
  if not Result then
    AResult := Empty;
end;

function TRSRect.GetUnion(const Other: TRSRect): TRSRect;
begin
  if Other.IsEmpty then
    Result := Self
  else
    Result := TRSRect.Create(IfThen(Other.Left < Left, Other.Left, Left),
                           IfThen(Other.Top < Top, Other.Top, Top),
                           IfThen(Other.Right > Right, Other.Right, Right),
                           IfThen(Other.Bottom > Bottom, Other.Bottom, Bottom));
end;

function TRSRect.GetUnion(const Other: TRSRect; out AResult: TRSRect): Boolean;
begin
  AResult := GetUnion(Other);
  Result := not AResult.IsEmpty;
  if not Result then
    AResult := Empty;
end;

function TRSRect.GetOffset(const X, Y: TRSGfxFloat): TRSRect;
begin
  Result := TRSRect.Create(Left + X, Top + Y, Right + X, Bottom + Y);
end;

function TRSRect.GetOffset(const Value: TRSPoint): TRSRect;
begin
  Result := GetOffset(Value.X, Value.Y);
end;

function TRSRect.GetInflate(const DeltaWidth, DeltaHeight: TRSGfxFloat): TRSRect;
begin
  Result := TRSRect.Create(Left - DeltaWidth, Top - DeltaHeight,
                         Right + DeltaWidth, Bottom + DeltaHeight);
end;

function TRSRect.GetInflate(const Delta: TRSPoint): TRSRect;
begin
  Result := GetInflate(Delta.X, Delta.Y);
end;
{$ENDIF}

{ TRSPoint }
{$IFNDEF DELPHI16PLUS}
class operator TRSPoint.Equal(const Lhs, Rhs: TRSPoint): Boolean;
begin
  Result := SameValue(Lhs.X, Rhs.X) and SameValue(Lhs.Y, Rhs.Y);
end;

class operator TRSPoint.NotEqual(const Lhs, Rhs: TRSPoint): Boolean;
begin
  result := not (Lhs = Rhs);
end;

class operator TRSPoint.Add(const Lhs, Rhs: TRSPoint): TRSPoint;
begin
  Result.X := Lhs.X + Rhs.X;
  Result.Y := Lhs.Y + Rhs.Y;
end;

class operator TRSPoint.Subtract(const Lhs, Rhs: TRSPoint): TRSPoint;
begin
  Result.X := Lhs.X - Rhs.X;
  Result.Y := Lhs.Y - Rhs.Y;
end;

class operator TRSPoint.Multiply(const Lhs, Rhs: TRSPoint): TRSPoint;
begin
  Result.X := Lhs.X * Rhs.X;
  Result.Y := Lhs.Y * Rhs.Y;
end;
{$ENDIF}

{$IFDEF CLR}
class function TRSPoint.Create(const AX, AY: TRSGfxFloat): TRSPoint;
begin
  Result.X := AX;
  Result.Y := AY;
end;

class function TRSPoint.Empty: TRSPoint;
begin
  Result := Create(0, 0);
end;

function TRSPoint.ToString: string;
begin
  Result := '{X=' + X.ToString + // DO NOT LOCALIZE
            ',Y=' + Y.ToString + '}'; // DO NOT LOCALIZE
end;

function TRSPoint.Left: TRSGfxFloat;
begin
  Result := X;
end;

function TRSPoint.Top: TRSGfxFloat;
begin
  Result := Y;
end;

function TRSPoint.IsEmpty: Boolean;
begin
  Result := (X = 0) and (Y = 0);
end;

function TRSPoint.IsInvalid: Boolean;
begin
  Result := (X = -1) and (Y = -1);
end;

function TRSPoint.Equals(const AX, AY: TRSGfxFloat): Boolean;
begin
  Result := (X = AX) and (Y = AY);
end;

function TRSPoint.Equals(const Value: TRSPoint): Boolean;
begin
  Result := (X = Value.X) and (Y = Value.Y);
end;
{$ENDIF}

{$IFNDEF FMX}
{ TVclCanvasHelper }

procedure TVclCanvasHelper.DrawLine(P1, P2: TRSPoint);
begin
  MoveTo(Round(P1.X), Round(P1.Y));
  LineTo(Round(P2.X), Round(P2.Y));
end;

procedure TVclCanvasHelper.DrawLine(P1, P2: TPoint);
begin
  MoveTo(P1.X, P1.Y);
  LineTo(P2.X, P2.Y);
end;

procedure TVclCanvasHelper.Polygon(const Points: TRSPoints);
begin
  Polygon(RSPointsToPoints(Points));
end;

procedure TVclCanvasHelper.Polyline(const Points: TRSPoints);
begin
  Polyline(RSPointsToPoints(Points));
end;

function TVclCanvasHelper.GetPenWidth: Integer;
begin
  result := Pen.Width;
end;

procedure TVclCanvasHelper.SetPenWidth(const Value: Integer);
begin
  Pen.Width := Round(Value);
end;
{$ENDIF}

function DoubleToByte(D: Double): Byte;
var
  I: Integer;
begin
  I := Round(D);
  if (I > 255) then I := 255;
  if (I < 0) then I := 0;
  Result := I;
end;

{$IFDEF FMX}
function SetGradientColors( const Gradient: TGradient; const Colors: array of TCanvasColor ): TGradient;
{ Purpose: To set a gradients color collection to an array of colors }
var
  I: Integer;
begin
  // fix gradient points
//  if Gradient.Points.Count > 2 then
//    while Gradient.Points.Count > Length(Colors) do
//      Gradient.Points.Delete(Gradient.Points.Count-2);
//  if Length(Colors) > 0 then Gradient.Color := Colors[Low(Colors)];
//  if Length(Colors) > 1 then Gradient.Color1 := Colors[High(Colors)];
//  for I := Low(Colors)+1 to High(Colors)-1 do
//    with Gradient.Points.Add as TGradientPoint do
//    begin
//      Color := Colors[i];
//      Offset := (1/(Length(Colors)-1))*i;
//    end;
    while Gradient.Points.Count > 0 do
      Gradient.Points.Delete(0);
  for I := Low(Colors) to High(Colors) do
    with Gradient.Points.Add as TGradientPoint do
    begin
      Color := Colors[i];
      Offset := (1/(Length(Colors)-1))*i;
    end;
  result := Gradient;
end;

procedure DrawGradient( const Canvas: TCanvas; const ARect: TCanvasRect;
  const Colors: array of TCanvasColor; Style: TGradientStyle;
  StartPt, StopPt: TPointF; FillAsRect: Boolean; const Opacity: Single );
var
  CanvasState: {$IFDEF USE_RSCANVASSAVE}TRSCanvasSaveState{$ELSE}TCanvasSaveState{$ENDIF};
begin
  CanvasState := Canvas.SaveState{$IFDEF USE_RSCANVASSAVE}(AllParts){$ELSE}{$ENDIF};
  try
    SetGradientColors( Canvas.Brush.Gradient, Colors );
    if Style = TGradientStyle.{$IFDEF DELPHI20PLUS}Linear{$ELSE}gsLinear{$ENDIF} then
    begin
      Canvas.Brush.Gradient.StartPosition.Point := StartPt;
      Canvas.Brush.Gradient.StopPosition.Point := StopPt;
    end
    else
      Canvas.Brush.Gradient.RadialTransform.RotationCenter.Point := StartPt;
    Canvas.Brush.Gradient.Style := Style;
    Canvas.Fill.Kind := TBrushKind.{$IFDEF DELPHI20PLUS}Gradient{$ELSE}bkGradient{$ENDIF};
    if FillAsRect then
      Canvas.FillRect(ARect, Opacity)
    else
      Canvas.FillEllipse(ARect,Opacity);
  finally
    Canvas.RestoreState(CanvasState);
  end;
end;
{$ENDIF}

procedure DirectionGradient( const Canvas: TCanvas; const CanvasRect: TCanvasRect;
   const Colors: array of TCanvasColor; Direction: TGradientDirection;
   LineSize: Integer; const Opacity: Single);
{$IFDEF FMX}
var
  Reverse: Array of TCanvasColor;
  procedure ReverseColors;
  var
    I: Integer;
  begin
    SetLength(Reverse, Length(Colors));
    for I := Low(Colors) to High(Colors) do
      Reverse[High(Colors)-i] := Colors[i];
  end;
{$ENDIF}
begin
{$IFDEF FMX}
  case Direction of
    gdDownRight:
      DrawGradient(Canvas, CanvasRect,
        Colors, TGradientStyle.{$IFDEF DELPHI20PLUS}Linear{$ELSE}gsLinear{$ENDIF},
        PointF(0,0), PointF(1,1),
        True, Opacity);
    gdDown:
      LineGradient(Canvas, CanvasRect,
        Colors, LineSize,
        diVertical,
        Opacity);
    gdDownLeft:
      DrawGradient(Canvas, CanvasRect,
        Colors, TGradientStyle.{$IFDEF DELPHI20PLUS}Linear{$ELSE}gsLinear{$ENDIF},
        PointF(1,0), PointF(0,1),
        True, Opacity);
    gdRight:
      LineGradient(Canvas, CanvasRect,
        Colors, LineSize,
        diHorizontal,
        Opacity);
    gdLeft:
    begin
      ReverseColors;
      LineGradient(Canvas, CanvasRect,
        Reverse, LineSize,
        diHorizontal,
        Opacity);
    end;
    gdUpRight:
      DrawGradient(Canvas, CanvasRect,
        Colors, TGradientStyle.{$IFDEF DELPHI20PLUS}Linear{$ELSE}gsLinear{$ENDIF},
        PointF(0,1), PointF(1,0),
        True, Opacity);
    gdUp:
    begin
      ReverseColors;
      LineGradient(Canvas, CanvasRect,
        Reverse, LineSize,
        diVertical,
        Opacity);
    end;
    gdUpLeft:
      DrawGradient(Canvas, CanvasRect,
        Colors, TGradientStyle.{$IFDEF DELPHI20PLUS}Linear{$ELSE}gsLinear{$ENDIF},
        PointF(1,1), PointF(0,0),
        True, Opacity);
  end;
{$ELSE}
  RSVclGraphics.DirectionGradient( Canvas, CanvasRect, Colors, Direction, LineSize );
  Exit;
{$ENDIF}
end;

procedure LineGradient(const Canvas: TCanvas; const ARect: TCanvasRect;
  const Colors: array of TCanvasColor; LineSize: Integer;
  const Orientation: TGradientOrientation; const Opacity: Single);
{ Purpose: Draws a line gradient in the canvas bounding rect in the specified   }
{          orientation.                                                         }
{ Arguments:                                                                    }
{          Canvas        - The canvas on which to draw                          }
{          CanvasRect    - The rectangle defining the bounds of the Canvas      }
{          Colors        - An array of ordered colors to be used in the gradient}
{          LineSize      - The step (or coarseness) of the gradient             }
{          Orientation   - Indicates whether the gradient draws horizontal or   }
{                          vertical                                             }
{          Opacity       - transparency (0=clear, 1=solid) of the gradient,     }
{                          FMX ONLY                                             }
var
  R,G,B: Double;
{$IFDEF FMX}
  A: Double;
  RGB1: TAlphaColorRec;
  RGB2: TAlphaColorRec;
  CRGB: TAlphaColorRec;
{$ELSE}
  RGB1: TRGB;
  RGB2: TRGB;
  CRGB: TRGB;
{$ENDIF}
  Color1, Color2: TCanvasColor;
  i: Integer;
  X1, X2, Y1, Y2: TCanvasPixel;
  WideByLineSize: Double;
  Wide, CWide, Start, Finish, Cp: TCanvasPixel;
  CR: TCanvasRect;
  ColorCount: Integer;
  HoldBrushColor: TCanvasColor;
begin
  if (ARect.Right = ARect.Left) or (ARect.Bottom = ARect.Top) then Exit;
{$IFDEF FMX}
  // optimize with FMX
  if LineSize <= 1 then
  begin
    if (Orientation = diHorizontal) then
      DrawGradient( Canvas, ARect, Colors, TGradientStyle.{$IFDEF DELPHI20PLUS}Linear{$ELSE}gsLinear{$ENDIF},
        PointF(0,0), PointF(1,0), True, Opacity )
    else {diVertical}
      DrawGradient( Canvas, ARect, Colors, TGradientStyle.{$IFDEF DELPHI20PLUS}Linear{$ELSE}gsLinear{$ENDIF},
        PointF(0, 0), PointF(0, 1), True, Opacity );
    Exit;
  end;
{$ENDIF}
  if (LineSize <= 0) then LineSize := 1;
  ColorCount := Length(Colors);
  if (ColorCount <= 0) then Exit;

  i := Low(Colors);
  X1 := ARect.Left;
  Y1 := ARect.Top;
  X2 := ARect.Right;
  if X1 > X2 then
  begin
       X1 := X2;
       X2 := ARect.Left;
  end;
  Y2 := ARect.Bottom;
  if Y1 > Y2 then
  begin
       Y1 := Y2;
       Y2 := ARect.Top;
  end;

  if (Orientation = diHorizontal) then begin
    Wide := CanvasMax([1,CanvasDiv((X2 - X1), CanvasMax([1,(ColorCount-1)]))]);
    Start := X1;
    Finish := X2;
    Cp := X1;
  end
  else begin
    Wide := CanvasMax([1,CanvasDiv((Y2 - Y1), CanvasMax([1,(ColorCount-1)]))]);
    Start := Y1;
    Finish := Y2;
    Cp := Y1;
  end;
  WideByLineSize := Wide*LineSize;

  if (ColorCount <= 1) then begin
    Color1 := Colors[i];
    Color2 := Color1;
  end
  else begin
    Color1 := Colors[i];
    Color2 := Colors[i+1];
    Inc(i);
  end;
  CWide := Wide;
{$IFDEF FMX}
  RGB1.Color := Color1;
  RGB2.Color := Color2;
  A := RGB1.A;
{$ELSE}
  RGB1 := ColorToRGB(Color1);
  RGB2 := ColorToRGB(Color2);
{$ENDIF}
  R := RGB1.R;
  G := RGB1.G;
  B := RGB1.B;

  { Save the current settings }
  HoldBrushColor := Canvas.Brush.Color;
  while (Cp <= Finish) do begin
    if (Orientation = diHorizontal) then
      CR := CanvasRect(Cp, Y1, Cp+LineSize, Y2)
    else {diVertical}
      CR := CanvasRect(X1, Cp, X2, Cp+LineSize);

    CRGB.R := DoubleToByte(R);
    CRGB.G := DoubleToByte(G);
    CRGB.B := DoubleToByte(B);
{$IFDEF FMX}
    CRGB.A := DoubleToByte(A);
    Canvas.Brush.Color := CRGB.Color;
    Canvas.FillRect(CR, Opacity);
{$ELSE}
    Canvas.Brush.Color := RGBToColor(CRGB);
    Canvas.FillRect(CR);
{$ENDIF}


    Cp := Cp + LineSize;
    if ((Cp-Start) >= CWide) then begin
      CWide := CWide + Wide;
      Inc(i);
      if (i >= Low(Colors) + ColorCount) then Dec(i);
      Color1 := Color2;
      Color2 := Colors[i];
{$IFDEF FMX}
      RGB1.Color := Color1;
      RGB2.Color := Color2;
{$ELSE}
      RGB1 := ColorToRGB(Color1);
      RGB2 := ColorToRGB(Color2);
{$ENDIF}
    end;
    R := R + (RGB2.R - RGB1.R) / WideByLineSize;
    G := G + (RGB2.G - RGB1.G) / WideByLineSize;
    B := B + (RGB2.B - RGB1.B) / WideByLineSize;
{$IFDEF FMX}
    A := A + (RGB2.A - RGB1.A) / WideByLineSize;
{$ENDIF}
  end;
  { Restore the brush }
  Canvas.Brush.Color := HoldBrushColor;
end;

procedure EllipseGradient(const Canvas: TCanvas; const ARect: TCanvasRect;
  FocalX, FocalY: Single;
  const Colors: array of TCanvasColor; LineSize: Single;
  const Opacity: Single); overload;
{ Purpose: Draws an ellipse gradient in the canvas bounding rect.               }
{ Arguments:                                                                    }
{          Canvas        - The canvas on which to draw                          }
{          CanvasRect    - The rectangle defining the bounds of the Canvas      }
{          FocalX, FocalY- Focal point of radial gradient, between 0 and 1      }
{          Colors        - An array of ordered colors to be used in the gradient}
{          LineSize      - The step (or coarseness) of the gradient             }
{          Opacity       - transparency (0=clear, 1=solid) of the gradient,     }
{                          FMX ONLY                                             }
var
  R,G,B : Double;
{$IFDEF FMX}
  A: Double;
  RGB1: TAlphaColorRec;
  RGB2: TAlphaColorRec;
  CRGB: TAlphaColorRec;
  CanvasState: {$IFDEF USE_RSCANVASSAVE}TRSCanvasSaveState{$ELSE}TCanvasSaveState{$ENDIF};
  Bitmap: TBitmap;
{$ELSE}
  RGB1: TRGB;
  RGB2: TRGB;
  CRGB: TRGB;
  HoldPenWidth: TCanvasPixel;
  HoldPenStyle: TPenStyle;
  HoldBrushColor: TCanvasColor;
{$ENDIF}
  Color1, Color2: TCanvasColor;
  i: Integer;
  X1, X2, Y1, Y2: TCanvasPixel;
  Wide, CWide: TCanvasPixel;
  Cx, Cy: Single;
  Cx2, Cy2: Single;
  FX, FY: TCanvasPixel;
  X1Step, X2Step: Single;
  Y1Step, Y2Step: Single;
  ColorCount: Integer;
  ACanvas: TCanvas;
  j: Single;
begin
  if (ARect.Right = 0) or (ARect.Bottom = 0) then Exit;
{$IFDEF FMX}
  // optimize with FMX
  if LineSize <= 1 then
  begin
    DrawGradient( Canvas, ARect, Colors, TGradientStyle.{$IFDEF DELPHI20PLUS}Radial{$ELSE}gsRadial{$ENDIF},
        PointF(FocalX, FocalY), PointF(0,0), False, Opacity );
    Exit;
  end;
{$ENDIF}
  if (LineSize <= 0) then LineSize := 1;
  ColorCount := Length(Colors);
  if (ColorCount <= 0) then Exit;

  i := Low(Colors);
  X1 := ARect.Left;
  Y1 := ARect.Top;
  X2 := ARect.Right;
  Y2 := ARect.Bottom;
  FX := round(X1+abs((X2-X1)*FocalX));
  FY := round(Y1+abs((Y2-Y1)*FocalY));
  Wide := CanvasMax([1, CanvasDiv(CanvasMax([abs(FX-X1), abs(FX-X2), abs(FY-Y2), abs(FY - Y1)]), CanvasMax([1,ColorCount-1]))]);

  if (ColorCount <= 1) then begin
    Color1 := Colors[i];
    Color2 := Color1;
  end
  else begin
    Color1 := Colors[i];
    Color2 := Colors[i+1];
    Inc(i);
  end;
  CWide := Wide;
{$IFDEF FMX}
  RGB1.Color := Color1;
  RGB2.Color := Color2;
  A := RGB1.A;
{$ELSE}
  RGB1 := ColorToRGB(Color1);
  RGB2 := ColorToRGB(Color2);
{$ENDIF}
  R := RGB1.R;
  G := RGB1.G;
  B := RGB1.B;
  Cx := X1;
  Cy := Y1;
  Cx2 := X2;
  Cy2 := Y2;
  X1Step := (FX-aRect.Left)/(RSRectWidth(aRect)/LineSize);
  X2Step := (aRect.Right-FX)/(RSRectWidth(aRect)/LineSize);
  Y1Step := (FY-aRect.Top)/(RSRectHeight(aRect)/LineSize);
  Y2Step := (aRect.Bottom-FY)/(RSRectHeight(aRect)/LineSize);
//  LineSize := MaxValue([X1Step, X2Step, Y1Step, Y2Step]);
  j := 0;
  ACanvas := Canvas;
{$IFDEF FMX}
  CanvasState := Canvas.SaveState{$IFDEF USE_RSCANVASSAVE}(AllParts){$ELSE}{$ENDIF};
  Bitmap := nil;
  if Opacity < 1 then
  begin
    Bitmap := TBitmap.Create(Round(CanvasMax([ARect.Right,ARect.Left])), Round(CanvasMax([ARect.Bottom,ARect.Top])));
    Bitmap.Canvas.BeginScene;
    ACanvas := Bitmap.Canvas;
    CopyCanvasState(ACanvas, Canvas);
  end;
  try
{$ENDIF}
    with ACanvas do begin
      { Save the current settings }
  {$IFDEF FMX}
      Stroke.Kind := TBrushKind.{$IFDEF DELPHI20PLUS}None{$ELSE}bkNone{$ENDIF};
      Brush.Kind := TBrushKind.{$IFDEF DELPHI20PLUS}Solid{$ELSE}bkSolid{$ENDIF};
  {$ELSE}
      HoldPenWidth := Pen.Width;
      HoldPenStyle := Pen.Style;
      HoldBrushColor := Brush.Color;
      { Modify the Pen }
      Pen.Width := 0;
      Pen.Style := psClear;
  {$ENDIF}
      while (Cx < FX) and (Cy < FY) do begin
        CRGB.R := DoubleToByte(R);
        CRGB.G := DoubleToByte(G);
        CRGB.B := DoubleToByte(B);
    {$IFDEF FMX}
        CRGB.A := DoubleToByte(A);
        ACanvas.Brush.Color := CRGB.Color;
    {$ELSE}
        ACanvas.Brush.Color := RGBToColor(CRGB);
    {$ENDIF}

        ACanvas.Ellipse(Round(Cx), Round(Cy), Round(Cx2), Round(Cy2));

        Cx := Cx + X1Step;
        Cx2 := Cx2 - X2Step;
        Cy := Cy + Y1Step;
        Cy2 := Cy2 - Y2Step;
        j := j + MaxValue([x1Step, x2Step, y1Step, y2Step]);
        if (J >= CWide) then begin
          CWide := CWide + Wide;
          Inc(i);
          if (i > Low(Colors) + ColorCount) then Dec(i);
          Color1 := Color2;
          Color2 := Colors[i];
    {$IFDEF FMX}
          RGB1.Color := Color1;
          RGB2.Color := Color2;
    {$ELSE}
          RGB1 := ColorToRGB(Color1);
          RGB2 := ColorToRGB(Color2);
    {$ENDIF}
        end;
        R := R + (RGB2.R - RGB1.R) / Wide * LineSize;
        G := G + (RGB2.G - RGB1.G) / Wide * LineSize;
        B := B + (RGB2.B - RGB1.B) / Wide * LineSize;
    {$IFDEF FMX}
        A := A + (RGB2.A - RGB1.A) / Wide * LineSize;
    {$ENDIF}
      end;

      { Restore the old settings }
  {$IFDEF FMX}
      RestoreState(CanvasState);
  {$ELSE}
      Pen.Width := HoldPenWidth;
      Pen.Style := HoldPenStyle;
      Brush.Color := HoldBrushColor;
  {$ENDIF}
    end;
{$IFDEF FMX}
  finally
    if Bitmap <> nil then
    begin
      Bitmap.Canvas.EndScene;
      Canvas.DrawBitmap( Bitmap,
        ARect,
        ARect, Opacity );
      Bitmap.Free;
    end;
  end;
{$ENDIF}
end;

procedure EllipseGradient(const Canvas: TCanvas; const ARect: TCanvasRect;
  const Colors: array of TCanvasColor; LineSize: Integer;
  const Opacity: Single); overload;
{ Purpose: Draws an ellipse gradient in the canvas bounding rect.               }
{ Arguments:                                                                    }
{          Canvas        - The canvas on which to draw                          }
{          CanvasRect    - The rectangle defining the bounds of the Canvas      }
{          Colors        - An array of ordered colors to be used in the gradient}
{          LineSize      - The step (or coarseness) of the gradient             }
{          Opacity       - transparency (0=clear, 1=solid) of the gradient,     }
{                          FMX ONLY                                             }
var
  R,G,B : Double;
{$IFDEF FMX}
  A: Double;
  RGB1: TAlphaColorRec;
  RGB2: TAlphaColorRec;
  CRGB: TAlphaColorRec;
  CanvasState: {$IFDEF USE_RSCANVASSAVE}TRSCanvasSaveState{$ELSE}TCanvasSaveState{$ENDIF};
  Bitmap: TBitmap;
{$ELSE}
  RGB1: TRGB;
  RGB2: TRGB;
  CRGB: TRGB;
  HoldPenWidth: TCanvasPixel;
  HoldPenStyle: TPenStyle;
  HoldBrushColor: TCanvasColor;
{$ENDIF}
  Color1, Color2: TCanvasColor;
  i: Integer;
  X1, X2, Y1, Y2: TCanvasPixel;
  Wide, CWide: TCanvasPixel;
  Cx, Cy, Cp: TCanvasPixel;
  ColorCount: Integer;
  ControlAxis: TCanvasPixel;
  ACanvas: TCanvas;
begin
  if (ARect.Right = 0) or (ARect.Bottom = 0) then Exit;
{$IFDEF FMX}
  // optimize with FMX
  if LineSize <= 1 then
  begin
    DrawGradient( Canvas, ARect, Colors, TGradientStyle.{$IFDEF DELPHI20PLUS}Radial{$ELSE}gsRadial{$ENDIF},
        PointF(0.5,0.5), PointF(0,0), False, Opacity );
    Exit;
  end;
{$ENDIF}
  if (LineSize <= 0) then LineSize := 1;
  ColorCount := Length(Colors);
  if (ColorCount <= 0) then Exit;

  i := Low(Colors);
  X1 := ARect.Left;
  Y1 := ARect.Top;
  X2 := ARect.Right;
  Y2 := ARect.Bottom;
  if (Y2-Y1)>(X2-X1) then begin
    Wide := CanvasMax([1, CanvasDiv(CanvasDiv((Y2 - Y1), CanvasMax([1,ColorCount-1])), 2)]);
    ControlAxis := Y2;
  end
  else begin
    Wide := CanvasMax([1, CanvasDiv(CanvasDiv((X2 - X1), CanvasMax([1,ColorCount-1])), 2)]);
    ControlAxis := X2;
  end;

  if (ColorCount <= 1) then begin
    Color1 := Colors[i];
    Color2 := Color1;
  end
  else begin
    Color1 := Colors[i];
    Color2 := Colors[i+1];
    Inc(i);
  end;
  CWide := Wide;
{$IFDEF FMX}
  RGB1.Color := Color1;
  RGB2.Color := Color2;
  A := RGB1.A;
{$ELSE}
  RGB1 := ColorToRGB(Color1);
  RGB2 := ColorToRGB(Color2);
{$ENDIF}
  R := RGB1.R;
  G := RGB1.G;
  B := RGB1.B;
  Cx := X1;
  Cy := Y1;

  ACanvas := Canvas;
{$IFDEF FMX}
  CanvasState := Canvas.SaveState{$IFDEF USE_RSCANVASSAVE}(AllParts){$ELSE}{$ENDIF};
  Bitmap := nil;
  if Opacity < 1 then
  begin
    Bitmap := TBitmap.Create(Round(CanvasMax([ARect.Right,ARect.Left])), Round(CanvasMax([ARect.Bottom,ARect.Top])));
    Bitmap.Canvas.BeginScene;
    ACanvas := Bitmap.Canvas;
    CopyCanvasState(ACanvas, Canvas);
  end;
  try
{$ENDIF}
    with ACanvas do begin
      { Save the current settings }
  {$IFDEF FMX}
      Stroke.Kind := TBrushKind.{$IFDEF DELPHI20PLUS}None{$ELSE}bkNone{$ENDIF};
      Brush.Kind := TBrushKind.{$IFDEF DELPHI20PLUS}Solid{$ELSE}bkSolid{$ENDIF};
  {$ELSE}
      HoldPenWidth := Pen.Width;
      HoldPenStyle := Pen.Style;
      HoldBrushColor := Brush.Color;
      { Modify the Pen }
      Pen.Width := 0;
      Pen.Style := psClear;
  {$ENDIF}

      while (Cx < (X1+CanvasDiv((X2-X1), 2))) and (Cy < (Y1+CanvasDiv((Y2-Y1), 2))) do begin
        CRGB.R := DoubleToByte(R);
        CRGB.G := DoubleToByte(G);
        CRGB.B := DoubleToByte(B);
    {$IFDEF FMX}
        CRGB.A := DoubleToByte(A);
        ACanvas.Brush.Color := CRGB.Color;
    {$ELSE}
        ACanvas.Brush.Color := RGBToColor(CRGB);
    {$ENDIF}

        ACanvas.Ellipse(Cx, Cy, X1+(X2-Cx), Y1+(Y2-Cy));

        Cx := Cx + LineSize;
        Cy := Cy + LineSize;
        if (ControlAxis = Y2) then
          Cp := Cy - Y1
        else
          Cp := Cx - X1;
        if (Cp >= CWide) then begin
          CWide := CWide + Wide;
          Inc(i);
          if (i > Low(Colors) + ColorCount) then Dec(i);
          Color1 := Color2;
          Color2 := Colors[i];
    {$IFDEF FMX}
          RGB1.Color := Color1;
          RGB2.Color := Color2;
    {$ELSE}
          RGB1 := ColorToRGB(Color1);
          RGB2 := ColorToRGB(Color2);
    {$ENDIF}
        end;
        R := R + (RGB2.R - RGB1.R) / Wide * LineSize;
        G := G + (RGB2.G - RGB1.G) / Wide * LineSize;
        B := B + (RGB2.B - RGB1.B) / Wide * LineSize;
    {$IFDEF FMX}
        A := A + (RGB2.A - RGB1.A) / Wide * LineSize;
    {$ENDIF}
      end;

      { Restore the old settings }
  {$IFDEF FMX}
      RestoreState(CanvasState);
  {$ELSE}
      Pen.Width := HoldPenWidth;
      Pen.Style := HoldPenStyle;
      Brush.Color := HoldBrushColor;
  {$ENDIF}
    end;
{$IFDEF FMX}
  finally
    if Bitmap <> nil then
    begin
      Bitmap.Canvas.EndScene;
      Canvas.DrawBitmap( Bitmap,
        ARect,
        ARect, Opacity );
      Bitmap.Free;
    end;
  end;
{$ENDIF}
end;

procedure RectGradient(const Canvas: TCanvas; const ARect: TCanvasRect;
  const Colors: array of TCanvasColor; LineSize: Integer;
  const Opacity: Single);
{ Purpose: Draws a rectangle gradient in the canvas bounding rect.              }
{ Arguments:                                                                    }
{          Canvas        - The canvas on which to draw                          }
{          CanvasRect    - The rectangle defining the bounds of the Canvas      }
{          Colors        - An array of ordered colors to be used in the gradient}
{          LineSize      - The step (or coarseness) of the gradient             }
{          Opacity       - transparency (0=clear, 1=solid) of the gradient,     }
{                          FMX ONLY                                             }
var
  R,G,B : Double;
{$IFDEF FMX}
  RGB1: TAlphaColorRec;
  RGB2: TAlphaColorRec;
  CRGB: TAlphaColorRec;
  A: Double;
  CanvasState: {$IFDEF USE_RSCANVASSAVE}TRSCanvasSaveState{$ELSE}TCanvasSaveState{$ENDIF};
{$ELSE}
  RGB1: TRGB;
  RGB2: TRGB;
  CRGB: TRGB;
  HoldBrushColor: TCanvasColor;
{$ENDIF}
  Color1, Color2: TCanvasColor;
  i: Integer;
  X1, X2, Y1, Y2: TCanvasPixel;
  Wide, CWide: TCanvasPixel;
  Cx, Cy, Cp: TCanvasPixel;
  ColorCount: Integer;
  ControlAxis: TCanvasPixel;
  Rect1: TCanvasRect;
begin
  if (ARect.Right = 0) or (ARect.Bottom = 0) then Exit;
{$IFDEF FMX}
  // optimize with FMX
  if LineSize <= 1 then
  begin
    DrawGradient( Canvas, ARect, Colors, TGradientStyle.{$IFDEF DELPHI20PLUS}Radial{$ELSE}gsRadial{$ENDIF},
        PointF(0.5,0.5), PointF(0,0), True, Opacity );
    Exit;
  end;
{$ENDIF}
  if (LineSize <= 0) then LineSize := 1;
  ColorCount := Length(Colors);
  if (ColorCount <= 0) then Exit;
  i := Low(Colors);
  X1 := ARect.Left;
  Y1 := ARect.Top;
  X2 := ARect.Right;
  Y2 := ARect.Bottom;

  if ((Y2-Y1)>(X2-X1)) then begin
    Wide := CanvasMax([1, CanvasDiv(CanvasDiv((Y2 - Y1), CanvasMax([1,ColorCount-1])), 2)]);
    ControlAxis := Y2;
  end
  else begin
    Wide := CanvasMax([1, CanvasDiv(CanvasDiv((X2 - X1), CanvasMax([1,ColorCount-1])), 2)]);
    ControlAxis := X2;
  end;

  if (ColorCount <= 1) then begin
    Color1 := Colors[i];
    Color2 := Color1;
  end
  else begin
    Color1 := Colors[i];
    Color2 := Colors[i+1];
    Inc(i);
  end;
  CWide := Wide;
{$IFDEF FMX}
  RGB1.Color := Color1;
  RGB2.Color := Color2;
  A := RGB1.A;
{$ELSE}
  RGB1 := ColorToRGB(Color1);
  RGB2 := ColorToRGB(Color2);
{$ENDIF}
  R := RGB1.R;
  G := RGB1.G;
  B := RGB1.B;
  Cx := X1;
  Cy := Y1;

  { Save the current settings }
{$IFDEF FMX}
  CanvasState := Canvas.SaveState{$IFDEF USE_RSCANVASSAVE}(AllParts){$ELSE}{$ENDIF};
{$ELSE}
  HoldBrushColor := Canvas.Pen.Color;
{$ENDIF}
  try
    with Canvas do begin
{$IFDEF FMX}
      Pen.Kind := TBrushKind.{$IFDEF DELPHI20PLUS}Solid{$ELSE}bkSolid{$ENDIF};
      StrokeThickness := LineSize;
      StrokeJoin := TStrokeJoin.{$IFDEF DELPHI20PLUS}Miter{$ELSE}sjMiter{$ENDIF};
{$ELSE}
      Pen.Width := LineSize;
{$ENDIF}
      while (Cx < (X1+CanvasDiv((X2-X1), 2))) and (Cy < (Y1+CanvasDiv((Y2-Y1), 2))) do begin
        Rect1 := CanvasRect(Cx, Cy, X1+(X2-Cx), Y1+(Y2-Cy));
        CRGB.R := DoubleToByte(R);
        CRGB.G := DoubleToByte(G);
        CRGB.B := DoubleToByte(B);
    {$IFDEF FMX}
        CRGB.A := DoubleToByte(A);
        Pen.Color := CRGB.Color;
        DrawRect(Rect1, 0, 0, [], Opacity);
    {$ELSE}
        Pen.Color := RGBToColor(CRGB);
        Rectangle(Rect1);
    {$ENDIF}

        Cx := Cx + LineSize;
        Cy := Cy + LineSize;
        if (ControlAxis = Y2) then
          Cp := Cy - Y1
        else
          Cp := Cx - X1;
        if (Cp >= CWide) then begin
          CWide := CWide + Wide;
          Inc(i);
          if (i > Low(Colors) + ColorCount) then Dec(i);
          Color1 := Color2;
          Color2 := Colors[i];
    {$IFDEF FMX}
          RGB1.Color := Color1;
          RGB2.Color := Color2;
    {$ELSE}
          RGB1 := ColorToRGB(Color1);
          RGB2 := ColorToRGB(Color2);
    {$ENDIF}
        end;
        R := R + (RGB2.R - RGB1.R) / Wide * LineSize;
        G := G + (RGB2.G - RGB1.G) / Wide * LineSize;
        B := B + (RGB2.B - RGB1.B) / Wide * LineSize;
    {$IFDEF FMX}
        A := A + (RGB2.A - RGB1.A) / Wide * LineSize;
    {$ENDIF}
      end;

    end;
  finally
    { Restore the old settings }
{$IFDEF FMX}
    Canvas.RestoreState(CanvasState);
{$ELSE}
    Canvas.Pen.Color := HoldBrushColor;
    Canvas.Brush.Color := RGBToColor(CRGB);
    Canvas.FillRect(Rect1);
    Canvas.Brush.Color := HoldBrushColor;
{$ENDIF}
  end;
end;

function GetMultilineTextHeight(const Canvas: TCanvas; Text: string): TRSGfxFloat;
var
   i: Integer;
   TextHeight: TRSGfxFloat;
begin
     // returns the text height for a string that includes multiple lines
     TextHeight := Canvas.TextHeight(Text);
     result := TextHeight;
     i := Pos(#10, Text);
     while i > 0 do
     begin
          Text := Copy(Text, i+1, Length(Text));
          result := result + TextHeight;
          i := Pos(#10, Text);
     end;
end;

function WrapCanvasText(const Canvas: TCanvas; Text: string; MaxWidth: TRSGfxFloat): String;
begin
     result := WrapCanvasText(Canvas, Text, NewLineString, [#9, ' ', '-'], MaxWidth);
end;

function GetRotatedTextExtent( const Canvas: TCanvas; Angle: Integer; const Str: String; var Origin: TPoint ): TSize;
var
   W, H, X, Y: Integer;
begin
     Origin := Point(0,0);
     result := Canvas.TextExtent(Str);
     if Angle <> 0 then
     begin
          {adjust height and width as necessary}
          H := result.cy;
          W := result.cx;
          {width (X) of text at new angle}
          X := Round(W * Cos(Angle*Pi/180));
          {height (y) of text at new angle}
          Y := Round(W * Sin(Angle*Pi/180));
          result.cy := Abs(Y) + 2*H;
          result.cx := Abs(X) + 2*H;
          if X < 0 then
             Origin.X := result.cx-H
          else
              Origin.X := H;
          if Y < 0 then
             Origin.Y := H
          else
          begin
               if X < 0 then
                  Origin.Y := result.cy - H
               else
                   Origin.Y := result.cy - H - H div 2;
          end;
      end;
end;

function GetRotatedTextExtent( const Canvas: TCanvas; Angle: Integer; const Str: String; var Origin: TRSPoint ): TRSSize;
var
   W, H, X, Y: Single;
begin
     Origin := RSPoint(0,0);
     result.cx := Canvas.TextWidth(Str);
     result.cy := Canvas.TextHeight(Str);
     if Angle <> 0 then
     begin
          {adjust height and width as necessary}
          H := result.cy;
          W := result.cx;
          {width (X) of text at new angle}
          X := (W * Cos(Angle*Pi/180));
          {height (y) of text at new angle}
          Y := (W * Sin(Angle*Pi/180));
          result.cy := Abs(Y) + 2*H;
          result.cx := Abs(X) + 2*H;
          if X < 0 then
             Origin.X := result.cx-H
          else
              Origin.X := H;
          if Y < 0 then
             Origin.Y := H
          else
          begin
               if X < 0 then
                  Origin.Y := result.cy - H
               else
                   Origin.Y := result.cy - H - H / 2;
          end;
      end;
end;

{$IFDEF NEXTGEN}
function WrapCanvasText( const Canvas: TCanvas; Text: string;
                         const BreakStr: string; BreakChars: Array of Char;
                         MaxWidth: TRSGfxFloat ): String;
var
  QuoteChars: Array of Char;

  function GetWord( const Text: String; Index: Integer ): Integer;
  begin
       result := Index;
       while (result < Length(Text)) and
             (not RSCharInSet(Text[result], BreakChars)) and
             (not RSCharInSet(Text[result], [#10, #13])) do
             if RSCharInSet(Text[result], QuoteChars) then
             begin
                  Inc(result);
                  while (result < Length(Text)) and
                        (not RSCharInSet(Text[result], QuoteChars)) do
                        Inc(Result);

             end
             else
                 Inc(result);
       while (result < Length(Text)) and
             RSCharInSet(Text[result], BreakChars) do
             Inc(result);
       while (result < Length(Text)) and
             RSCharInSet(Text[result], [#10, #13]) do
             Inc(result);
       result := result - Index;
  end;
var
  i: Integer;
  Len: Integer;
  Width: TRSGfxFloat;
  TextWidth: TRSGfxFloat;
  FirstWord: Boolean;
  TheWord: String;
begin
     result := '';
     if Length(Text) = 0 then Exit;

     SetLength(QuoteChars, 2);
     QuoteChars[0] := '''';
     QuoteChars[1] := '"';
//     BreakChars := BreakChars + [#10, #13];
     // loop until end of text
     i := 1;
     Width := 0;
     FirstWord := True;
     while (i < Length(Text)) do
     begin
          Len := GetWord(Text, i);
          TheWord := Copy(Text,i,Len);
          TextWidth := Canvas.TextWidth(TheWord);
          // if line is less than maxwidth, append, else break and then append
          if FirstWord or ((Width + TextWidth) < MaxWidth) then
          begin
               result := result + TheWord;
               Width := Width + TextWidth;
               FirstWord := AnsiEndsText(NewLineString, TheWord);
               if FirstWord then
                  Width := 0;
          end
          else
          begin
               if AnsiEndsText(NewLineString, TheWord) then
               begin
                    FirstWord := True;
                    Width := 0;
                    result := result + TheWord;
               end
               else
               begin
                    Width := TextWidth;
                    result := result + BreakStr + TheWord;
               end;
          end;
          i := i + Len;
     end;
     result := result + Text[i];
end;
{$ELSE}
function WrapCanvasText( const Canvas: TCanvas; Text: string;
                         const BreakStr: string; BreakChars: TSysCharSet;
                         MaxWidth: TRSGfxFloat ): String;
const
  QuoteChars = ['''', '"'];

  function GetWord( const Text: String; Index: Integer ): Integer;
  begin
       result := Index;
       while (result < Length(Text)) and
             (not RSCharInSet(Text[result], BreakChars)) and
             (not RSCharInSet(Text[result], [#10, #13])) do
             if RSCharInSet(Text[result], QuoteChars) then
             begin
                  Inc(result);
                  while (result < Length(Text)) and
                        (not RSCharInSet(Text[result], QuoteChars)) do
                        Inc(Result);

             end
             else
                 Inc(result);
       while (result < Length(Text)) and
             RSCharInSet(Text[result], BreakChars) do
             Inc(result);
       while (result < Length(Text)) and
             RSCharInSet(Text[result], [#10, #13]) do
             Inc(result);
       result := result - Index;
  end;
var
  i: Integer;
  Len: Integer;
  Width: TRSGfxFloat;
  TextWidth: TRSGfxFloat;
  FirstWord: Boolean;
  TheWord: String;
begin
     result := '';
     if Length(Text) = 0 then Exit;

//     BreakChars := BreakChars + [#10, #13];
     // loop until end of text
     i := 1;
     Width := 0;
     FirstWord := True;
     while (i < Length(Text)) do
     begin
          Len := GetWord(Text, i);
          TheWord := Copy(Text,i,Len);
          TextWidth := Canvas.TextWidth(TheWord);
          // if line is less than maxwidth, append, else break and then append
          if FirstWord or ((Width + TextWidth) < MaxWidth) then
          begin
               result := result + TheWord;
               Width := Width + TextWidth;
               FirstWord := AnsiEndsText(NewLineString, TheWord);
               if FirstWord then
                  Width := 0;
          end
          else
          begin
               if AnsiEndsText(NewLineString, TheWord) then
               begin
                    FirstWord := True;
                    Width := 0;
                    result := result + TheWord;
               end
               else
               begin
                    Width := TextWidth;
                    result := result + BreakStr + TheWord;
               end;
          end;
          i := i + Len;
     end;
     result := result + Text[i];
end;
{$ENDIF !NEXTGEN}

function Arc(StartAngle, EndAngle, InnerRadius, OuterRadius,
  ArcOffset: Double): String;
var
  a0: Double;
  a1: Double;
  da: Double;
  df: String;
  c0: Double;
  s0: Double;
  c1: Double;
  s1: Double;
  sInnerRadius: String;
  sOuterRadius: String;
begin
  // create a Arc (donut-like arc) path string

  // Ensure that the outer radius is always larger than the inner radius.
  if OuterRadius < InnerRadius then
  begin
    a0 := OuterRadius;
    OuterRadius := InnerRadius;
    InnerRadius := a0;
  end;
  a0 := DegToRad( StartAngle + ArcOffset );
  a1 := DegToRad( EndAngle + ArcOffset );
  if a1 < a0 then
  begin
    da := a0;
    a0 := a1;
    a1 := da;
  end
  else
    da := a1 - a0;
  if da < PI then
    df := '0'
  else
    df := '1';
  c0 := Cos(a0);
  s0 := Sin(a0);
  c1 := Cos(a1);
  s1 := Sin(a1);

  sInnerRadius := FloatToStr(InnerRadius, USFormatSettings);
  sOuterRadius := FloatToStr(OuterRadius, USFormatSettings);
  if da >= (2*PI-FloatError) then
  begin
    if InnerRadius <> 0 then
      // draw circle with hole
      result :=
        'M0,' + sOuterRadius
        + 'A' + sOuterRadius + ',' + sOuterRadius + ' 0 1,1 0,-' + sOuterRadius
        + 'A' + sOuterRadius + ',' + sOuterRadius + ' 0 1,1 0,' + sOuterRadius
        + 'M0,' + sInnerRadius
        + 'A' + sInnerRadius + ',' + sInnerRadius + ' 0 1,0 0,-' + sInnerRadius
        + 'A' + sInnerRadius + ',' + sInnerRadius + ' 0 1,0 0,' + sInnerRadius
        + 'Z'
    else
      // draw circle with no hole
      result := 'M0,' + sOuterRadius
        + 'A' + sOuterRadius + ',' + sOuterRadius + ' 0 1,1 0,-' + sOuterRadius
        + 'A' + sOuterRadius + ',' + sOuterRadius + ' 0 1,1 0,' + sOuterRadius
        + 'Z';
  end
  else
    if InnerRadius <> 0 then
      // draw donut piece
      result :=
       'M' + FloatToStr(OuterRadius * c0, USFormatSettings) + ',' + FloatToStr(OuterRadius * s0, USFormatSettings)
        + 'A' + sOuterRadius + ',' + sOuterRadius + ' 0 ' + df + ',1 '
        + FloatToStr(OuterRadius * c1, USFormatSettings) + ',' + FloatToStr(OuterRadius * s1, USFormatSettings)
        + 'L' + FloatToStr(InnerRadius * c1, USFormatSettings) + ',' + FloatToStr(InnerRadius * s1, USFormatSettings)
        + 'A' + sInnerRadius + ',' + sInnerRadius + ' 0 ' + df + ',0 '
        + FloatToStr(InnerRadius * c0, USFormatSettings) + ',' + FloatToStr(InnerRadius * s0, USFormatSettings)
        + 'Z'
    else
      // draw pie piece
      result := 'M' + FloatToStr(OuterRadius * c0, USFormatSettings) + ',' + FloatToStr(OuterRadius * s0, USFormatSettings)
        + 'A' + sOuterRadius + ',' + sOuterRadius + ' 0 ' + df + ',1 '
        + FloatToStr(OuterRadius * c1, USFormatSettings) + ',' + FloatToStr(OuterRadius * s1, USFormatSettings)
        + 'L0,0'
        + 'Z';
end;

function Arc(const Path: TRSPathData; StartAngle, EndAngle, InnerRadius,
  OuterRadius, ArcOffset: Double): TRSPathData;
  procedure AddCircleSegment(const r1: Double; const Sweep: Boolean);
  begin
    Path.MoveTo( RSPoint(0,r1) );
    Path.AddArcSvg( RSPoint(0,r1), RSPoint(r1, r1), 0,
      True, Sweep, RSPoint(0,-r1));
    Path.AddArcSvg( RSPoint(0,-r1), RSPoint(r1, r1), 0,
      True, Sweep, RSPoint(0,r1));
  end;
  function ArcSweep(x0, y0, x1, y1: Double): Boolean;
  begin
    //  (x0 - x1) * y0 - (y0 - y1) * x0 > 0 ? 0 : 1;
    result := (((x0 - x1) * y0) - ((y0 - y1) * x0)) <= 0;
  end;

const
  HalfPI = PI/2;
  PadAngle = 0;
var
  a0: Double;
  a1: Double;
  l1: Double;
  l0: Double;
  Sweep: Boolean;
  x0, y0,
  x1, y1: Double;
  x2, y2,
  x3, y3: Double;
//  temp: Double;
//  IsOuterCollasped,
//  IsInnerCollasped: Boolean;
begin
  // create a Arc (donut-like arc) path, modified from arc.js source
  // https://github.com/mbostock/d3/blob/a6cc4cc84e10996ef83385c1b25f4b38129bb427/src/svg/arc.js
  result := Path;
  // Ensure that the outer radius is always larger than the inner radius.
  if OuterRadius < InnerRadius then
  begin
    a0 := OuterRadius;
    OuterRadius := InnerRadius;
    InnerRadius := a0;
  end;

  a0 := DegToRad( StartAngle + ArcOffset );
  a1 := DegToRad( EndAngle + ArcOffset );
  l1 := abs(a1-a0);
  l0 := abs(a0-a1);
  Sweep := a0 <= a1;

  // full circle?
  if l1 >= (2*PI-FloatError) then
  begin
    AddCircleSegment(OuterRadius, Sweep);
    if InnerRadius > 0 then
      AddCircleSegment(InnerRadius, Sweep);
    Path.ClosePath;
    Exit;
  end;

  // compute the outer radius end points
  x0 := OuterRadius*Cos(a0);
  y0 := OuterRadius*Sin(a0);
  x1 := OuterRadius*Cos(a1);
  y1 := OuterRadius*Sin(a1);
  // Detect whether the outer corners are collapsed.
//  IsOuterCollasped := ArcSweep(x0, y0, x1, y1) = (Sweep xor (l1 > PI));
//  if IsOuterCollasped then
//  begin
//    Temp := (a0 + a1) / 2;
//    x0 := OuterRadius*Cos(Temp);
//    y0 := OuterRadius*Sin(Temp);
//  end;

  // compute the inner radius end points
  x2 := InnerRadius*Cos(a1);
  y2 := InnerRadius*Sin(a1);
  x3 := InnerRadius*Cos(a0);
  y3 := InnerRadius*Sin(a0);
  // Detect whether the inner corners are collapsed.
//  IsInnerCollasped := ArcSweep(x2, y2, x3, y3) = ((not Sweep) xor (l0 > PI));
//  if IsInnerCollasped then
//  begin
//    Temp := (a0 + a1) / 2;
//    x2 := InnerRadius*Cos(Temp);
//    y2 := InnerRadius*Sin(Temp);
//  end;

  Path.MoveTo( RSPoint(x0, y0) );
//  if not IsOuterCollasped then
    Path.AddArcSvg( RSPoint(x0, y0), RSPoint(OuterRadius, OuterRadius), 0,
      l1 > PI, Sweep, RSPoint(x1, y1));
  Path.LineTo(RSPoint(x2, y2));
  if (InnerRadius > 0) then //and (not IsInnerCollasped) then
    Path.AddArcSvg( RSPoint(x2, y2), RSPoint(InnerRadius, InnerRadius), 0,
      l0 > PI, not Sweep, RSPoint(x3, y3));
  Path.ClosePath;
end;

procedure DrawBezierPolygon( const Canvas: TCanvas; Vc: TPoints;
                         Smooth: Boolean = False;
                         Closed: Boolean = False;
                         Filled: Boolean = False;
                         PointStep: Integer = 1;
                         Opacity: Single = 1 );
begin
  if Opacity > 0 then
    if Filled then
      Canvas.Polygon(BezierPolygonPoints(VC, Smooth, Closed, PointStep){$IFDEF FMX}, Opacity{$ENDIF})
    else
      Canvas.Polyline(BezierPolygonPoints(VC, Smooth, Closed, PointStep){$IFDEF FMX}, Opacity{$ENDIF});
end;

procedure DrawBezierPolygon( const Canvas: TCanvas; Vc: TRSPoints;
                         Smooth: Boolean = False;
                         Closed: Boolean = False;
                         Filled: Boolean = False;
                         PointStep: Integer = 1;
                         Opacity: Single = 1 );
begin
  if Opacity > 0 then
    if Filled then
       Canvas.Polygon(BezierPolygonPoints(VC, Smooth, Closed, PointStep){$IFDEF FMX}, Opacity{$ENDIF})
    else
       Canvas.Polyline(BezierPolygonPoints(VC, Smooth, Closed, PointStep){$IFDEF FMX}, Opacity{$ENDIF});
end;

function BezierPolygonPoints( Vc: TPoints;
                         Smooth: Boolean = False;
                         Closed: Boolean = False;
                         PointStep: Integer = 1 ): TPoints;
  function Bezier(x:Double;j:byte): Double;
  begin
      result:=0;
      case j of
          0:  begin x:=1-x; result:=x*x*x/6 end;
          1:  result:=x*x*(x/2-1)+4/6;
          2:  result:=x*(1+x*(1-x))/2+1/6;
          3:  result:=x*x*x/6;
         10:  result:=-(1-x)*(1-x)*x/2;
         11:  result:=1+x*x*(-5+3*x)/2;
         12:  result:=x*(1+x*(4-3*x))/2;
         13:  result:=-(1-x)*x*x/2;
      end;
  end;

var
  i,j,k:Integer;
  l,step,sc,x,y:Double;
  Point:TPoint;
  MaxPoint: Integer;
begin
  MaxPoint := Length(VC);
  SetLength(result, 0);
  if MaxPoint<2 then exit;
  if not Closed then
  begin
    SetLength(result, Length(VC)+2);
    result[0] := VC[0];
{$IFDEF CLR}
    System.Array.Copy(VC,result,Length(VC));
//          Move(Ptr^, BufferPtr^, NumBytes);
{$ELSE}
    Move(VC[0], result[1], Length(VC)*SizeOf(VC[0]));
{$ENDIF}
    result[Length(result)-1] := result[Length(result)-2];
    VC := result;
    SetLength(result,0);
    Inc(MaxPoint, 2);
  end;
  if smooth then k:=0 else k:=10; (* Type der Interpolation *)
  (*------- handle start between 1 & 2-------- *)
  if closed then
  begin
      sc:=sqrt(sqr(vc[1].x-vc[0].x)+sqr(vc[1].y-vc[0].y));
      if sc=0 then sc:=1;
      step:=PointStep/sc;
      l:=0;
      repeat
       for j:=0 to 3 do begin
       x:=Bezier(l,k)*vc[MaxPoint-1].x
         +Bezier(l,k+1)*vc[0].x
         +Bezier(l,k+2)*vc[1].x
         +Bezier(l,k+3)*vc[2].x;
       y:=Bezier(l,k)*vc[MaxPoint-1].y
         +Bezier(l,k+1)*vc[0].y
         +Bezier(l,k+2)*vc[1].y
         +Bezier(l,k+3)*vc[2].y;
       end; // end For
       Point.X:=round(X);Point.Y:=round(Y);
       SetLength(result, Length(result)+1);
       result[Length(result)-1] := Point;
       l:=l+step;
      until l>1;
  end;
  (* ----------- handle inside index 2 .. Max-2 -------------- *)
  for i:=1 to MaxPoint-3 do
  begin
     sc:=sqrt(sqr(vc[i+1].x-vc[i].x)+sqr(vc[i+1].y-vc[i].y));
     if sc=0 then sc:=1;
     step:=PointStep/sc;
     l:=0;
     repeat
      x:=Bezier(l,k)  *vc[i-1].x
        +Bezier(l,k+1)*vc[i  ].x
        +Bezier(l,k+2)*vc[i+1].x
        +Bezier(l,k+3)*vc[i+2].x;
      y:=Bezier(l,k)  *vc[i-1].y
        +Bezier(l,k+1)*vc[i  ].y
        +Bezier(l,k+2)*vc[i+1].y
        +Bezier(l,k+3)*vc[i+2].y;
       Point.X:=round(X);Point.Y:=round(Y);
       SetLength(result, Length(result)+1);
       result[Length(result)-1] := Point;
      l:=l+step;
     until l>1;
   end;
   (* ---- handle end of vector max-1 & max ---- *)
   if Closed then
   begin

      sc:=sqrt(sqr(vc[MaxPoint-1].x-vc[MaxPoint-2].x)
              +sqr(vc[MaxPoint-1].y-vc[MaxPoint-2].y));

      if sc=0 then sc:=1;
      step:=PointStep/sc;
      l:=0;

      Repeat
       for j:=0 to 3 do begin
       x:=Bezier(l,k)  *vc[MaxPoint-2].x
         +Bezier(l,k+1)*vc[MaxPoint-1].x
         +Bezier(l,k+2)*vc[MaxPoint-1  ].x
         +Bezier(l,k+3)*vc[2          ].x;
       y:=Bezier(l,k)  *vc[MaxPoint-2].y
         +Bezier(l,k+1)*vc[MaxPoint-1].y
         +Bezier(l,k+2)*vc[MaxPoint-1  ].y
         +Bezier(l,k+3)*vc[          2].y;
       end;
       Point.X:=round(X);Point.Y:=round(Y);
       SetLength(result, Length(result)+1);
       result[Length(result)-1] := Point;
       l:=l+step;
      until l>1;
   end;
end;

function BezierPolygonPoints( Vc: TRSPoints;
                         Smooth: Boolean = False;
                         Closed: Boolean = False;
                         PointStep: Integer = 1 ): TRSPoints;
  function Bezier(x:Double;j:byte): Double;
  begin
      result:=0;
      case j of
          0:  begin x:=1-x; result:=x*x*x/6 end;
          1:  result:=x*x*(x/2-1)+4/6;
          2:  result:=x*(1+x*(1-x))/2+1/6;
          3:  result:=x*x*x/6;
         10:  result:=-(1-x)*(1-x)*x/2;
         11:  result:=1+x*x*(-5+3*x)/2;
         12:  result:=x*(1+x*(4-3*x))/2;
         13:  result:=-(1-x)*x*x/2;
      end;
  end;

var
  i,j,k:Integer;
  l,step,sc,x,y:Double;
  Point:TRSPoint;
  MaxPoint: Integer;
begin
  MaxPoint := Length(VC);
  SetLength(result, 0);
  if MaxPoint<2 then exit;
  if not Closed then
  begin
    SetLength(result, Length(VC)+2);
    result[0] := VC[0];
{$IFDEF CLR}
    System.Array.Copy(VC,result,Length(VC));
//          Move(Ptr^, BufferPtr^, NumBytes);
{$ELSE}
    Move(VC[0], result[1], Length(VC)*SizeOf(VC[0]));
{$ENDIF}
    result[Length(result)-1] := result[Length(result)-2];
    VC := result;
    SetLength(result,0);
    Inc(MaxPoint, 2);
  end;
  if smooth then k:=0 else k:=10; (* Type der Interpolation *)
  (*------- handle start between 1 & 2-------- *)
  if closed then
  begin
      sc:=sqrt(sqr(vc[1].x-vc[0].x)+sqr(vc[1].y-vc[0].y));
      if sc=0 then sc:=1;
      step:=PointStep/sc;
      l:=0;
      repeat
       for j:=0 to 3 do begin
       x:=Bezier(l,k)*vc[MaxPoint-1].x
         +Bezier(l,k+1)*vc[0].x
         +Bezier(l,k+2)*vc[1].x
         +Bezier(l,k+3)*vc[2].x;
       y:=Bezier(l,k)*vc[MaxPoint-1].y
         +Bezier(l,k+1)*vc[0].y
         +Bezier(l,k+2)*vc[1].y
         +Bezier(l,k+3)*vc[2].y;
       end; // end For
       Point.X:=X;Point.Y:=Y;
       SetLength(result, Length(result)+1);
       result[Length(result)-1] := Point;
       l:=l+step;
      until l>1;
  end;
  (* ----------- handle inside index 2 .. Max-2 -------------- *)
  for i:=1 to MaxPoint-3 do
  begin
     sc:=sqrt(sqr(vc[i+1].x-vc[i].x)+sqr(vc[i+1].y-vc[i].y));
     if sc=0 then sc:=1;
     step:=PointStep/sc;
     l:=0;
     repeat
      x:=Bezier(l,k)  *vc[i-1].x
        +Bezier(l,k+1)*vc[i  ].x
        +Bezier(l,k+2)*vc[i+1].x
        +Bezier(l,k+3)*vc[i+2].x;
      y:=Bezier(l,k)  *vc[i-1].y
        +Bezier(l,k+1)*vc[i  ].y
        +Bezier(l,k+2)*vc[i+1].y
        +Bezier(l,k+3)*vc[i+2].y;
       Point.X:=X;Point.Y:=Y;
       SetLength(result, Length(result)+1);
       result[Length(result)-1] := Point;
      l:=l+step;
     until l>1;
   end;
   (* ---- handle end of vector max-1 & max ---- *)
   if Closed then
   begin

      sc:=sqrt(sqr(vc[MaxPoint-1].x-vc[MaxPoint-2].x)
              +sqr(vc[MaxPoint-1].y-vc[MaxPoint-2].y));

      if sc=0 then sc:=1;
      step:=PointStep/sc;
      l:=0;

      Repeat
       for j:=0 to 3 do begin
       x:=Bezier(l,k)  *vc[MaxPoint-2].x
         +Bezier(l,k+1)*vc[MaxPoint-1].x
         +Bezier(l,k+2)*vc[MaxPoint-1  ].x
         +Bezier(l,k+3)*vc[2          ].x;
       y:=Bezier(l,k)  *vc[MaxPoint-2].y
         +Bezier(l,k+1)*vc[MaxPoint-1].y
         +Bezier(l,k+2)*vc[MaxPoint-1  ].y
         +Bezier(l,k+3)*vc[          2].y;
       end;
       Point.X:=X;Point.Y:=Y;
       SetLength(result, Length(result)+1);
       result[Length(result)-1] := Point;
       l:=l+step;
      until l>1;
   end;
end;

procedure DrawArrow(const Canvas:  TCanvas;
                    const Points: TRSPoints;
                    const HeadLength:  TCanvasPixel;
                    Caption: String;
                    Options: TArrowOptions;
                    Width, CaptionSpan: TCanvasPixel;
                    Opacity: Single
                    {$IFDEF FMX}; FontColor: TCanvasColor{$ENDIF}); overload;
var
   i: Integer;
   Ops: TArrowOptions;
begin
     // CaptionSpan marks which line segment should contain the caption, -1 means the
     // middle
     if Length(Points) < 2 then Exit;
     if (CaptionSpan = -1) or (CaptionSpan >= (Length(Points)-1)) then
        CaptionSpan := (Length(Points)-1) div 2;
     for i := 0 to Length(Points) - 2 do
     begin
          Ops := Options;
          if i > 0 then
             Ops := Ops - [ aoArrowAtStart ];
          if i < (Length(Points)-2) then
             Ops := Ops - [ aoArrowAtEnd ];
          if i <> CaptionSpan then
             Ops := Ops - [ aoDrawCaption ];
          DrawArrow( Canvas,
                     CanvasRound(Points[i].X), CanvasRound(Points[i].Y),
                     CanvasRound(Points[i+1].X), CanvasRound(Points[i+1].Y),
                     HeadLength, Caption, Ops, Width, Opacity{$IFDEF FMX}, FontColor{$ENDIF} );
     end;
end;


procedure DrawArrow(const Canvas:  TCanvas;
                    const Points: TPoints;
                    const HeadLength:  TCanvasPixel;
                    Caption: String;
                    Options: TArrowOptions;
                    Width, CaptionSpan: TCanvasPixel;
                    Opacity: Single
                    {$IFDEF FMX}; FontColor: TCanvasColor{$ENDIF}); overload;
var
   i: Integer;
   Ops: TArrowOptions;
begin
     // CaptionSpan marks which line segment should contain the caption, -1 means the
     // middle
     if Length(Points) < 2 then Exit;
     if (CaptionSpan = -1) or (CaptionSpan >= (Length(Points)-1)) then
        CaptionSpan := (Length(Points)-1) div 2;
     for i := 0 to Length(Points) - 2 do
     begin
          Ops := Options;
          if i > 0 then
             Ops := Ops - [ aoArrowAtStart ];
          if i < (Length(Points)-2) then
             Ops := Ops - [ aoArrowAtEnd ];
          if i <> CaptionSpan then
             Ops := Ops - [ aoDrawCaption ];
          DrawArrow( Canvas, Points[i].X, Points[i].Y, Points[i+1].X, Points[i+1].Y,
                     HeadLength, Caption, Ops, Width, Opacity{$IFDEF FMX}, FontColor{$ENDIF} );
     end;
end;

procedure DrawArrow(const Canvas:  TCanvas;
                    const x1,y1, x2,y2:  TCanvasPixel;
                    const HeadLength:  TCanvasPixel;
                    Caption: String;
                    Options: TArrowOptions;
                    Width: TCanvasPixel;
                    Opacity: Single
                    {$IFDEF FMX}; FontColor: TCanvasColor{$ENDIF}); overload;
var
   Triangle        :  Array of TCanvasPoint;
   xbase           :  TCanvasPixel;
   xLineDelta      :  TCanvasPixel;
   xLineUnitDelta  :  Double;
   xNormalDelta    :  TCanvasPixel;
   xNormalUnitDelta:  Double;
   ybase           :  TCanvasPixel;
   yLineDelta      :  TCanvasPixel;
   yLineUnitDelta  :  Double;
   yNormalDelta    :  TCanvasPixel;
   yNormalUnitDelta:  Double;
   i, j: Integer;
{$IFDEF FMX}
   AColor: TCanvasColor;
{$ELSE}
   Brush: TBrush;
{$ENDIF}
begin
  xLineDelta := x2 - x1;
  yLineDelta := y2 - y1;
  if (xLineDelta = 0) and (yLineDelta = 0) then Exit;
  if Caption <> '' then
  begin
{$IFDEF FMX}
    AColor := Canvas.Fill.Color;
    Canvas.Fill.Color := FontColor;
    try
{$ELSE}
    Brush := TBrush.Create;
    Brush.Assign( Canvas.Brush );
    try
       Canvas.Brush.Style := bsClear;
{$ENDIF}
       if aoDrawCaption in Options then
       begin
         i := pos( NewLineString, Caption );
         j := 0;
         while i > 0 do
         begin
              Canvas.TextOut(x1+CanvasDiv(XLineDelta, 2), y1+CanvasDiv(YLineDelta, 2)+j*Canvas.TextHeight('A'), Copy(Caption,1,i-1));
              Caption := Copy(Caption,i+Length(NewLineString),Length(Caption));
              i := pos( NewLineString, Caption );
              inc(j);
         end;
         Canvas.TextOut(x1+CanvasDiv(XLineDelta, 2), y1+CanvasDiv(YLineDelta, 2)+j*Canvas.TextHeight('A'), Caption);
       end;
    finally
{$IFDEF FMX}
       Canvas.Fill.Color := AColor;
{$ELSE}
       Canvas.Brush := Brush;
       Brush.Free;
{$ENDIF}
    end;
  end;

  // Optimize the denominator
  xLineUnitDelta := SQRT( SQR(xLineDelta) + SQR(yLineDelta) );
  yLineUnitDelta := yLineDelta / xLineUnitDelta;
  xLineUnitDelta := xLineDelta / xLineUnitDelta;

  xNormalDelta :=  yLineDelta;
  yNormalDelta := -xLineDelta;
  // Optimize the denominator
  xNormalUnitDelta := SQRT( SQR(xNormalDelta) + SQR(yNormalDelta) );
  yNormalUnitDelta := yNormalDelta / xNormalUnitDelta;
  xNormalUnitDelta := xNormalDelta / xNormalUnitDelta;

  // draw line part
  if Width = 1 then
  begin
    Canvas.DrawLine(RSPoint(X1,Y1), RSPoint(X2, Y2){$IFDEF FMX}, Opacity{$ENDIF});
//       Canvas.MoveTo(x1,y1);
//       Canvas.LineTo(x2,y2);
  end
  else
  begin
       SetLength( Triangle, 5 );
       xBase := x1 + CanvasRound(Width * xLineUnitDelta);
       if aoArrowAtStart in Options then
         yBase := y1 + CanvasRound(Width * yLineUnitDelta)
       else
         yBase := y1;
       Triangle[0] := CanvasPoint( xBase + CanvasRound(Width*xNormalUnitDelta/2),
                             yBase + CanvasRound(Width*yNormalUnitDelta/2));
       Triangle[1] := CanvasPoint( xBase - Round(Width*xNormalUnitDelta/2),
                             yBase - CanvasRound(Width*yNormalUnitDelta/2));
       xBase := x2 - CanvasRound(HeadLength * xLineUnitDelta);
       if aoArrowAtEnd in Options then
         yBase := y2 - CanvasRound(HeadLength * yLineUnitDelta)
       else
         yBase := y2;
       Triangle[2] := CanvasPoint( xBase - CanvasRound(Width*xNormalUnitDelta/2),
                             yBase - CanvasRound(Width*yNormalUnitDelta/2));
       Triangle[3] := CanvasPoint( xBase + CanvasRound(Width*xNormalUnitDelta/2),
                             yBase + CanvasRound(Width*yNormalUnitDelta/2));
       Triangle[4] := Triangle[0];
       if aoFillHead in Options then
          Canvas.Polygon(Triangle)
       else
           Canvas.Polyline(Triangle);
  end;

  if aoClosedHead in Options then
     SetLength( Triangle, 4 )
  else
      SetLength( Triangle, 3 );
  if aoArrowAtStart in Options then
  begin
       // (xBase,yBase) is where arrow line is perpendicular to base of triangle.
       xBase := x1 + CanvasRound(HeadLength * xLineUnitDelta);
       yBase := y1 + CanvasRound(HeadLength * yLineUnitDelta);
       Triangle[0] := CanvasPoint( xBase + CanvasRound(HeadLength*xNormalUnitDelta),
                             yBase + CanvasRound(HeadLength*yNormalUnitDelta));
       Triangle[1] := CanvasPoint( x1, y1 );
       Triangle[2] := CanvasPoint( xBase - CanvasRound(HeadLength*xNormalUnitDelta),
                             yBase - CanvasRound(HeadLength*yNormalUnitDelta));
       if aoClosedHead in Options then
          Triangle[3] := Triangle[0];

       // Draw the arrow tip
       if aoFillHead in Options then
          Canvas.Polygon(Triangle)
       else
           Canvas.Polyline(Triangle);
  end;
  if aoArrowAtEnd in Options then
  begin
       // (xBase,yBase) is where arrow line is perpendicular to base of triangle.
       xBase := x2 - CanvasRound(HeadLength * xLineUnitDelta);
       yBase := y2 - CanvasRound(HeadLength * yLineUnitDelta);

       Triangle[0] := CanvasPoint( xBase + CanvasRound(HeadLength*xNormalUnitDelta),
                             yBase + CanvasRound(HeadLength*yNormalUnitDelta));
       Triangle[1] := CanvasPoint( x2, y2 );
       Triangle[2] := CanvasPoint( xBase - CanvasRound(HeadLength*xNormalUnitDelta),
                             yBase - CanvasRound(HeadLength*yNormalUnitDelta));
       if aoClosedHead in Options then
          Triangle[3] := Triangle[0];

       // Draw the arrow tip
       if aoFillHead in Options then
          Canvas.Polygon(Triangle)
       else
           Canvas.Polyline(Triangle);
  end;
end;

procedure DrawDonutSlice( const Canvas: TCanvas; ARect: TCanvasRect;
  StartDegrees, StopDegrees: Double; Offset: Integer;
  InnerRadius, OuterRadius: Double;
  Opacity: Single );
var
  Path: TRSPathData;
  B: TRSRect;
  M: TRSMatrix;
begin
  ARect := FixRect( ARect );

  Path := TRSPathData.Create;
  try
    Arc( Path, -StartDegrees, -StopDegrees, InnerRadius*100, OuterRadius*100, -Offset );

    // transform arc path into aRect
    B := RSRect(-100,-100,100,100);
    M := CreateTranslateRSMatrix(-B.Left, -B.Top);
    M := RSMatrixMultiply(M, CreateScaleRSMatrix(RSRectWidth(ARect) / RSRectWidth(B), RSRectHeight(ARect) / RSRectHeight(B)));
    M := RSMatrixMultiply(M, CreateTranslateRSMatrix(ARect.Left, ARect.Top));
    Path.ApplyMatrix(M);

    Path.FillPath(Canvas, Opacity);
    Path.DrawPath(Canvas, Opacity);
  finally
    Path.Free;
  end;
end;

procedure DrawPieSlice( const Canvas: TCanvas; ARect: TCanvasRect;
  StartDegrees, StopDegrees: Double; OuterRadius: Double; Offset: Integer;
  Opacity: Single );
var
  bRect: TRSRect;
begin
  // OuterRadius is percentage between 0 and 1+
  ARect := FixRect( ARect );
  if OuterRadius <> 1 then
  begin
    bRect := RSRect(aRect);
    RSInflateRect( bRect, -(1-OuterRadius)/2*RSRectWidth(aRect), -(1-OuterRadius)/2*RSRectHeight(aRect));
    aRect := CanvasRect(bRect);
  end;
  DrawPieSlice( Canvas, aRect, StartDegrees, StopDegrees, Offset, Opacity );
end;

procedure DrawPieSlice( const Canvas: TCanvas; ARect: TCanvasRect;
  StartDegrees, StopDegrees: Double; Offset: Integer;
  Opacity: Single );
var
   Center: TCanvasPoint;
   hdiv2, wdiv2: TCanvasPixel;
{$IFDEF FMX}
  Path: TPathData;
{$ELSE}
   x3, x4, y3, y4: TCanvasPixel;
{$ENDIF}
begin
  ARect := FixRect( ARect );
  Center := CanvasCenterPoint(ARect);
  hdiv2 := CanvasDiv(ARect.Bottom - ARect.Top, 2);
  wdiv2 := CanvasDiv(ARect.Right - ARect.Left, 2);
{$IFDEF FMX}
  Path := TPathData.Create;
  try
    Path.MoveTo(Center);
    Path.AddArc(Center, RSPoint(wdiv2, hdiv2),
          -StartDegrees-Offset, -(StopDegrees-StartDegrees));
    Path.LineTo(Center);
    Canvas.FillPath(Path, Opacity);
    Canvas.DrawPath(Path, Opacity);
  finally
    Path.Free;
  end;
{$ELSE}
  // negative signs on "Y" values to correct for "flip" from normal
  // math defintion for "Y" dimension
  X3 := Center.X + CanvasRound( wdiv2*Cos( DegToRad(Offset+StartDegrees) ) );
  Y3 := Center.Y - CanvasRound( hdiv2*Sin( DegToRad(Offset+StartDegrees) ) );

  X4 := Center.X + CanvasRound( wdiv2*Cos( DegToRad(Offset+StopDegrees) ) );
  Y4 := Center.Y - CanvasRound( hdiv2*Sin( DegToRad(Offset+StopDegrees) ) );

  Canvas.Pie(ARect.Left, ARect.Top, ARect.Right, ARect.Bottom, X3, Y3, X4, Y4);
{$ENDIF}
end;

function DrawShape(const Canvas: TCanvas; const CanvasRect: TCanvasRect;
   const Shape: TRSShapeType; const ShapeOrientation: TRSShapeOrientation;
   const Bevel: Boolean; const BevelSize: TCanvasPixel): TCanvasRect;
{ Purpose: Draws a shape, defined by the bounding rect, on the canvas.        }
{ Arguments:                                                                  }
{          Canvas           - The canvas on which to draw                     }
{          CanvasRect       - The rectangle defining the bounds of the shape  }
{          Shape            - The shape to be drawn                           }
{          ShapeOrientation - The orientation of the shape (doesn't apply to  }
{                             rects or circles)                               }

var
  Points: TCanvasPoints;
  BevelPoints: array [0..2] of TCanvasPoints;
  ARect: TCanvasRect;
{$IFDEF FMX}
  CanvasState: {$IFDEF USE_RSCANVASSAVE}TRSCanvasSaveState{$ELSE}TCanvasSaveState{$ENDIF};
{$ELSE}
  HoldBrushColor: TCanvasColor;
  HoldPenStyle: TPenStyle;
{$ENDIF}
  BevelColorLeft: TCanvasColor;
  BevelColorUp: TCanvasColor;
  BevelColorRight: TCanvasColor;
  BevelColorDown: TCanvasColor;
begin
  { Initialize bevel colors }
  case ShapeOrientation of
    soDown:  begin
               BevelColorLeft := clxDarkGray;
               BevelColorUp := clxDarkGray;
               BevelColorRight := clxLightGray;
               BevelColorDown := clxLightGray;
             end;
    soLeft:  begin
               BevelColorLeft := clxLightGray;
               BevelColorUp := clxDarkGray;
               BevelColorRight := clxDarkGray;
               BevelColorDown := clxLightGray;
             end;
    soRight: begin
               BevelColorLeft := clxDarkGray;
               BevelColorUp := clxLightGray;
               BevelColorRight := clxLightGray;
               BevelColorDown := clxDarkGray;
             end;
  else {soUp}
   BevelColorLeft := clxLightGray;
   BevelColorUp := clxLightGray;
   BevelColorRight := clxDarkGray;
   BevelColorDown := clxDarkGray;
  end;

  { Draw the shape }
  ARect := CanvasRect;
  case Shape of
    sRect:               begin
                           if Bevel then begin
                              Frame3DExt(Canvas, ARect, BevelColorUp, BevelColorDown, BevelSize);
                           end;
                           Canvas.FillRect(ARect);
                         end;
    sCircle:             begin
                           if Bevel then begin
{$IFDEF FMX}
                             CanvasState := Canvas.SaveState{$IFDEF USE_RSCANVASSAVE}(AllParts){$ELSE}{$ENDIF};
                             Canvas.Pen.Kind := TBrushKind.{$IFDEF DELPHI20PLUS}None{$ELSE}bkNone{$ENDIF};
{$ELSE}
                             HoldBrushColor := Canvas.Brush.Color;
                             HoldPenStyle := Canvas.Pen.Style;
                             Canvas.Pen.Style := psClear;
{$ENDIF}
                             Canvas.Brush.Color := BevelColorLeft;
                             Canvas.Chord(ARect.Left, ARect.Top, ARect.Right, ARect.Bottom, ARect.Right, ARect.Top, ARect.Left, ARect.Bottom);
                             Canvas.Brush.Color := BevelColorRight;
                             Canvas.Chord(ARect.Left, ARect.Top, ARect.Right, ARect.Bottom, ARect.Left, ARect.Bottom, ARect.Right, ARect.Top);
                             InflateRect(ARect, -BevelSize, -BevelSize);
{$IFDEF FMX}
                             Canvas.RestoreState(CanvasState);
{$ELSE}
                             Canvas.Pen.Style := HoldPenStyle;
                             Canvas.Brush.Color := HoldBrushColor;
{$ENDIF}
                           end;
                           Canvas.Ellipse(ARect);
                         end;
    sIsoscelesTriangle:  begin
                           { Bevel the triangle first }
                           if Bevel then begin
{$IFDEF FMX}
                             CanvasState := Canvas.SaveState{$IFDEF USE_RSCANVASSAVE}(AllParts){$ELSE}{$ENDIF};
                             Canvas.Pen.Kind := TBrushKind.{$IFDEF DELPHI20PLUS}None{$ELSE}bkNone{$ENDIF};
{$ELSE}
                             HoldBrushColor := Canvas.Brush.Color;
                             HoldPenStyle := Canvas.Pen.Style;
                             Canvas.Pen.Style := psClear;
{$ENDIF}

                             { Initialize three sub triangles (BevelPoints[0][2] will be the center point }
                             case ShapeOrientation of
                               soDown:   begin
                                           SetLength(BevelPoints[0], 3);
                                           BevelPoints[0][0].X := ARect.Left + CanvasDiv((ARect.Right - ARect.Left), 2);
                                           BevelPoints[0][0].Y := ARect.Bottom;
                                           BevelPoints[0][1].X := ARect.Left;
                                           BevelPoints[0][1].Y := ARect.Top;
                                           BevelPoints[0][2].X := BevelPoints[0][0].X;
                                           BevelPoints[0][2].Y := ARect.Top + CanvasDiv((ARect.Bottom - ARect.Top), 2);
                                           SetLength(BevelPoints[1], 3);
                                           BevelPoints[1][0].X := ARect.Left;
                                           BevelPoints[1][0].Y := ARect.Top;
                                           BevelPoints[1][1].X := ARect.Right;
                                           BevelPoints[1][1].Y := ARect.Top;
                                           BevelPoints[1][2].X := BevelPoints[0][2].X;
                                           BevelPoints[1][2].Y := BevelPoints[0][2].Y;
                                           SetLength(BevelPoints[2], 3);
                                           BevelPoints[2][0].X := ARect.Right;
                                           BevelPoints[2][0].Y := ARect.Top;
                                           BevelPoints[2][1].X := BevelPoints[0][0].X;
                                           BevelPoints[2][1].Y := ARect.Bottom;
                                           BevelPoints[2][2].X := BevelPoints[0][2].X;
                                           BevelPoints[2][2].Y := BevelPoints[0][2].Y;

                                           { Draw the three Isosceles triangles }
                                           Canvas.Brush.Color := clxLightGray;
                                           Canvas.Polygon(BevelPoints[0]);
                                           Canvas.Brush.Color := clxLightGray;
                                           Canvas.Polygon(BevelPoints[1]);
                                           Canvas.Brush.Color := clxDarkGray;
                                           Canvas.Polygon(BevelPoints[2]);
                                         end;
                               soLeft:   begin
                                           SetLength(BevelPoints[0], 3);
                                           BevelPoints[0][0].X := ARect.Left;
                                           BevelPoints[0][0].Y := ARect.Top + CanvasDiv((ARect.Bottom - ARect.Top), 2);
                                           BevelPoints[0][1].X := ARect.Right;
                                           BevelPoints[0][1].Y := ARect.Top;
                                           BevelPoints[0][2].X := ARect.Left + CanvasDiv((ARect.Right - ARect.Left), 2);
                                           BevelPoints[0][2].Y := BevelPoints[0][0].Y;
                                           SetLength(BevelPoints[1], 3);
                                           BevelPoints[1][0].X := ARect.Right;
                                           BevelPoints[1][0].Y := ARect.Top;
                                           BevelPoints[1][1].X := ARect.Right;
                                           BevelPoints[1][1].Y := ARect.Bottom;
                                           BevelPoints[1][2].X := BevelPoints[0][2].X;
                                           BevelPoints[1][2].Y := BevelPoints[0][2].Y;
                                           SetLength(BevelPoints[2], 3);
                                           BevelPoints[2][0].X := ARect.Right;
                                           BevelPoints[2][0].Y := ARect.Bottom;
                                           BevelPoints[2][1].X := ARect.Left;
                                           BevelPoints[2][1].Y := BevelPoints[0][0].Y;
                                           BevelPoints[2][2].X := BevelPoints[0][2].X;
                                           BevelPoints[2][2].Y := BevelPoints[0][2].Y;

                                           { Draw the three Isosceles triangles }
                                           Canvas.Brush.Color := clxLightGray;
                                           Canvas.Polygon(BevelPoints[0]);
                                           Canvas.Brush.Color := clxDarkGray;
                                           Canvas.Polygon(BevelPoints[1]);
                                           Canvas.Brush.Color := clxDarkGray;
                                           Canvas.Polygon(BevelPoints[2]);
                                         end;
                               soRight:  begin
                                           SetLength(BevelPoints[0], 3);
                                           BevelPoints[0][0].X := ARect.Right;
                                           BevelPoints[0][0].Y := ARect.Top + CanvasDiv((ARect.Bottom - ARect.Top), 2);
                                           BevelPoints[0][1].X := ARect.Left;
                                           BevelPoints[0][1].Y := ARect.Top;
                                           BevelPoints[0][2].X := ARect.Left + CanvasDiv((ARect.Right - ARect.Left), 2);
                                           BevelPoints[0][2].Y := BevelPoints[0][0].Y;
                                           SetLength(BevelPoints[1], 3);
                                           BevelPoints[1][0].X := ARect.Left;
                                           BevelPoints[1][0].Y := ARect.Bottom;
                                           BevelPoints[1][1].X := ARect.Left;
                                           BevelPoints[1][1].Y := ARect.Top;
                                           BevelPoints[1][2].X := BevelPoints[0][2].X;
                                           BevelPoints[1][2].Y := BevelPoints[0][2].Y;
                                           SetLength(BevelPoints[2], 3);
                                           BevelPoints[2][0].X := ARect.Left;
                                           BevelPoints[2][0].Y := ARect.Bottom;
                                           BevelPoints[2][1].X := ARect.Right;
                                           BevelPoints[2][1].Y := BevelPoints[0][0].Y;
                                           BevelPoints[2][2].X := BevelPoints[0][2].X;
                                           BevelPoints[2][2].Y := BevelPoints[0][2].Y;

                                           { Draw the three Isosceles triangles }
                                           Canvas.Brush.Color := clxLightGray;
                                           Canvas.Polygon(BevelPoints[0]);
                                           Canvas.Brush.Color := clxLightGray;
                                           Canvas.Polygon(BevelPoints[1]);
                                           Canvas.Brush.Color := clxDarkGray;
                                           Canvas.Polygon(BevelPoints[2]);
                                         end;
                             else {soUp}
                               SetLength(BevelPoints[0], 3);
                               BevelPoints[0][0].X := ARect.Left + CanvasDiv((ARect.Right - ARect.Left), 2);
                               BevelPoints[0][0].Y := ARect.Top;
                               BevelPoints[0][1].X := ARect.Left;
                               BevelPoints[0][1].Y := ARect.Bottom;
                               BevelPoints[0][2].X := BevelPoints[0][0].X;
                               BevelPoints[0][2].Y := ARect.Top + CanvasDiv((ARect.Bottom - ARect.Top), 2);
                               SetLength(BevelPoints[1], 3);
                               BevelPoints[1][0].X := ARect.Right;
                               BevelPoints[1][0].Y := ARect.Bottom;
                               BevelPoints[1][1].X := BevelPoints[0][0].X;
                               BevelPoints[1][1].Y := ARect.Top;
                               BevelPoints[1][2].X := BevelPoints[0][2].X;
                               BevelPoints[1][2].Y := BevelPoints[0][2].Y;
                               SetLength(BevelPoints[2], 3);
                               BevelPoints[2][0].X := ARect.Left;
                               BevelPoints[2][0].Y := ARect.Bottom;
                               BevelPoints[2][1].X := ARect.Right;
                               BevelPoints[2][1].Y := ARect.Bottom;
                               BevelPoints[2][2].X := BevelPoints[0][2].X;
                               BevelPoints[2][2].Y := BevelPoints[0][2].Y;

                               { Draw the three Isosceles triangles }
                               Canvas.Brush.Color := clxLightGray;
                               Canvas.Polygon(BevelPoints[0]);
                               Canvas.Brush.Color := clxDarkGray;
                               Canvas.Polygon(BevelPoints[1]);
                               Canvas.Brush.Color := clxDarkGray;
                               Canvas.Polygon(BevelPoints[2]);
                             end;
                             InflateRect(ARect, -BevelSize, -BevelSize);
{$IFDEF FMX}
                             Canvas.RestoreState(CanvasState);
{$ELSE}
                             Canvas.Pen.Style := HoldPenStyle;
                             Canvas.Brush.Color := HoldBrushColor;
{$ENDIF}
                           end;

                           { Initialize the points for the Isosceles triangle }
                           SetLength(Points, 3);
                           case ShapeOrientation of
                             soDown:   begin
                                         Points[0].X := ARect.Left + CanvasDiv((ARect.Right - ARect.Left), 2);
                                         Points[0].Y := ARect.Bottom;
                                         Points[1].X := ARect.Left;
                                         Points[1].Y := ARect.Top;
                                         Points[2].X := ARect.Right;
                                         Points[2].Y := ARect.Top;
                                       end;
                             soLeft:   begin
                                         Points[0].X := ARect.Left;
                                         Points[0].Y := ARect.Top + CanvasDiv((ARect.Bottom - ARect.Top), 2);
                                         Points[1].X := ARect.Right;
                                         Points[1].Y := ARect.Top;
                                         Points[2].X := ARect.Right;
                                         Points[2].Y := ARect.Bottom;
                                       end;
                             soRight:  begin
                                         Points[0].X := ARect.Right;
                                         Points[0].Y := ARect.Top + CanvasDiv((ARect.Bottom - ARect.Top), 2);
                                         Points[1].X := ARect.Left;
                                         Points[1].Y := ARect.Top;
                                         Points[2].X := ARect.Left;
                                         Points[2].Y := ARect.Bottom;
                                       end;
                           else {soUp}
                             Points[0].X := ARect.Left + CanvasDiv((ARect.Right - ARect.Left), 2);
                             Points[0].Y := ARect.Top;
                             Points[1].X := ARect.Left;
                             Points[1].Y := ARect.Bottom;
                             Points[2].X := ARect.Right;
                             Points[2].Y := ARect.Bottom;
                           end;

                           { Draw the Isosceles triangle }
                           Canvas.Polygon(Points);
                         end;
    sRightTriangle:      begin
                           { Bevel the triangle first }
                           if Bevel then begin
{$IFDEF FMX}
                             CanvasState := Canvas.SaveState{$IFDEF USE_RSCANVASSAVE}(AllParts){$ELSE}{$ENDIF};
                             Canvas.Pen.Kind := TBrushKind.{$IFDEF DELPHI20PLUS}None{$ELSE}bkNone{$ENDIF};
{$ELSE}
                             HoldBrushColor := Canvas.Brush.Color;
                             HoldPenStyle := Canvas.Pen.Style;
                             Canvas.Pen.Style := psClear;
{$ENDIF}
                             Canvas.Brush.Color := BevelColorLeft;
                             Canvas.Brush.Color := BevelColorRight;
                             InflateRect(ARect, -BevelSize, -BevelSize);
{$IFDEF FMX}
                             Canvas.RestoreState(CanvasState);
{$ELSE}
                             Canvas.Pen.Style := HoldPenStyle;
                             Canvas.Brush.Color := HoldBrushColor;
{$ENDIF}
                           end;

                           { Initialize the points for the Right triangle }
                           SetLength(Points, 3);
                           case ShapeOrientation of
                             soDown:   begin
                                         Points[0].X := ARect.Left;
                                         Points[0].Y := ARect.Bottom;
                                         Points[1].X := ARect.Left;
                                         Points[1].Y := ARect.Top;
                                         Points[2].X := ARect.Right;
                                         Points[2].Y := ARect.Top;
                                       end;
                             soLeft:   begin
                                         Points[0].X := ARect.Left;
                                         Points[0].Y := ARect.Top;
                                         Points[1].X := ARect.Right;
                                         Points[1].Y := ARect.Top;
                                         Points[2].X := ARect.Right;
                                         Points[2].Y := ARect.Bottom;
                                       end;
                             soRight:  begin
                                         Points[0].X := ARect.Right;
                                         Points[0].Y := ARect.Bottom;
                                         Points[1].X := ARect.Left;
                                         Points[1].Y := ARect.Top;
                                         Points[2].X := ARect.Left;
                                         Points[2].Y := ARect.Bottom;
                                       end;
                           else {soUp}
                             Points[0].X := ARect.Right;
                             Points[0].Y := ARect.Top;
                             Points[1].X := ARect.Left;
                             Points[1].Y := ARect.Bottom;
                             Points[2].X := ARect.Right;
                             Points[2].Y := ARect.Bottom;
                           end;

                           { Draw the Right triangle }
                           Canvas.Polygon(Points);
                         end;
  end;
  result := ARect;
end;

procedure CopyCanvasState( const CanvasTo, CanvasFrom: TCanvas );
begin
    CanvasTo.Brush.Assign(CanvasFrom.Brush);
    CanvasTo.Pen.Assign(CanvasFrom.Pen);
    CanvasTo.Font.Assign(CanvasFrom.Font);
{$IFDEF FMX}
    CanvasTo.SetMatrix(CanvasFrom.Matrix);
    CanvasTo.StrokeThickness := CanvasFrom.StrokeThickness;
  {$IFDEF DELPHI17PLUS}
    CanvasTo.StrokeCap := CanvasFrom.Stroke.Cap;
    CanvasTo.StrokeJoin := CanvasFrom.Stroke.Join;
    CanvasTo.StrokeDash := CanvasFrom.Stroke.Dash;
  {$ELSE}
    CanvasTo.StrokeCap := CanvasFrom.StrokeCap;
    CanvasTo.StrokeJoin := CanvasFrom.StrokeJoin;
    CanvasTo.StrokeDash := CanvasFrom.StrokeDash;
  {$ENDIF}
{$ENDIF}
end;

function DrawShape( const Canvas: TCanvas; ARect: TCanvasRect;
                     Image: TBitmap;
                     Style: TRSShapeStyle;
                     StretchImage: Boolean;
                     BevelHighlight: TCanvasColor;
                     BevelShadow: TCanvasColor;
                     BevelSize: TCanvasPixel;
                     Opacity: Single ): TCanvasRect;
var
   ptCenter: TCanvasPoint;
   ptInCenter: TCanvasPoint;
   H, W, S: TCanvasPixel;
   ACanvas: TCanvas;
{$IFDEF FMX}
   CanvasState: {$IFDEF USE_RSCANVASSAVE}TRSCanvasSaveState{$ELSE}TCanvasSaveState{$ENDIF};
   OrigRect: TCanvasRect;
   Bitmap: TBitmap;
   ShapeOpacity: Single;
{$ELSE}
   Color: TCanvasColor;
{$ENDIF}
   PtTop, PtLeft, ptRight, ptBottom: TCanvasPoint;
   PtHalfTopLeft: TCanvasPoint;
   PtTopLeft, PtBottomLeft, ptTopRight, ptBottomRight: TCanvasPoint;
begin
  ACanvas := Canvas;
  ARect := FixRect(ARect);
{$IFDEF FMX}
  OrigRect := ARect;
  if (Opacity < 1) and (Style in ShapeStyles3D) then
  begin
    Bitmap := TBitmap.Create(Round(ARect.Right), Round(ARect.Bottom));
    Bitmap.Canvas.BeginScene;
    ACanvas := Bitmap.Canvas;
    CopyCanvasState(ACanvas, Canvas);
    ShapeOpacity := 1;
  end
  else
  begin
    Bitmap := nil;
    ShapeOpacity := Opacity;
  end;
  try
{$ENDIF}
     case Style of
          ssRectangle,
          ss3DRectangle:
          begin
               if Style = ss3DRectangle then
                  Frame3DExt( ACanvas, ARect, BevelHighlight, BevelShadow, BevelSize);
               ACanvas.Rectangle(ARect{$IFDEF FMX}, ShapeOpacity{$ENDIF});
          end;
          ssEllipse,
          ss3DEllipse:
          begin
               if Style = ss3DEllipse then
                  Frame3DEllipse(ACanvas, ARect, BevelHighlight, BevelShadow, BevelSize );
               ACanvas.Ellipse(ARect{$IFDEF FMX}, ShapeOpacity{$ENDIF});
          end;
          ssImage:
          begin
               if not StretchImage then
                  ACanvas.Draw(ARect.Left, ARect.Top, Image{$IFDEF FMX}, ShapeOpacity{$ENDIF})
               else
                   ACanvas.StretchDraw(ARect, Image{$IFDEF FMX}, ShapeOpacity{$ENDIF} );
          end;
          ssDiamond,
          ssLeftTriangle,
          ssRightTriangle,
          ssUpTriangle,
          ssDownTriangle,
          ss3DDiamond,
          ss3DLeftTriangle,
          ss3DRightTriangle,
          ss3DUpTriangle,
          ss3DDownTriangle:
          begin
               ptCenter := CanvasCenterPoint( ARect );
               ptLeft := CanvasPoint( ARect.Left, ptCenter.Y );
               ptTop := CanvasPoint( ptCenter.X, ARect.Top );
               ptRight := CanvasPoint( ARect.Right, ptCenter.Y );
               ptBottom := CanvasPoint( ptCenter.X, ARect.Bottom );
               ptTopLeft := ARect.TopLeft;
               ptTopRight := CanvasPoint( ARect.Right, ARect.Top );
               ptBottomLeft := CanvasPoint( ARect.Left, ARect.Bottom );
               ptBottomRight := ARect.BottomRight;
               ptHalfTopLeft := CanvasPoint( ARect.Left + CanvasDiv((ARect.Right-ARect.Left), 4),
                                       ARect.Top + CanvasDiv((ARect.Bottom-ARect.Top), 4) );
//               ptHalfTopRight := Point( ARect.Left + Round((ARect.Right-ARect.Left) * 0.75),
//                                        ARect.Top + ((ARect.Bottom-ARect.Top) div 4) );
//               ptHalfBottomLeft := Point( ARect.Left + ((ARect.Right-ARect.Left) div 4),
//                                          ARect.Top + Round((ARect.Bottom-ARect.Top) * 0.75) );
//               ptHalfBottomRight := Point( ARect.Left + Round((ARect.Right-ARect.Left) * 0.75),
//                                           ARect.Top + Round((ARect.Bottom-ARect.Top) * 0.75) );
               // do we need to bevel
               if Style in ShapeStyles3D then
               begin
{$IFDEF FMX}
                    CanvasState := Canvas.SaveState{$IFDEF USE_RSCANVASSAVE}(AllParts){$ELSE}{$ENDIF};
{$ELSE}
                    Color := ACanvas.Brush.Color;
{$ENDIF}
                    try
                       // draw the highlight
                       {$IFDEF FMX}
                       ACanvas.Pen.Kind := TBrushKind.{$IFDEF DELPHI20PLUS}None{$ELSE}bkNone{$ENDIF};
                       {$ELSE}
                       ACanvas.Pen.Style := psClear;
                       {$ENDIF}
                       ACanvas.Brush.Color := BevelHighlight;
                       case Style of
                         ss3DDiamond:        ACanvas.Polygon( [ ptLeft, ptTop, ptBottom]{$IFDEF FMX}, ShapeOpacity{$ENDIF} );
                         ss3DLeftTriangle:   ACanvas.Polygon( [ ptLeft, ptTopRight, ptBottomRight]{$IFDEF FMX}, ShapeOpacity{$ENDIF} );
                         ss3DRightTriangle:  ACanvas.Polygon( [ ptRight, ptTopLeft, ptBottomLeft]{$IFDEF FMX}, ShapeOpacity{$ENDIF} );
                         ss3DUpTriangle:     ACanvas.Polygon( [ ptBottomLeft, ptTop, ptBottomRight]{$IFDEF FMX}, ShapeOpacity{$ENDIF} );
                         ss3DDownTriangle:   ACanvas.Polygon( [ ptTopLeft, ptBottom, ptTopRight]{$IFDEF FMX}, ShapeOpacity{$ENDIF} );
                       end;
                       // draw the shadow
                       ACanvas.Brush.Color := BevelShadow;
                       case Style of
                         ss3DDiamond:
                         begin
                              ACanvas.Polygon( [ ptRight, ptTop, ptBottom]{$IFDEF FMX}, ShapeOpacity{$ENDIF} );
                              OffsetPoint( ptLeft, BevelSize, 0 );
                              OffsetPoint( ptRight, -BevelSize, 0 );
                              OffsetPoint( ptTop, 0, BevelSize );
                              OffsetPoint( ptBottom, 0, -BevelSize );
                         end;
                         ss3DLeftTriangle:
                         begin
                              ptInCenter := InCenterPoint( ptLeft, ptTopRight, ptBottomRight );
                              ACanvas.Polygon( [ ptInCenter, ptTopRight, ptBottomRight ]{$IFDEF FMX}, ShapeOpacity{$ENDIF} );
                              OffsetPoint( ptLeft, ptInCenter, BevelSize );
                              OffsetPoint( ptTopRight, ptInCenter, BevelSize );
                              OffsetPoint( ptBottomRight, ptInCenter, BevelSize );
                         end;
                         ss3DRightTriangle:
                         begin
                              ptInCenter := InCenterPoint( ptRight, ptTopLeft, ptBottomLeft );
                              ACanvas.Polygon( [ ptInCenter, ptTopLeft, ptRight ]{$IFDEF FMX}, ShapeOpacity{$ENDIF} );
                              ACanvas.Polygon( [ ptInCenter, ptRight, ptBottomLeft ]{$IFDEF FMX}, ShapeOpacity{$ENDIF} );
                              OffsetPoint( ptRight, ptInCenter, BevelSize );
                              OffsetPoint( ptTopLeft, ptInCenter, BevelSize );
                              OffsetPoint( ptBottomLeft, ptInCenter, BevelSize );
                         end;
                         ss3DUpTriangle:
                         begin
                              ptInCenter := InCenterPoint( ptBottomLeft, ptTop, ptBottomRight );
                              ACanvas.Polygon( [ ptTop, ptInCenter, ptBottomRight]{$IFDEF FMX}, ShapeOpacity{$ENDIF} );
                              ACanvas.Polygon( [ ptBottomLeft, ptInCenter, ptBottomRight]{$IFDEF FMX}, ShapeOpacity{$ENDIF} );
                              OffsetPoint( ptBottomRight, ptInCenter, BevelSize );
                              OffsetPoint( ptTop, ptInCenter, BevelSize );
                              OffsetPoint( ptBottomLeft, ptInCenter, BevelSize );
                         end;
                         ss3DDownTriangle:
                         begin
                              ptInCenter := InCenterPoint( ptTopLeft, ptBottom, ptTopRight );
                              ACanvas.Polygon( [ ptBottom, ptInCenter, ptTopRight ]{$IFDEF FMX}, ShapeOpacity{$ENDIF} );
                              OffsetPoint( ptTopRight, ptInCenter, BevelSize );
                              OffsetPoint( ptBottom, ptInCenter, BevelSize );
                              OffsetPoint( ptTopLeft, ptInCenter, BevelSize );
                         end;
                       end;
                    finally
{$IFDEF FMX}
                       ACanvas.RestoreState(CanvasState);
{$ELSE}
                       ACanvas.Brush.Color := Color;
{$ENDIF}
                    end;
               end;
               // draw the shape
               // note the starting point in the "middle" of a segment ensures
               // the corners are smoothed in FireMonkey
               case Style of
                 ssDiamond, ss3DDiamond:
                   ACanvas.Polygon([ PtLeft,
                                    ptTop,
                                    ptRight,
                                    ptBottom ]{$IFDEF FMX}, ShapeOpacity{$ENDIF});
                 ssLeftTriangle, ss3DLeftTriangle:
                   ACanvas.Polygon([ PtLeft,
                                    ptTopRight,
                                    ptBottomRight ]{$IFDEF FMX}, ShapeOpacity{$ENDIF});
                 ssRightTriangle, ss3DRightTriangle:
                   ACanvas.Polygon([ ptTopLeft,
                                    ptRight,
                                    ptBottomLeft ]{$IFDEF FMX}, ShapeOpacity{$ENDIF});
                 ssUpTriangle, ss3DUpTriangle:
                   ACanvas.Polygon([ PtBottomLeft,
                                    ptTop,
                                    ptBottomRight ]{$IFDEF FMX}, ShapeOpacity{$ENDIF});
                 ssDownTriangle, ss3DDownTriangle:
                   ACanvas.Polygon([ PtTopLeft,
                                    ptTopRight,
                                    ptBottom ]{$IFDEF FMX}, ShapeOpacity{$ENDIF});
//                 ssDiamond, ss3DDiamond:
//                   ACanvas.Polygon([ PtHalfTopLeft,
//                                    PtLeft,
//                                    ptTop,
//                                    ptRight,
//                                    ptBottom,
//                                    PtLeft ]{$IFDEF FMX}, ShapeOpacity{$ENDIF});
//                 ssLeftTriangle, ss3DLeftTriangle:
//                   ACanvas.Polygon([ PtRight,
//                                    ptBottomRight,
//                                    PtLeft,
//                                    PtTopRight ]{$IFDEF FMX}, ShapeOpacity{$ENDIF});
//                 ssRightTriangle, ss3DRightTriangle:
//                   ACanvas.Polygon([ PtLeft,
//                                    ptTopLeft,
//                                    ptRight,
//                                    ptBottomLeft ]{$IFDEF FMX}, ShapeOpacity{$ENDIF});
//                 ssUpTriangle, ss3DUpTriangle:
//                   ACanvas.Polygon([ PtBottom,
//                                    PtBottomLeft,
//                                    ptTop,
//                                    ptBottomRight ]{$IFDEF FMX}, ShapeOpacity{$ENDIF});
//                 ssDownTriangle, ss3DDownTriangle:
//                   ACanvas.Polygon([ PtTop,
//                                    PtTopRight,
//                                    ptBottom,
//                                    ptTopLeft ]{$IFDEF FMX}, ShapeOpacity{$ENDIF});
               end;
          end;
          ssSquare,
          ssCircle,
          ssRoundRect,
          ssRoundSquare,
          ss3DSquare,
          ss3DCircle,
          ss3DRoundRect,
          ss3DRoundSquare:
          begin
               w := ARect.Right - ARect.Left;
               h := ARect.Bottom - ARect.Top;
               if W < H then S := W else S := H;
               if Style in [ssSquare, ssRoundSquare, ssCircle, ss3DRoundSquare, ss3DSquare, ss3DCircle] then
               begin
                    ARect.Left := ARect.Left + CanvasDiv((W-S), 2);
                    ARect.Top := ARect.Top + CanvasDiv((H-S), 2);
                    ARect.Right := ARect.Left + S;
                    ARect.Bottom := ARect.Top + S;
               end;
               case Style of
                    ssSquare,
                    ss3DSquare:
                    begin
                         if Style = ss3DSquare then
                            Frame3DExt( ACanvas, ARect, BevelHighlight, BevelShadow, BevelSize);
                         ACanvas.Rectangle(ARect{$IFDEF FMX}, ShapeOpacity{$ENDIF});
                    end;
                    ssRoundRect,
                    ssRoundSquare,
                    ss3DRoundRect,
                    ss3DRoundSquare:
                    begin
                         if Style in [ss3DRoundRect, ss3DRoundSquare] then
                            Frame3DExt( ACanvas, ARect, BevelHighlight,
                              BevelShadow, BevelSize, True,
                              CanvasDiv(S, 4), CanvasDiv(S, 4) );
                         ACanvas.RoundRect( ARect.Left, Arect.Top,
                                           ARect.Right, ARect.Bottom,
                                           CanvasDiv(S, 4), CanvasDiv(S, 4){$IFDEF FMX}, ShapeOpacity{$ENDIF} );
                    end;
                    ssCircle,
                    ss3DCircle:
                    begin
                         if Style = ss3DCircle then
                            Frame3DEllipse(ACanvas, ARect, BevelHighlight,
                              BevelShadow, BevelSize);
                         ACanvas.Ellipse(ARect{$IFDEF FMX}, ShapeOpacity{$ENDIF});
                    end;
               end;
          end;
     end;
     result := ARect;
{$IFDEF FMX}
  finally
    if Bitmap <> nil then
    begin
      Bitmap.Canvas.EndScene;
      {$IFNDEF DELPHI17PLUS}
      Bitmap.BitmapChanged;
      {$ENDIF}
      Canvas.DrawBitmap( Bitmap,
        OrigRect,
        OrigRect, Opacity );
      Bitmap.Free;
    end;
  end;
{$ENDIF}
end;

procedure Frame3DExt(Canvas: TCanvas; var Rect: TCanvasRect;
  TopColor, BottomColor: TCanvasColor; Width: TCanvasPixel; Rounded: Boolean; X, Y: TCanvasPixel );

  procedure DoRect;
  var
    TopRight, BottomLeft: TCanvasPoint;
  begin
    with Canvas, Rect do
    begin
      TopRight.X := Right;
      TopRight.Y := Top;
      BottomLeft.X := Left;
      BottomLeft.Y := Bottom;
      Pen.Color := TopColor;
      PolyLine([BottomLeft, TopLeft, TopRight]);
      Pen.Color := BottomColor;
      BottomLeft.X := BottomLeft.X - 1;
      PolyLine([TopRight, BottomRight, BottomLeft]);
    end;
  end;

  procedure DoRoundRect;
  var
{$IFDEF FMX}
    CanvasState: {$IFDEF USE_RSCANVASSAVE}TRSCanvasSaveState{$ELSE}TCanvasSaveState{$ENDIF};
{$ELSE}
    Color: TCanvasColor;
{$ENDIF}
  begin
{$IFDEF FMX}
       CanvasState := Canvas.SaveState{$IFDEF USE_RSCANVASSAVE}(AllParts){$ELSE}{$ENDIF};
{$ELSE}
       Color := Canvas.Brush.Color;
{$ENDIF}
       try
          Canvas.Brush.Color := TopColor;
          // Draw Round Rect in highlight color
          Canvas.RoundRect(Rect.Left, Rect.Top, Rect.Right, Rect.Bottom, X, Y);
          Canvas.Brush.Color := BottomColor;
{$IFDEF FMX}
          Canvas.Pen.Kind := TBrushKind.{$IFDEF DELPHI20PLUS}None{$ELSE}bkNone{$ENDIF};
{$ELSE}
          Canvas.Pen.Style := psClear;
{$ENDIF}
          // Draw bisected ellipse in top right
          Canvas.Chord(Rect.Right-X, Rect.Top, Rect.Right, Rect.Top+Y, Rect.Right-X, Rect.Top+Y, Rect.Right, Rect.Top );
          // Draw bisected ellipse in lower left
          Canvas.Chord(Rect.Left, Rect.Bottom-Y, Rect.Left+X, Rect.Bottom, Rect.Left, Rect.Bottom, Rect.Left+X, Rect.Bottom-Y );
          // Draw ellipse in bottom right
          Canvas.Ellipse(Rect.Right-X, Rect.Bottom-Y, Rect.Right, Rect.Bottom);
          // Now fill in the rest of the shadow
          Canvas.Rectangle( CanvasDiv(Rect.Right-X, 2), CanvasDiv(Rect.Top+Y, 2), Rect.Right, CanvasDiv(Rect.Bottom-Y, 2));
          Canvas.Rectangle( CanvasDiv(Rect.Left+X, 2), Rect.Bottom, CanvasDiv(Rect.Right-X, 2), CanvasDiv(Rect.Bottom-Y, 2));
       finally
{$IFDEF FMX}
          Canvas.RestoreState(CanvasState);
{$ELSE}
          Canvas.Brush.Color := Color;
{$ENDIF}
       end;
  end;
begin
  Canvas.PenWidth := 1;
  if Rounded then
  begin
       DoRoundRect;
       InflateRect(Rect, -Width, -Width);
       Exit;
  end;
  Rect.Bottom := Rect.Bottom - 1;
  Rect.Right := Rect.Right - 1;
  while Width > 0 do
  begin
    Width := Width - 1;
    DoRect;
    InflateRect(Rect, -1, -1);
  end;
  Rect.Bottom := Rect.Bottom + 1;
  Rect.Right := Rect.Right + 1;
end;

procedure Frame3DEllipse(Canvas: TCanvas; var Rect: TCanvasRect;
  TopColor, BottomColor: TCanvasColor; Width: TCanvasPixel);
var
{$IFDEF FMX}
   CanvasState: {$IFDEF USE_RSCANVASSAVE}TRSCanvasSaveState{$ELSE}TCanvasSaveState{$ENDIF};
{$ELSE}
   Color: TCanvasColor;
{$ENDIF}
begin
{$IFDEF FMX}
     CanvasState := Canvas.SaveState{$IFDEF USE_RSCANVASSAVE}(AllParts){$ELSE}{$ENDIF};
     Canvas.Pen.Kind := TBrushKind.{$IFDEF DELPHI20PLUS}None{$ELSE}bkNone{$ENDIF};
{$ELSE}
     Color := Canvas.Brush.Color;
     Canvas.Pen.Style := psClear;
{$ENDIF}
     try
        Canvas.Brush.Color := TopColor;
        Canvas.Chord( Rect.Left, Rect.Top, Rect.Right, Rect.Bottom,
                      Rect.Right, Rect.Top, Rect.Left, Rect.Bottom );
        Canvas.Brush.Color := BottomColor;
        Canvas.Chord( Rect.Left, Rect.Top, Rect.Right, Rect.Bottom,
                      Rect.Left, Rect.Bottom, Rect.Right, Rect.Top );
        InflateRect(Rect, -Width, -Width);
     finally
{$IFDEF FMX}
        Canvas.RestoreState(CanvasState);
{$ELSE}
        Canvas.Brush.Color := Color;
{$ENDIF}
     end;
end;

function IndexToColor(const Index: Integer; const Count: Integer; const Colors: TColors): TCanvasColor;
{ Purpose: Converts a position in a range of numbers to a color      }
{ Arguments:                                                         }
{          Index  - The index into the range of numbers (zero based) }
{          Count  - Range of numbers                                 }
{          Colors - An array of ordered colors                       }
var
{$IFDEF FMX}
  RGB1: TAlphaColorRec;
  RGB2: TAlphaColorRec;
  CRGB: TAlphaColorRec;
  AD: Double;
  AlphaS: Double;
{$ELSE}
  RGB1: TRGB;
  RGB2: TRGB;
  CRGB: TRGB;
{$ENDIF}
  RD,GD,BD: Double;
  RS,GS,BS: Double;
  ColorCount: Integer;
  SubDivPerColorRange: Double;
  IndexColorRange: Integer;
  LowColor: TCanvasColor;
  HighColor: TCanvasColor;
  RelativeIndex: Double;
begin
  { Figure out how many colors we're dealing with }
  ColorCount := Length(Colors);
  if (ColorCount <= 0) then begin
    Result := clxBlack;
    Exit;
  end;

  if (ColorCount > 1) then begin
    { Figure out what colors surround the index }
    SubDivPerColorRange := Count / (ColorCount-1);
    if (SubDivPerColorRange <= 0) then
      SubDivPerColorRange := Count;
    IndexColorRange := Index div Max(1,Round(SubDivPerColorRange));
    LowColor := Colors[IndexColorRange];
    HighColor := Colors[IndexColorRange+1];
    { Calculate where in the ColorRange the Index is. Similar to mod, but with Doubles }
    RelativeIndex := Index-(SubDivPerColorRange*IndexColorRange)+1;

    { We can't lose any precision, or color will be wrong }
{$IFDEF FMX}
    RGB1.Color := LowColor;
    RGB2.Color := HighColor;
    AD := (RGB1.A - RGB2.A) / SubDivPerColorRange;
{$ELSE}
    RGB1 := ColorToRGB(LowColor);
    RGB2 := ColorToRGB(HighColor);
{$ENDIF}
    RD := (RGB1.R - RGB2.R) / SubDivPerColorRange;
    GD := (RGB1.G - RGB2.G) / SubDivPerColorRange;
    BD := (RGB1.B - RGB2.B) / SubDivPerColorRange;
    RS := RGB1.R - (RD * RelativeIndex);
    GS := RGB1.G - (GD * RelativeIndex);
    BS := RGB1.B - (BD * RelativeIndex);
    CRGB.R := DoubleToByte(RS);
    CRGB.G := DoubleToByte(GS);
    CRGB.B := DoubleToByte(BS);
{$IFDEF FMX}
    AlphaS := RGB1.A - (AD * RelativeIndex);
    CRGB.A := DoubleToByte(AlphaS);
    Result := CRGB.Color;
{$ELSE}
    Result := RGB(DoubleToByte(RS),DoubleToByte(GS),DoubleToByte(BS));
{$ENDIF}
  end
  else begin
    { Just return the low color value }
    Result := Colors[0];
  end;
end;

procedure ProgressCircle(const Canvas: TCanvas; const CanvasR: TCanvasRect;
   const OuterPadding: Integer; const Progress: Integer;
   const Min, Max: Integer; const Colors: TColors; const AdjustPenColor: Boolean;
   const SubDivisions: Integer; const SubShape: TRSShapeType;
   const SubShapeOrientation: TRSShapeOrientation;
   const SubShapeBevel: Boolean; const SubShapeBevelSize: Integer;
   const Spacing: Integer;
   const Direction: TProgressDirection; const DisplayType: TProgressDisplayType);
{ Purpose: Similar to a progress bar, but in a circular orientation. Shapes are }
{          drawn on the perimeter of the circle up to the current progress      }
{          position. }
{ Arguments:                                                                    }
{          Canvas        - The canvas on which to draw                          }
{          CanvasR    - The rectangle defining the bounds of the Canvas      }
{          OuterPadding  - The number of pixels not to be used on the outer     }
{                          edges of the CanvasR                              }
{          Progress      - The current progress, not in percent, but relative   }
{                          to Max                                               }
{          Min           - The minimum scale value                              }
{          Max           - The maximum scale value                              }
{          Colors        - An array of ordered colors                           }
{          AdjustPenColor- Indicates whether the pen should be used, or left up }
{                          to the caller                                        }
{          SubDivisions  - The number of subdivisions in the scale (<= Max)     }
{          SubShape      - The type of shape to be drawn on the perimeter       }
{          SubShapeOrientation -                                                }
{                          The orientation of the SubShape (doen't apply to     }
{                          some shapes                                          }
{          SubShapeBevel - Indicates whether the subshapes are beveled          }
{          SubShapeBevelSize - Specifies the width in pixels of the subshape    }
{                          bevel                                                }
{          Spacing       - The space in pixels between SubDivisions             }
{          Direction     - Indicates whether the progress meter draws forward   }
{                          or backward                                          }
{          DisplayType   - Determines whether the shape representing only the   }
{                          current progress, or all of the shapes up to the     }
{                          current progress is drawn                            }
var
  i, Endi: Integer;
  ShapeW, ShapeH,
  ShapeOx, ShapeOy, { Drawing offset for the shape }
  AngleFactor: TCanvasPixel;
  w, h: TCanvasPixel;
  CRLeft, CRTop: TCanvasPixel;
  HoldBrushColor: TCanvasColor;
  HoldPenColor: TCanvasColor;
  ColorCount: Integer;
  MinColor: TCanvasColor;
begin
  { Figure out how many colors we're dealing with }
  ColorCount := Length(Colors);
  if (ColorCount <= 0) then Exit;
  MinColor := Colors[Low(Colors)];

  HoldBrushColor := Canvas.Brush.Color;
  HoldPenColor := Canvas.Pen.Color;
  CRLeft := CanvasMin([CanvasR.Left, CanvasR.Right]);
  CRTop := CanvasMin([CanvasR.Top, CanvasR.Bottom]);
  w := abs(CanvasR.Right - CanvasR.Left);
  h := abs(CanvasR.Bottom - CanvasR.Top);
  {TODO: The calculations for ShapeW & ShapeH are not optimal.  Below is close, but there is overlap}
  { Take the circumference, and divide by the subdivisions to determine the width and height of the shapes}
  ShapeW := CanvasRound(((w-(OuterPadding*2))*Pi) / SubDivisions) - Spacing;
  ShapeH := CanvasRound(((h-(OuterPadding*2))*Pi) / SubDivisions) - Spacing;
  if (Direction = pdBackward) then
    AngleFactor := -1
  else {pdForward}
    AngleFactor := 1;

  { Determine how far to draw }
  if (Min = Max) or (Progress = Max) then
    Endi := Subdivisions
  else
    EndI := Trunc((Progress / ((Max-Min)/Subdivisions)));

  { Draw the progress }
  Canvas.Brush.Color := MinColor;
  if AdjustPenColor then
    Canvas.Pen.Color := MinColor;
  for i:=1 to Endi do begin
    if (DisplayType = dtAllProgress) or (i = Endi) then begin
      { Calculate the position of the next subshape }
      ShapeOx := CanvasTrunc(((w-ShapeW-(OuterPadding*2))/2 + Cos(DegToRad(AngleFactor*((360*i)/Subdivisions)))*(w-ShapeW-(OuterPadding*2))/2))+OuterPadding;
      ShapeOy := CanvasTrunc(((h-ShapeH-(OuterPadding*2))/2 - Sin(DegToRad(AngleFactor*((360*i)/Subdivisions)))*(h-ShapeH-(OuterPadding*2))/2))+OuterPadding;

      { Fill the subdivision }
      DrawShape(Canvas, CanvasRect(CRLeft + ShapeOx, CRTop + ShapeOy, CRLeft + ShapeOx+ShapeW, CRTop + ShapeOy+ShapeH),
                SubShape, SubShapeOrientation, SubShapeBevel, SubShapeBevelSize);
    end;

    { Update the color }
    Canvas.Brush.Color := IndexToColor(i, SubDivisions, Colors);
    if AdjustPenColor then
      Canvas.Pen.Color := Canvas.Brush.Color;
  end;
  Canvas.Brush.Color := HoldBrushColor;
  Canvas.Pen.Color := HoldPenColor;
end;

procedure ProgressPie(const Canvas: TCanvas; const CanvasR: TCanvasRect;
   const OuterPadding: Integer; const Progress: Integer;
   const Min, Max: Integer; const Colors: TColors; const AdjustPenColor: Boolean;
   const SubDivisions: Integer; const Spacing: Integer;
   const Direction: TProgressDirection; const DisplayType: TProgressDisplayType);
{ Purpose: Similar to a progress bar, but in a Pie shape. Pie slices are drawn  }
{          up to the current progress position.                                 }
{ Arguments:                                                                    }
{          Canvas        - The canvas on which to draw                          }
{          CanvasR    - The rectangle defining the bounds of the Canvas      }
{          OuterPadding  - The number of pixels not to be used on the outer     }
{                          edges of the CanvasR                              }
{          Progress      - The current progress, not in percent, but relative   }
{                          to Max                                               }
{          Min           - The minimum scale value                              }
{          Max           - The maximum scale value                              }
{          Colors        - An array of ordered colors                           }
{          AdjustPenColor- Indicates whether the pen should be used, or left up }
{                          to the caller                                        }
{          SubDivisions  - The number of subdivisions in the scale (<= Max)     }
{          Spacing       - The space in DEGREES (not pixels anymore) between SubDivisions             }
{          Direction     - Indicates whether the progress meter draws forward   }
{                          or backward                                          }
{          DisplayType   - Determines whether the slice representing only the   }
{                          current progress, or all of the slices up to the     }
{                          current progress is drawn                            }
var
  i, Endi: Integer;
  HoldBrushColor: TCanvasColor;
  HoldPenColor: TCanvasColor;
  ColorCount: Integer;
  MinColor: TCanvasColor;
  DegreesPerSlice: Extended;
  Degrees: Extended;
begin
  { Figure out how many colors we're dealing with }
  ColorCount := Length(Colors);
  if (ColorCount <= 0) then Exit;
  MinColor := Colors[Low(Colors)];

  HoldBrushColor := Canvas.Brush.Color;
  HoldPenColor := Canvas.Pen.Color;
  DegreesPerSlice := 360 / SubDivisions;

  { Determine how far to draw }
  if (Min = Max) or (Progress = Max) then
    Endi := Subdivisions
  else
    EndI := Trunc((Progress / ((Max-Min)/Subdivisions)));

  { Draw the progress }
  Canvas.Brush.Color := MinColor;
  if AdjustPenColor then
    Canvas.Pen.Color := MinColor;
  for i:=0 to Endi-1 do begin
    if (DisplayType = dtAllProgress) or (i = (Endi-1)) then begin
      Degrees := (360*i)/Subdivisions;
      if (Direction = pdBackward) then
        Degrees := 360-Degrees+DegreesPerSlice; {Reverse direction}

      DrawPieSlice(Canvas, CanvasR, Degrees + (Spacing/2), Degrees + DegreesPerSlice - (Spacing/2));
//      { Calculate the position of the first pie coordinate }
//      X3 := CanvasTrunc(((w-(OuterPadding*2))/2 + Cos(DegToRad(Degrees-DegreesPerSlice+(Spacing/2)))*(w-(OuterPadding*2))/2))+OuterPadding;
//      Y3 := CanvasTrunc(((h-(OuterPadding*2))/2 - Sin(DegToRad(Degrees-DegreesPerSlice+(Spacing/2)))*(h-(OuterPadding*2))/2))+OuterPadding;
//
//      { Calculate the position of the second pie coordinate }
//      X4 := CanvasTrunc(((w-(OuterPadding*2))/2 + Cos(DegToRad(Degrees-(Spacing/2)))*(w-(OuterPadding*2))/2))+OuterPadding;
//      Y4 := CanvasTrunc(((h-(OuterPadding*2))/2 - Sin(DegToRad(Degrees-(Spacing/2)))*(h-(OuterPadding*2))/2))+OuterPadding;
//
//      { Fill the slice }
//      Canvas.Pie(CanvasR.Left, CanvasR.Top, CanvasR.Right, CanvasR.Bottom, X3, Y3, X4, Y4);
//      Canvas.Chord(CanvasR.Left, CanvasR.Top, CanvasR.Right, CanvasR.Bottom, X3, Y3, X4, Y4);
    end;

    { Update the color }
    Canvas.Brush.Color := IndexToColor(i, SubDivisions, Colors);
    if AdjustPenColor then
      Canvas.Pen.Color := Canvas.Brush.Color;
  end;
  Canvas.Brush.Color := HoldBrushColor;
  Canvas.Pen.Color := HoldPenColor;
end;

procedure ProgressRect(const Canvas: TCanvas; const CanvasR: TCanvasRect;
   const OuterPadding: Integer; const Progress: Integer;
   const Min, Max: Integer; const Colors: TColors; const AdjustPenColor: Boolean;
   const SubDivisions: Integer; const SubShape: TRSShapeType;
   const SubShapeOrientation: TRSShapeOrientation;
   const SubShapeBevel: Boolean; const SubShapeBevelSize: Integer;
   const Spacing: Integer;
   const Orientation: TProgressOrientation;
   const Direction: TProgressDirection; const DisplayType: TProgressDisplayType);
{ Purpose: Similar to a progress bar, either in veritcal or hoizontal           }
{          orientation. Shapes are drawn up to the current progress position.   }
{ Arguments:                                                                    }
{          Canvas        - The canvas on which to draw                          }
{          CanvasR    - The rectangle defining the bounds of the Canvas      }
{          OuterPadding  - The number of pixels not to be used on the outer     }
{                          edges of the CanvasR                              }
{          Progress      - The current progress, not in percent, but relative   }
{                          to Max                                               }
{          Min           - The minimum scale value                              }
{          Max           - The maximum scale value                              }
{          Colors        - An array of ordered colors                           }
{          AdjustPenColor- Indicates whether the pen should be used, or left up }
{                          to the caller                                        }
{          SubDivisions  - The number of subdivisions in the scale (<= Max)     }
{          SubShape      - The type of shape to be drawn in the SubDivision     }
{          SubShapeOrientation -                                                }
{                          The orientation of the SubShape (doen't apply to     }
{                          some shapes                                          }
{          SubShapeBevel - Indicates whether the subshapes are beveled          }
{          SubShapeBevelSize - Specifies the width in pixels of the subshape    }
{                          bevel                                                }
{          Spacing       - The space in pixels between SubDivisions             }
{          Orientation   - Determines whether the progress scale is drawn       }
{                          horizontal or vertical on the Canvas                 }
{          Direction     - Indicates whether the progress meter draws forward   }
{                          or backward                                          }
{          DisplayType   - Determines whether the shape representing only the   }
{                          current progress, or all of the shapes up to the     }
{                          current progress is drawn                            }
var
  i, Endi: Integer;
  ShapeW, ShapeH,
  ShapeOx, ShapeOy, { Drawing offset for the shape }
  ShapeDeltaX, ShapeDeltaY: Single;
  w, h: TCanvasPixel;
  HoldBrushColor: TCanvasColor;
  HoldPenColor: TCanvasColor;
  ColorCount: Integer;
  MinColor: TCanvasColor;
  CRLeft: TCanvasPixel;
  CRTop: TCanvasPixel;
begin
  { Figure out how many colors we're dealing with }
  ColorCount := Length(Colors);
  if (ColorCount <= 0) then Exit;
  MinColor := Colors[Low(Colors)];

  HoldBrushColor := Canvas.Brush.Color;
  HoldPenColor := Canvas.Pen.Color;
  CRLeft := CanvasMin([CanvasR.Left, CanvasR.Right]);
  CRTop := CanvasMin([CanvasR.Top, CanvasR.Bottom]);
  w := abs(CanvasR.Right - CanvasR.Left);
  h := abs(CanvasR.Bottom - CanvasR.Top);
  ShapeOx := OuterPadding;
  ShapeOy := ShapeOx;
  if (Orientation = poHorizontal) then begin
    ShapeW := (w-(OuterPadding*2)-(Spacing*SubDivisions)) / SubDivisions;
    ShapeH := h - ShapeOy*2;
    ShapeDeltaX := (Spacing + ShapeW);
    if (Direction = pdBackward) then begin
      ShapeOx := w - ShapeOx - ShapeW;
      ShapeDeltaX := -ShapeDeltaX;
    end;
    ShapeDeltaY := 0;
  end
  else begin {poVertical}
    ShapeH := (h-(OuterPadding*2)-(Spacing*SubDivisions)) / SubDivisions;
    ShapeW := w - ShapeOx*2;
    ShapeDeltaY := (Spacing + ShapeH);
    if (Direction = pdBackward) then begin
      ShapeOy := h - ShapeOy - ShapeH;
      ShapeDeltaY := -ShapeDeltaY;
    end;
    ShapeDeltaX := 0;
  end;

  { Determine how far to draw }
  if (Min = Max) or (Progress = Max) then
    Endi := Subdivisions
  else
    EndI := Trunc((Progress / ((Max-Min)/Subdivisions)));

  { Draw the progress }
  Canvas.Brush.Color := MinColor;
  if AdjustPenColor then
    Canvas.Pen.Color := MinColor;
  for i:=1 to Endi do begin
    { Check for clipping cases }
    if (DisplayType = dtAllProgress) or (i = Endi) then begin
      if (Orientation = poHorizontal) and (ShapeOx+ShapeW > w-OuterPadding) then
        DrawShape(Canvas, CanvasRect(CRLeft + CanvasTrunc(MinValue([Int(w-OuterPadding),ShapeOx])),
                                CRTop + CanvasTrunc(ShapeOy),
                                CRLeft + w-OuterPadding,
                                CRTop + CanvasTrunc(ShapeOy+ShapeH)),
                  SubShape, SubShapeOrientation, SubShapeBevel, SubShapeBevelSize)
      else if (Orientation = poVertical) and (ShapeOy+ShapeH > h-OuterPadding) then
        DrawShape(Canvas, CanvasRect(CRLeft + CanvasTrunc(ShapeOx),
                                CRTop + CanvasTrunc(MinValue([Int(h-OuterPadding),ShapeOy])),
                                CRLeft + CanvasTrunc(ShapeOx+ShapeW),
                                CRTop + h-OuterPadding),
                  SubShape, SubShapeOrientation, SubShapeBevel, SubShapeBevelSize)
      else
        { Fill the subdivision normally }
        DrawShape(Canvas, CanvasRect(CRLeft + CanvasTrunc(ShapeOx),
                              CRTop + CanvasTrunc(ShapeOy),
                              CRLeft + CanvasTrunc(ShapeOx+ShapeW),
                              CRTop + CanvasTrunc(ShapeOy+ShapeH)),
                  SubShape, SubShapeOrientation, SubShapeBevel, SubShapeBevelSize);
    end;

    { Calculate the position of the next subdivision shape }
    ShapeOx := ShapeOx + ShapeDeltaX;
    ShapeOy := ShapeOy + ShapeDeltaY;

    { Update the color }
    Canvas.Brush.Color := IndexToColor(i, SubDivisions, Colors);
    if AdjustPenColor then
      Canvas.Pen.Color := Canvas.Brush.Color;
  end;
  Canvas.Brush.Color := HoldBrushColor;
  Canvas.Pen.Color := HoldPenColor;
end;

{ TRSCustomRect }

procedure TRSCustomRect.Assign(Source: TPersistent);
{ Purpose: Used to assign the contents of the source to this object instance   }
{ Arguments: }
{          Source  - The source object containing the properties to be         }
{                    assigned to this object instance }
begin
  if Source is TRSCustomRect then begin
    FBoundsRect := TRSCustomRect(Source).BoundsRect;
    Change;
  end
  else
    inherited Assign(Source);
end;

procedure TRSCustomRect.Change;
{ Purpose: Used to notify listeners of a change in this object instance        }
begin
  if Assigned(FOnChange) then FOnChange(Self);
end;

function TRSCustomRect.GetSize(const Index: Integer): Integer;
{ Purpose: General property getter for size properties in this class           }
{ Arguments: }
{          Index - The index of the property being requested                   }
{ Results: The size value associated with the property defined by the index    }
begin
  case Index of
    0: Result := BoundsRect.Left;
    1: Result := BoundsRect.Top;
    2: Result := BoundsRect.Right;
    3: Result := BoundsRect.Bottom;
    4: Result := BoundsRect.Bottom - BoundsRect.Top;
  else
    Result := BoundsRect.Right - BoundsRect.Left;
  end;
end;

procedure TRSCustomRect.MoveTo(X, Y: Integer);
{ Purpose: Moves the object rect Top,Left to the specified X,Y. Retains the    }
{          width and height of the rect.                                       }
{ Arguments: }
{          X - The X screen position                                           }
{          Y - The Y screen position                                           }
begin
     if (X = BoundsRect.Left) and (Y = BoundsRect.Top) then Exit;
     FBoundsRect.Right := X + FBoundsRect.Right - FBoundsRect.Left;
     FBoundsRect.Left := X;
     FBoundsRect.Bottom := Y + FBoundsRect.Bottom - FBoundsRect.Top;
     FBoundsRect.Top := Y;
     Change;
end;

procedure TRSCustomRect.Scale(DX, DY: Integer);
{ Purpose: Scales the object rect by the specified DX,DY.                      }
{ Arguments: }
{          DX - The delta X screen value in pixels                             }
{          DY - The delta Y screen value in pixels                             }
begin
     FBoundsRect := HelperFunctions.ScaleRect( BoundsRect, DX, DY );
     Change;
end;

procedure TRSCustomRect.SetBoundsRect(const Value: TRect);
{ Purpose: Property setter for the BoundsRect property                         }
{ Arguments: }
{          Value - The new value for the BoundsRect property                   }
begin
  FBoundsRect := Value;
  Change;
end;

procedure TRSCustomRect.SetSize(const Index, Value: Integer);
{ Purpose: General property setter for size properties in this class           }
{ Arguments: }
{          Index - The index of the property being changed                     }
{          Value - The new property value                                      }
begin
  if (Value <> GetSize(Index)) then begin
    case Index of
      0: FBoundsRect.Left := Value;
      1: FBoundsRect.Top := Value;
      2: FBoundsRect.Right := Value;
      3: FBoundsRect.Bottom := Value;
      4: FBoundsRect.Bottom := BoundsRect.Top + Value;
    else
      FBoundsRect.Right := BoundsRect.Left + Value;
    end;
    Change;
  end;
end;

procedure TRSCustomRect.Translate(DX, DY: Integer);
{ Purpose: Translates the object rect Top,Left by the specified DX,DY. Retains }
{          the width and height of the rect.                                   }
{ Arguments: }
{          DX - The delta X screen position                                    }
{          DY - The delta Y screen position                                    }
begin
  if (DX <> 0) then begin
    FBoundsRect.Left := FBoundsRect.Left + DX;
    FBoundsRect.Right := FBoundsRect.Right + DX;
  end;
  if (DY <> 0) then begin
    FBoundsRect.Top := FBoundsRect.Top + DY;
    FBoundsRect.Bottom := FBoundsRect.Bottom + DY;
  end;
  if (DX <> 0) or (DY <> 0) then
    Change;
end;

{ TRSCustomRSRect }

procedure TRSCustomRSRect.Assign(Source: TPersistent);
{ Purpose: Used to assign the contents of the source to this object instance   }
{ Arguments: }
{          Source  - The source object containing the properties to be         }
{                    assigned to this object instance }
begin
  if Source is TRSCustomRSRect then begin
    FBoundsRect := TRSCustomRSRect(Source).BoundsRect;
    Change;
  end
  else
    inherited Assign(Source);
end;

procedure TRSCustomRSRect.Change;
{ Purpose: Used to notify listeners of a change in this object instance        }
begin
  if Assigned(FOnChange) then FOnChange(Self);
end;

function TRSCustomRSRect.GetSize(const Index: Integer): Single;
{ Purpose: General property getter for size properties in this class           }
{ Arguments: }
{          Index - The index of the property being requested                   }
{ Results: The size value associated with the property defined by the index    }
begin
  case Index of
    0: Result := BoundsRect.Left;
    1: Result := BoundsRect.Top;
    2: Result := BoundsRect.Right;
    3: Result := BoundsRect.Bottom;
    4: Result := BoundsRect.Bottom - BoundsRect.Top;
  else
    Result := BoundsRect.Right - BoundsRect.Left;
  end;
end;

procedure TRSCustomRSRect.MoveTo(X, Y: Single);
{ Purpose: Moves the object rect Top,Left to the specified X,Y. Retains the    }
{          width and height of the rect.                                       }
{ Arguments: }
{          X - The X screen position                                           }
{          Y - The Y screen position                                           }
begin
     if (X = BoundsRect.Left) and (Y = BoundsRect.Top) then Exit;
     FBoundsRect.Right := X + FBoundsRect.Right - FBoundsRect.Left;
     FBoundsRect.Left := X;
     FBoundsRect.Bottom := Y + FBoundsRect.Bottom - FBoundsRect.Top;
     FBoundsRect.Top := Y;
     Change;
end;

procedure TRSCustomRSRect.Scale(DX, DY: Single);
{ Purpose: Scales the object rect by the specified DX,DY.                      }
{ Arguments: }
{          DX - The delta X screen value in pixels                             }
{          DY - The delta Y screen value in pixels                             }
begin
     FBoundsRect := ScaleRect( BoundsRect, DX, DY );
     Change;
end;

procedure TRSCustomRSRect.SetBoundsRect(const Value: TRSRect);
{ Purpose: Property setter for the BoundsRect property                         }
{ Arguments: }
{          Value - The new value for the BoundsRect property                   }
begin
  FBoundsRect := Value;
  Change;
end;

procedure TRSCustomRSRect.SetSize(const Index: Integer; const Value: Single);
{ Purpose: General property setter for size properties in this class           }
{ Arguments: }
{          Index - The index of the property being changed                     }
{          Value - The new property value                                      }
begin
  if (Value <> GetSize(Index)) then begin
    case Index of
      0: FBoundsRect.Left := Value;
      1: FBoundsRect.Top := Value;
      2: FBoundsRect.Right := Value;
      3: FBoundsRect.Bottom := Value;
      4: FBoundsRect.Bottom := BoundsRect.Top + Value;
    else
      FBoundsRect.Right := BoundsRect.Left + Value;
    end;
    Change;
  end;
end;

procedure TRSCustomRSRect.Translate(DX, DY: Single);
{ Purpose: Translates the object rect Top,Left by the specified DX,DY. Retains }
{          the width and height of the rect.                                   }
{ Arguments: }
{          DX - The delta X screen position                                    }
{          DY - The delta Y screen position                                    }
begin
  if (DX <> 0) then begin
    FBoundsRect.Left := FBoundsRect.Left + DX;
    FBoundsRect.Right := FBoundsRect.Right + DX;
  end;
  if (DY <> 0) then begin
    FBoundsRect.Top := FBoundsRect.Top + DY;
    FBoundsRect.Bottom := FBoundsRect.Bottom + DY;
  end;
  if (DX <> 0) or (DY <> 0) then
    Change;
end;

{$IFDEF DELPHI16PLUS}
function RSVector( const aVector: TVector ): TRSVector;
begin
  result.X := aVector.X;
  result.Y := aVector.Y;
  result.W := aVector.W;
end;

function RSMatrix( const aMatrix: TMatrix ): TRSMatrix;
begin
  result.m11 := aMatrix.m11;
  result.m12 := aMatrix.m12;
  result.m13 := aMatrix.m13;
  result.m21 := aMatrix.m21;
  result.m22 := aMatrix.m22;
  result.m23 := aMatrix.m23;
  result.m31 := aMatrix.m31;
  result.m32 := aMatrix.m32;
  result.m33 := aMatrix.m33;
end;

function Vector( const aVector: TRSVector ): TVector;
begin
  result.X := aVector.X;
  result.Y := aVector.Y;
  result.W := aVector.W;
end;

function Matrix( const aMatrix: TRSMatrix ): TMatrix;
begin
  result.m11 := aMatrix.m11;
  result.m12 := aMatrix.m12;
  result.m13 := aMatrix.m13;
  result.m21 := aMatrix.m21;
  result.m22 := aMatrix.m22;
  result.m23 := aMatrix.m23;
  result.m31 := aMatrix.m31;
  result.m32 := aMatrix.m32;
  result.m33 := aMatrix.m33;
end;

{$ENDIF}

function RSPoint( const aVector: TRSVector ): TRSPoint;
begin
  result.X := aVector.X;
  result.Y := aVector.Y;
end;

function RSVector( const X, Y: Single; const W: Single = 1.0 ): TRSVector;
begin
  result.X := X;
  result.Y := Y;
  result.W := W;
end;

function RSVector( const aPoint: TRSPoint; const W: Single ): TRSVector;
begin
  result.X := aPoint.X;
  result.Y := aPoint.Y;
  result.W := W;
end;

function RSVectorAdd(const AFirstVector, ASecondVector : TRSVector) : TRSVector;
begin
//{$EXCESSPRECISION OFF}
  Result.V[0] := AFirstVector.V[0] + ASecondVector.V[0];
  Result.V[1] := AFirstVector.V[1] + ASecondVector.V[1];
  Result.W := 1.0;
//{$EXCESSPRECISION ON}
end;

function RSVectorSubtract(const AFirstVector, ASecondVector : TRSVector) : TRSVector;
begin
//{$EXCESSPRECISION OFF}
  Result.V[0] := AFirstVector.V[0] - ASecondVector.V[0];
  Result.V[1] := AFirstVector.V[1] - ASecondVector.V[1];
  Result.W := 1.0;
//{$EXCESSPRECISION ON}
end;

function RSVectorAngleCosine(const v1, v2: TRSVector): Single;
begin
  if (RSVectorLength(v1) <> 0) and (RSVectorLength(v2) <> 0) then
  begin
    Result := RSVectorDotProduct(v1, v2) / (RSVectorLength(v1) * RSVectorLength(v2));
    if Result > 1 then
      Result := 1;
  end
  else
    Result := 0;
end;

function RSVectorDotProduct(const v1, v2: TRSVector): Single;
begin
  Result := v1.V[0] * v2.V[0] + v1.V[1] * v2.V[1];
end;

function RSVectorCrossProductZ(const v1, v2: TRSVector): Single;
begin
  // 3D Cross with Z = 0
  Result := v1.X * v2.Y - v1.Y * v2.X;
end;

function RSVectorAngle(const V, N: TRSVector): Single;
begin
  if RSVectorCrossProductZ(V, N) < 0 then
    Result := RadToDeg(ArcCos(RSVectorAngleCosine(V, N)))
  else
    Result := -RadToDeg(ArcCos(RSVectorAngleCosine(V, N)));
end;

function RSVectorLength(const V: TRSVector): Single;
begin
  Result := Sqrt(RSVectorNorm(V));
end;

function RSVectorNorm(const V: TRSVector): Single;
begin
  Result := V.V[0] * V.V[0] + V.V[1] * V.V[1];
end;

function RSVectorNormalize(const V: TRSVector): TRSVector;
var
  VecLen: Single;
begin
  VecLen := RSVectorLength(V);

  if VecLen > 0.0 then
  begin
    Result.X := V.X / VecLen;
    Result.Y := V.Y / VecLen;
    Result.W := V.W;
  end else
    Result := V;
end;

function RSVectorMultiply(const AVector: TRSVector; const AFactor: Single): TRSVector;
begin
  Result.X := AVector.X * AFactor;
  Result.Y := AVector.Y * AFactor;
  Result.W := AVector.W;
end;

function RSVectorMultiply(const AFactor: Single; const AVector: TRSVector): TRSVector;
begin
  Result.X := AFactor * AVector.X;
  Result.Y := AFactor * AVector.Y;
  Result.W := AVector.W;
end;

function RSVectorDivide(const AVector: TRSVector; const AFactor: Single): TRSVector;
var
  InvFactor: Single;
begin
  if AFactor <> 0 then
  begin
    InvFactor := 1 / AFactor;

    Result.X := AVector.X * InvFactor;
    Result.Y := AVector.Y * InvFactor;
    Result.W := AVector.W;
  end else
    Result := AVector;
end;

function RSVectorReflect(const V1, V2: TRSVector): TRSVector;
begin
//  Result := Self + AVector * (-2 * Self.DotProduct(AVector));
  Result := RSVectorAdd(V1, RSVectorMultiply(V2, (-2 * RSVectorDotProduct(V1,V2))));
end;

function RSMatrixTransform(const AMatrix: TRSMatrix; const aRect: TRSRect): TRSRect;
begin
  result.TopLeft := RSPoint(RSMatrixTransform( aMatrix, RSVector(aRect.TopLeft) ));
  result.BottomRight := RSPoint(RSMatrixTransform( aMatrix, RSVector(aRect.BottomRight) ));
end;

function RSMatrixTransform(const AMatrix: TRSMatrix; const AVector: TRSVector): TRSVector;
begin
  Result.V[0] := AVector.V[0] * AMatrix.M[0].V[0] + AVector.V[1] * AMatrix.M[1].V[0]
    + AVector.V[2] * AMatrix.M[2].V[0];
  Result.V[1] := AVector.V[0] * AMatrix.M[0].V[1] + AVector.V[1] * AMatrix.M[1].V[1]
    + AVector.V[2] * AMatrix.M[2].V[1];
  Result.V[2] := 1.0;
end;

function RSMatrixEquals( const M1, M2: TRSMatrix; const Epsilon: Single ): Boolean;
begin
  Result := SameValue(M1.m11, M2.m11, Epsilon) and SameValue(M1.m12, M2.m12, Epsilon) and
    SameValue(M1.m13, M2.m13, Epsilon) and SameValue(M1.m21, M2.m21, Epsilon) and
    SameValue(M1.m22, M2.m22, Epsilon) and SameValue(M1.m23, M2.m23, Epsilon) and
    SameValue(M1.m31, M2.m31, Epsilon) and SameValue(M1.m32, M2.m32, Epsilon) and
    SameValue(M1.m33, M2.m33, Epsilon);
end;

function RSMatrixMultiply(const M1, M2: TRSMatrix): TRSMatrix;
begin
  Result.m11 := M1.m11 * M2.m11 + M1.m12 * M2.m21 + M1.m13 * M2.m31;
  Result.m12 := M1.m11 * M2.m12 + M1.m12 * M2.m22 + M1.m13 * M2.m32;
  Result.m13 := M1.m11 * M2.m13 + M1.m12 * M2.m23 + M1.m13 * M2.m33;
  Result.m21 := M1.m21 * M2.m11 + M1.m22 * M2.m21 + M1.m23 * M2.m31;
  Result.m22 := M1.m21 * M2.m12 + M1.m22 * M2.m22 + M1.m23 * M2.m32;
  Result.m23 := M1.m21 * M2.m13 + M1.m22 * M2.m23 + M1.m23 * M2.m33;
  Result.m31 := M1.m31 * M2.m11 + M1.m32 * M2.m21 + M1.m33 * M2.m31;
  Result.m32 := M1.m31 * M2.m12 + M1.m32 * M2.m22 + M1.m33 * M2.m32;
  Result.m33 := M1.m31 * M2.m13 + M1.m32 * M2.m23 + M1.m33 * M2.m33;
end;

function RSMatrixMultiply(const APoint: TRSPoint; const AMatrix: TRSMatrix): TRSPoint;
begin
  Result.X := APoint.X * AMatrix.M[0].V[0] + APoint.Y * AMatrix.M[1].V[0]
    + AMatrix.M[2].V[0];
  Result.Y := APoint.X * AMatrix.M[0].V[1] + APoint.Y * AMatrix.M[1].V[1]
    + AMatrix.M[2].V[1];
end;

function RSMatrixMultiply(const AVector: TRSVector; const AMatrix: TRSMatrix): TRSVector;
begin
  Result.V[0] := AVector.V[0] * AMatrix.M[0].V[0] + AVector.V[1] * AMatrix.M[1].V[0]
    + AVector.V[2] * AMatrix.M[2].V[0];
  Result.V[1] := AVector.V[0] * AMatrix.M[0].V[1] + AVector.V[1] * AMatrix.M[1].V[1]
    + AVector.V[2] * AMatrix.M[2].V[1];
  Result.V[2] := AVector.V[0] * AMatrix.M[0].V[2] + AVector.V[1] * AMatrix.M[1].V[2]
    + AVector.V[2] * AMatrix.M[2].V[2];
end;

function RSMatrixDeterminant(const M: TRSMatrix): Single;
begin
  Result := M.M[0].V[0] * (M.M[1].V[1] * M.M[2].V[2] - M.M[2].V[1] * M.M[1].V[2]) - M.M[0].V[1] *
    (M.M[1].V[0] * M.M[2].V[2] - M.M[2].V[0] * M.M[1].V[2]) + M.M[0].V[2] *
    (M.M[1].V[0] * M.M[2].V[1] - M.M[2].V[0] * M.M[1].V[1]);
end;

procedure AdjointRSMatrix(var M: TRSMatrix);
var
  a1, a2, a3, b1, b2, b3, c1, c2, c3: Single;
begin
  a1 := M.M[0].V[0];
  a2 := M.M[0].V[1];
  a3 := M.M[0].V[2];
  b1 := M.M[1].V[0];
  b2 := M.M[1].V[1];
  b3 := M.M[1].V[2];
  c1 := M.M[2].V[0];
  c2 := M.M[2].V[1];
  c3 := M.M[2].V[2];
  M.M[0].V[0] := (b2 * c3 - c2 * b3);
  M.M[1].V[0] := -(b1 * c3 - c1 * b3);
  M.M[2].V[0] := (b1 * c2 - c1 * b2);

  M.M[0].V[1] := -(a2 * c3 - c2 * a3);
  M.M[1].V[1] := (a1 * c3 - c1 * a3);
  M.M[2].V[1] := -(a1 * c2 - c1 * a2);

  M.M[0].V[2] := (a2 * b3 - b2 * a3);
  M.M[1].V[2] := -(a1 * b3 - b1 * a3);
  M.M[2].V[2] := (a1 * b2 - b1 * a2);
end;

procedure ScaleRSMatrix(var M: TRSMatrix; const factor: Single);
var
  I: Integer;
begin
  for I := 0 to 2 do
  begin
    M.M[I].V[0] := M.M[I].V[0] * factor;
    M.M[I].V[1] := M.M[I].V[1] * factor;
    M.M[I].V[2] := M.M[I].V[2] * factor;
  end;
end;

procedure InvertRSMatrix(var M: TRSMatrix);
const
  Epsilon: Single = 1E-40;
var
  det: Single;
begin
  det := RSMatrixDeterminant(M);
  if Abs(det) < Epsilon then
    M := RSIdentityMatrix
  else
  begin
    AdjointRSMatrix(M);
    ScaleRSMatrix(M, 1 / det);
  end;
end;

function CreateRotationRSMatrix(const Angle: Single): TRSMatrix;
var
  cosine, sine: Extended;
begin
  // angle is in RADIANS
  SinCos(Angle, sine, cosine);

  Result.m11 := cosine;
  Result.m12 := sine;
  Result.m13 := 0;
  Result.m21 := -sine;
  Result.m22 := cosine;
  Result.m23 := 0;

  Result.m31 := 0;
  Result.m32 := 0;
  Result.m33 := 1;
end;

function CreateRotationRSMatrix(const Point: TRSPoint; const Angle: Single): TRSMatrix;
begin
  result := CreateTranslateRSMatrix(-Point.X, -Point.Y);
  result := RSMatrixMultiply(result, CreateRotationRSMatrix(Angle));
  result := RSMatrixMultiply(result, CreateTranslateRSMatrix(Point.X, Point.Y));
end;

function CreateScaleRSMatrix(const ScaleX, ScaleY: Single): TRSMatrix;
begin
  Result := RSIdentityMatrix;
  Result.m11 := ScaleX;
  Result.m22 := ScaleY;
end;

function CreateTranslateRSMatrix(const DX, DY: Single): TRSMatrix;
begin
  Result := RSIdentityMatrix;
  Result.m31 := DX;
  Result.m32 := DY;
end;

function RSMatrixEulerAngles( const M: TRSMatrix ): TRSVector;
begin
  // returns euler angles in radians
  result.x := arctan2(M.m32, M.m33);
  result.y := arctan2(-M.M31, sqrt(M.m32*M.m32 + M.m33*M.m33));
  result.w := arctan2(M.m21, m.m11);
end;

function RSMatrixRotationAngle( const M: TRSMatrix ): Single;
begin
  // returns rotation angle in degrees
  result := RadToDeg(arctan2(M.m21, m.m11));
  result := (360 - result);
  if (result / 360) >= 1 then
    result := result - trunc(result/360)*360;
end;

function RSMatrixTranslationVector( const M: TRSMatrix ): TRSPoint;
begin
  result.X := M.m31;
  result.Y := M.m32;
end;

function RSMatrixScaleVector( const M: TRSMatrix ): TRSPoint;
begin
  result.X := M.m11;
  result.Y := M.m22;
end;

function IsIdentityMatrix( const M: TRSMatrix ): Boolean;
begin
  result := (M.m11 = RSIdentityMatrix.m11) and
            (M.m12 = RSIdentityMatrix.m12) and
            (M.m13 = RSIdentityMatrix.m13) and
            (M.m21 = RSIdentityMatrix.m21) and
            (M.m22 = RSIdentityMatrix.m22) and
            (M.m23 = RSIdentityMatrix.m23) and
            (M.m31 = RSIdentityMatrix.m31) and
            (M.m32 = RSIdentityMatrix.m32) and
            (M.m33 = RSIdentityMatrix.m33);
end;

{ TRSVector }

{$IFNDEF DELPHI16PLUS}
function TRSVector.RSqrt(const AValue: Single): Single;
begin
  if (AValue > 0) then
    Result := 1 / Sqrt(AValue)
  else if (AValue < 0) then
    Result := 1 / Sqrt(-AValue)
  else
    Result := 1;
end;

function TRSVector.Add(const AVector: TRSVector): TRSVector;
begin
//{$EXCESSPRECISION OFF}
  Result.V[0] := Self.V[0] + AVector.V[0];
  Result.V[1] := Self.V[1] + AVector.V[1];
  Result.W := 1.0;
//{$EXCESSPRECISION ON}
end;

function TRSVector.Subtract(const AVector: TRSVector): TRSVector;
begin
//{$EXCESSPRECISION OFF}
  Result.V[0] := Self.V[0] - AVector.V[0];
  Result.V[1] := Self.V[1] - AVector.V[1];
  Result.W := 1.0;
//{$EXCESSPRECISION ON}
end;

function TRSVector.ToPointF: TRSPoint;
begin
  Result.X := Self.X;
  Result.Y := Self.Y;
end;

function TRSVector.ToRSPoint: TRSPoint;
begin
  Result.X := Self.X;
  Result.Y := Self.Y;
end;

function TRSVector.Norm: Single;
begin
  Result := Self.V[0] * Self.V[0] + Self.V[1] * Self.V[1];
end;

function TRSVector.Normalize: TRSVector;
var
  LInvLen: Single;
begin
//{$EXCESSPRECISION OFF}
  LInvLen := RSqrt((Self.Norm));
  Result.V[0] := Self.V[0] * LInvLen;
  Result.V[1] := Self.V[1] * LInvLen;
  Result.V[2] := 0.0;
//{$EXCESSPRECISION ON}
end;

class operator TRSVector.NotEqual(const AFirstVector,
  ASecondVector: TRSVector): Boolean;
begin
  Result := not (AFirstVector =  ASecondVector);
end;

function TRSVector.Scale(const AFactor: Single): TRSVector;
begin
//{$EXCESSPRECISION OFF}
  Result.V[0] := Self.V[0] * AFactor;
  Result.V[1] := Self.V[1] * AFactor;
  Result.W := 1;
//{$EXCESSPRECISION ON}
end;

class operator TRSVector.Subtract(const AFirstVector,
  ASecondVector: TRSVector): TRSVector;
begin
  Result := AFirstVector.Subtract(ASecondVector);
end;

function TRSVector.Length: Single;
begin
  Result := Sqrt(Self.Norm);
end;

function TRSVector.DotProduct(const AVector: TRSVector): Single;
begin
  Result := Self.V[0] * AVector.V[0] + Self.V[1] * AVector.V[1];
end;

class operator TRSVector.Equal(const AFirstVector,
  ASecondVector: TRSVector): Boolean;
begin
  Result := SameValue(AFirstVector.X, ASecondVector.X)
    and SameValue(AFirstVector.Y, ASecondVector.Y);
end;

class operator TRSVector.Explicit(AVector: TRSVector): TRSPoint;
begin
  Result := RSPoint(AVector);
end;

class operator TRSVector.Explicit(APointF: TRSPoint): TRSVector;
begin
  Result := RSVector(APointF);
end;

class operator TRSVector.Explicit(ASizeF: TRSSize): TRSVector;
begin
  Result := RSVector(ASizeF.cx,ASizeF.cy);
end;

class operator TRSVector.Implicit(AVector: TRSVector): TRSPoint;
begin
  Result := RSPoint(AVector);
end;

class operator TRSVector.Implicit(APointF: TRSPoint): TRSVector;
begin
  Result := RSVector(APointF);
end;

class operator TRSVector.Implicit(ASizeF: TRSSize): TRSVector;
begin
  Result := RSVector(ASizeF.cx,ASizeF.cy);
end;

constructor TRSVector.Create(const APoint: TRSPoint);
begin
  Self.X := APoint.X;
  Self.Y := APoint.Y;
  Self.W := 1;
end;

constructor TRSVector.Create(const AX, AY, AW: Single);
begin
  Self.X := AX;
  Self.Y := AY;
  Self.W := AW;
end;

function TRSVector.Reflect(const AVector: TRSVector): TRSVector;
begin
  Result := Self.Add( AVector.Scale(-2 * Self.DotProduct(AVector)));
end;

class operator TRSVector.Add(const AFirstVector, ASecondVector: TRSVector): TRSVector;
begin
  Result := AFirstVector.Add(ASecondVector);
end;

{ TRSMatrix }

function TRSMatrix.Transform(const AVector: TRSVector): TRSVector;
begin
  Result.V[0] := AVector.V[0] * Self.M[0].V[0] + AVector.V[1] * Self.M[1].V[0]
    + AVector.V[2] * Self.M[2].V[0];
  Result.V[1] := AVector.V[0] * Self.M[0].V[1] + AVector.V[1] * Self.M[1].V[1]
    + AVector.V[2] * Self.M[2].V[1];
  Result.V[2] := 1.0;
end;
{$ENDIF}

{ TRSPathData }

function TRSPathData.Clone: TRSPathData;
begin
  result := TRSPathDataClass(ClassType).Create;
  result.Assign(Self);
end;

constructor TRSPathData.Create;
begin
  inherited Create;
  FFormatStr := LongFormatStr;
{$IFDEF FMX}
  inherited OnChanged := PathChange;
{$ELSE}
  FRecalcBounds := True;
{$ENDIF}
end;

procedure TRSPathData.AddPolygon( const Poly: TPoints );
var
  i: Integer;
begin
  if Length(Poly) = 0 then Exit;
  BeginUpdate;
  try
    MoveTo( PointToRSPoint(Poly[0]) );
    for i := 1 to Length(Poly)-1 do
      LineTo( PointToRSPoint(Poly[i]) );
    if (Poly[0].X = Poly[Length(Poly)-1].X) and
      (Poly[0].Y = Poly[Length(Poly)-1].Y) then
      ClosePath;
  finally
    EndUpdate;
  end;
end;

procedure TRSPathData.AddPolygon( const Poly: TRSPoints );
var
  i: Integer;
begin
  if Length(Poly) = 0 then Exit;
  BeginUpdate;
  try
    MoveTo( Poly[0] );
    for i := 1 to Length(Poly)-2 do
      LineTo( Poly[i] );
    if (Poly[0].X = Poly[Length(Poly)-1].X) and
      (Poly[0].Y = Poly[Length(Poly)-1].Y) then
      ClosePath
    else
      LineTo( Poly[Length(Poly)-1] );
  finally
    EndUpdate;
  end;
end;

procedure TRSPathData.AddPolygons( const Poly: TPointPolygons );
var
  i: Integer;
begin
  BeginUpdate;
  try
    for i := 0 to Length(Poly) - 1 do
      AddPolygon(Poly[i]);
  finally
    EndUpdate;
  end;
end;

procedure TRSPathData.AddPolygons( const Poly: TRSPolygons );
var
  i: Integer;
begin
  BeginUpdate;
  try
    for i := 0 to Length(Poly) - 1 do
      AddPolygon(Poly[i]);
  finally
    EndUpdate;
  end;
end;


procedure TRSPathData.Append(Source: TRSPathData);
var
  j: Integer;
begin
  if Count = 0 then
    Assign(Source)
  else
  begin
{$IFDEF FMX}
    j := 0;
    BeginUpdate;
    try
      while j < Source.Count do
      begin
        case Source.Points[j].Kind of
          TPathPointKind.{$IFDEF DELPHI20PLUS}MoveTo{$ELSE}ppMoveTo{$ENDIF}: MoveTo(Source.Points[j].Point);
          TPathPointKind.{$IFDEF DELPHI20PLUS}LineTo{$ELSE}ppLineTo{$ENDIF}: LineTo(Source.Points[j].Point);
          TPathPointKind.{$IFDEF DELPHI20PLUS}CurveTo{$ELSE}ppCurveTo{$ENDIF}:
          begin
            if (j+2) >= Source.Count then
              raise Exception.Create('Path Copying Error');
            CurveTo(Source.Points[j].Point, Source.Points[j+1].Point, Source.Points[j+2].Point);
            Inc(j, 2);
          end;
          TPathPointKind.{$IFDEF DELPHI20PLUS}Close{$ELSE}ppClose{$ENDIF}: ClosePath;
        end;
        Inc(j);
      end;
    finally
      EndUpdate;
    end;
{$ELSE}
    j := Length(FPathData);
    SetLength(FPathData, Length(FPathData)+Source.Count);
    System.Move(Source.FPathData[0], FPathData[j], SizeOf(TRSPathPoint) * Source.Count);
    Change;
{$ENDIF}
  end;
end;

procedure TRSPathData.Append(const PathString: String);
var
  Source: TRSPathData;
begin
  Source := TRSPathData.Create;
  try
    Source.Data := {$IFDEF FMX}{$IFNDEF DELPHI17PLUS}AnsiString(PathString){$ELSE}PathString{$ENDIF}{$ELSE}PathString{$ENDIF};
    Append(Source);
  finally
    Source.{$IFDEF AUTOREFCOUNT}DisposeOf{$ELSE}Free{$ENDIF};
  end;
end;

procedure TRSPathData.SetFillMode(const Value: TRSFillMode);
begin
  if Value <> FillMode then
  begin
    FFillMode := Value;
    {$IFNDEF FMX}Change;{$ENDIF}
  end;
end;

procedure TRSPathData.SetFormatStr(const Value: String);
begin
  if Value <> FormatStr then
  begin
    // test format string
    FormatFloat(Value, 123.45, USFormatSettings);
    FFormatStr := Value;
  end;
end;

function TRSPathData.GetPathString: {$IF Defined(FMX) and not Defined(DELPHI17PLUS)}AnsiString{$ELSE}String{$IFEND};
//{$IFDEF FMX}
//begin
//  result := inherited Data;
//end;
//{$ELSE}
var
  I: Integer;
  Builder: TStringBuilder;
begin
  Builder := TStringBuilder.Create;
  Result := Builder.ToString;
  try
    I := 0;
    while I < Count do
    begin
//      case Points[I].Kind of
//        TRSPathPointKind.ppMoveTo:
//          Builder.Append('M')
//                 .Append(FloatToStr(Points[I].Point.X, USFormatSettings))
//                 .Append(',')
//                 .Append(FloatToStr(Points[I].Point.Y, USFormatSettings))
//                 .Append(' ');
//        TRSPathPointKind.ppLineTo:
//          Builder.Append('L')
//                 .Append(FloatToStr(Points[I].Point.X, USFormatSettings))
//                 .Append(',')
//                 .Append(FloatToStr(Points[I].Point.Y, USFormatSettings))
//                 .Append(' ');
//        TRSPathPointKind.ppCurveTo:
//          begin
//            Builder.Append('C')
//                   .Append(FloatToStr(Points[I].Point.X, USFormatSettings))
//                   .Append(',')
//                   .Append(FloatToStr(Points[I].Point.Y, USFormatSettings))
//                   .Append(' ');
//
//            Builder.Append(FloatToStr(Points[I + 1].Point.X, USFormatSettings))
//                   .Append(',')
//                   .Append(FloatToStr(Points[I + 1].Point.Y, USFormatSettings))
//                   .Append(' ');
//
//            Builder.Append(FloatToStr(Points[I + 2].Point.X, USFormatSettings))
//                   .Append(',')
//                   .Append(FloatToStr(Points[I + 2].Point.Y, USFormatSettings))
//                   .Append(' ');
//
//            Inc(I, 2);
//          end;
//        TRSPathPointKind.ppClose:
//          Builder.Append('Z ');
//      end;
      case Points[I].Kind of
        TRSPathPointKind.ppMoveTo:
          Builder.Append('M')
                 .Append(FormatFloat(FormatStr, Points[I].Point.X, USFormatSettings))
                 .Append(',')
                 .Append(FormatFloat(FormatStr, Points[I].Point.Y, USFormatSettings))
                 .Append(' ');
        TRSPathPointKind.ppLineTo:
          Builder.Append('L')
                 .Append(FormatFloat(FormatStr, Points[I].Point.X, USFormatSettings))
                 .Append(',')
                 .Append(FormatFloat(FormatStr, Points[I].Point.Y, USFormatSettings))
                 .Append(' ');
        TRSPathPointKind.ppCurveTo:
          begin
            Builder.Append('C')
                   .Append(FormatFloat(FormatStr, Points[I].Point.X, USFormatSettings))
                   .Append(',')
                   .Append(FormatFloat(FormatStr, Points[I].Point.Y, USFormatSettings))
                   .Append(' ');

            Builder.Append(FormatFloat(FormatStr, Points[I + 1].Point.X, USFormatSettings))
                   .Append(',')
                   .Append(FormatFloat(FormatStr, Points[I + 1].Point.Y, USFormatSettings))
                   .Append(' ');

            Builder.Append(FormatFloat(FormatStr, Points[I + 2].Point.X, USFormatSettings))
                   .Append(',')
                   .Append(FormatFloat(FormatStr, Points[I + 2].Point.Y, USFormatSettings))
                   .Append(' ');

            Inc(I, 2);
          end;
        TRSPathPointKind.ppClose:
          Builder.Append('Z ');
      end;
      Inc(I);
    end;
    Result := Builder.ToString;
  finally
    Builder.Free;
  end;
end;
//{$ENDIF}

procedure TRSPathData.SetPathString(const Value: {$IF Defined(FMX) and not Defined(DELPHI17PLUS)}AnsiString{$ELSE}String{$IFEND});
var
  PathString, toks: string;
  Token: Char;
  R, CP1, CP2: TRSPoint;
  CP0, CP3: TRSPoint;
  QP1: TRSPoint;
  Angle: Single;
  floatBool: Single;
  large, sweep: Boolean;
  Pos, I: Integer;
  Builder: TStringBuilder;
  TokenBuilder: TStringBuilder;
  LastLength: Integer;
begin
  Builder := TStringBuilder.Create;
  TokenBuilder := TStringBuilder.Create;
  BeginUpdate;
  try
    { change every #10#13 to space }
    for I := 1 to Length(Value) do
    begin
      if (Value[I] = #9) or
         (Value[I] = #10) or
         (Value[I] = #13) then
        Builder.Append(' ')
      else
        Builder.Append(Value[I]);
    end;
    PathString := Builder.ToString;

    { }
    Clear;
//    SetLength(FPathData, 0);
    Pos := 1;
    LastLength := -1;
    Token := #0;
    while (Builder.Length >= Pos) and (LastLength <> Pos) do
    begin
      LastLength := Pos;
      toks := GetTok(PathString, Pos);
      // Added the following 7 lines of code to get around TPathData.SetPathString
      // bug (Fixed bug in TPathData.SetPathString which would not properly parse
      // consecutive arcs (A/a) elements if the A/a token did not exist between
      // arc points)
      if (toks <> '') or (Pos >= Length(PathString)) then
      begin
        TokenBuilder.Clear;
        TokenBuilder.Append(toks);
      end
      else
        TokenBuilder.Append(Token);

      while TokenBuilder.Length > 0 do
      begin
        Token := TokenBuilder[0];
        TokenBuilder.Remove(0, 1);
        try
          case Token of
            'z', 'Z':
              ClosePath;
            'M':
            begin
              MoveTo(GetPointFromStr(PathString, Pos));
              while (Pos <= Length(PathString)) and ((RSIsDigit(PathString[Pos])) or (PathString[Pos] = '-')) do
              begin
                { next points }
                LineTo(GetPointFromStr(PathString, Pos));
              end;
              QP1 := LastPoint; // control point should be coincident with the current point
            end;
            'm':
            begin
              MoveToRel(GetPointFromStr(PathString, Pos));
              while (Pos <= Length(PathString)) and (RSIsDigit(PathString[Pos]) or (PathString[Pos] = '-')) do
              begin
                { next points }
                LineToRel(GetPointFromStr(PathString, Pos));
              end;
              QP1 := LastPoint; // control point should be coincident with the current point
            end;
            'L':
            begin
              LineTo(GetPointFromStr(PathString, Pos));
              while (Pos <= Length(PathString)) and (RSIsDigit(PathString[Pos]) or (PathString[Pos] = '-')) do
              begin
                { next points }
                LineTo(GetPointFromStr(PathString, Pos));
              end;
              QP1 := LastPoint; // control point should be coincident with the current point
            end;
            'l':
            begin
              LineToRel(GetPointFromStr(PathString, Pos));
              while (Pos <= Length(PathString)) and (RSIsDigit(PathString[Pos]) or (PathString[Pos] = '-')) do
              begin
                { next points }
                LineToRel(GetPointFromStr(PathString, Pos));
              end;
              QP1 := LastPoint; // control point should be coincident with the current point
            end;
            'C':
            begin
              CP1 := GetPointFromStr(PathString, Pos);
              CP2 := GetPointFromStr(PathString, Pos);
              CurveTo(CP1, CP2, GetPointFromStr(PathString, Pos));
              while (Pos <= Length(PathString)) and (RSIsDigit(PathString[Pos]) or (PathString[Pos] = '-')) do
              begin
                { next points }
                CP1 := GetPointFromStr(PathString, Pos);
                CP2 := GetPointFromStr(PathString, Pos);
                CurveTo(CP1, CP2, GetPointFromStr(PathString, Pos));
              end;
              QP1 := LastPoint; // control point should be coincident with the current point
            end;
            'c':
            begin
              CP1 := GetPointFromStr(PathString, Pos);
              CP2 := GetPointFromStr(PathString, Pos);
              CurveToRel(CP1, CP2, GetPointFromStr(PathString, Pos));
              while (Pos <= Length(PathString)) and (RSIsDigit(PathString[Pos]) or (PathString[Pos] = '-')) do
              begin
                { next points }
                CP1 := GetPointFromStr(PathString, Pos);
                CP2 := GetPointFromStr(PathString, Pos);
                CurveToRel(CP1, CP2, GetPointFromStr(PathString, Pos));
              end;
              QP1 := LastPoint; // control point should be coincident with the current point
            end;
            'S':
            begin
              CP2 := GetPointFromStr(PathString, Pos);
              SmoothCurveTo(CP2, GetPointFromStr(PathString, Pos));
              while (Pos <= Length(PathString)) and (RSIsDigit(PathString[Pos]) or (PathString[Pos] = '-')) do
              begin
                { next points }
                CP2 := GetPointFromStr(PathString, Pos);
                SmoothCurveTo(CP2, GetPointFromStr(PathString, Pos));
              end;
              QP1 := LastPoint; // control point should be coincident with the current point
            end;
            's':
            begin
              CP2 := GetPointFromStr(PathString, Pos);
              SmoothCurveToRel(CP2, GetPointFromStr(PathString, Pos));
              while (Pos <= Length(PathString)) and (RSIsDigit(PathString[Pos]) or (PathString[Pos] = '-')) do
              begin
                { next points }
                CP2 := GetPointFromStr(PathString, Pos);
                SmoothCurveToRel(CP2, GetPointFromStr(PathString, Pos));
              end;
              QP1 := LastPoint; // control point should be coincident with the current point
            end;
            'H':
            begin
              HLineTo(StrToFloat(GetNum(PathString, Pos), USFormatSettings));
              QP1 := LastPoint; // control point should be coincident with the current point
            end;
            'h':
            begin
              HLineToRel(StrToFloat(GetNum(PathString, Pos), USFormatSettings));
              QP1 := LastPoint; // control point should be coincident with the current point
            end;
            'V':
            begin
              VLineTo(StrToFloat(GetNum(PathString, Pos), USFormatSettings));
              QP1 := LastPoint; // control point should be coincident with the current point
            end;
            'v':
            begin
              VLineToRel(StrToFloat(GetNum(PathString, Pos), USFormatSettings));
              QP1 := LastPoint; // control point should be coincident with the current point
            end;
            'Q':
            begin
              //  CP0 = QP0
              //  CP3 = QP2
              //  The two control points for the cubic are:
              //  CP1 = QP0 + 2/3 *(QP1-QP0)
              //  CP2 = QP2 + 2/3 *(QP1-QP2)
              CP0 := LastPoint;
              QP1 := GetPointFromStr(PathString, Pos);
              CP3 := GetPointFromStr(PathString, Pos);
              CP1.x := CP0.x + 2/3*(QP1.x - CP0.x);
              CP1.y := CP0.y + 2/3*(QP1.y - CP0.y);
              CP2.x := CP3.x + 2/3*(QP1.x - CP3.x);
              CP2.y := CP3.y + 2/3*(QP1.y - CP3.y);
              CurveTo(CP1, CP2, CP3);
              while (Pos <= Length(PathString)) and (RSIsDigit(PathString[Pos]) or (PathString[Pos] = '-')) do
              begin
                CP0 := CP3;
                QP1 := GetPointFromStr(PathString, Pos);
                CP3 := GetPointFromStr(PathString, Pos);
                CP1.x := CP0.x + 2/3*(QP1.x - CP0.x);
                CP1.y := CP0.y + 2/3*(QP1.y - CP0.y);
                CP2.x := CP3.x + 2/3*(QP1.x - CP3.x);
                CP2.y := CP3.y + 2/3*(QP1.y - CP3.y);
                CurveTo(CP1, CP2, CP3);
              end;
            end;
            'q':
            begin
              //  CP0 = QP0
              //  CP3 = QP2
              //  The two control points for the cubic are:
              //  CP1 = QP0 + 2/3 *(QP1-QP0)
              //  CP2 = QP2 + 2/3 *(QP1-QP2)
              CP0 := LastPoint;
              QP1 := CP0+GetPointFromStr(PathString, Pos);
              CP3 := CP0+GetPointFromStr(PathString, Pos);
              CP1.x := CP0.x + 2/3*(QP1.x - CP0.x);
              CP1.y := CP0.y + 2/3*(QP1.y - CP0.y);
              CP2.x := CP3.x + 2/3*(QP1.x - CP3.x);
              CP2.y := CP3.y + 2/3*(QP1.y - CP3.y);
              CurveTo(CP1, CP2, CP3);
              while (Pos <= Length(PathString)) and (RSIsDigit(PathString[Pos]) or (PathString[Pos] = '-')) do
              begin
                CP0 := CP3;
                QP1 := CP0+GetPointFromStr(PathString, Pos);
                CP3 := CP0+GetPointFromStr(PathString, Pos);
                CP1.x := CP0.x + 2/3*(QP1.x - CP0.x);
                CP1.y := CP0.y + 2/3*(QP1.y - CP0.y);
                CP2.x := CP3.x + 2/3*(QP1.x - CP3.x);
                CP2.y := CP3.y + 2/3*(QP1.y - CP3.y);
                CurveTo(CP1, CP2, CP3);
              end;
            end;
            'T':
            begin
              //  CP0 = QP0
              //  CP3 = QP2
              //  The two control points for the cubic are:
              //  CP1 = QP0 + 2/3 *(QP1-QP0)
              //  CP2 = QP2 + 2/3 *(QP1-QP2)
              CP0 := LastPoint;
              QP1 := CP0 + (CP0 - QP1);
              CP3 := GetPointFromStr(PathString, Pos);
              CP1.x := CP0.x + 2/3*(QP1.x - CP0.x);
              CP1.y := CP0.y + 2/3*(QP1.y - CP0.y);
              CP2.x := CP3.x + 2/3*(QP1.x - CP3.x);
              CP2.y := CP3.y + 2/3*(QP1.y - CP3.y);
              CurveTo(CP1, CP2, CP3);
              while (Pos <= Length(PathString)) and (RSIsDigit(PathString[Pos]) or (PathString[Pos] = '-')) do
              begin
                CP0 := CP3;
                QP1 := CP0 + (CP0 - QP1);
                CP3 := GetPointFromStr(PathString, Pos);
                CP1.x := CP0.x + 2/3*(QP1.x - CP0.x);
                CP1.y := CP0.y + 2/3*(QP1.y - CP0.y);
                CP2.x := CP3.x + 2/3*(QP1.x - CP3.x);
                CP2.y := CP3.y + 2/3*(QP1.y - CP3.y);
                CurveTo(CP1, CP2, CP3);
              end;
            end;
            't':
            begin
              //  CP0 = QP0
              //  CP3 = QP2
              //  The two control points for the cubic are:
              //  CP1 = QP0 + 2/3 *(QP1-QP0)
              //  CP2 = QP2 + 2/3 *(QP1-QP2)
              CP0 := LastPoint;
              QP1 := CP0 + (CP0 - QP1);
              CP3 := CP0+GetPointFromStr(PathString, Pos);
              CP1.x := CP0.x + 2/3*(QP1.x - CP0.x);
              CP1.y := CP0.y + 2/3*(QP1.y - CP0.y);
              CP2.x := CP3.x + 2/3*(QP1.x - CP3.x);
              CP2.y := CP3.y + 2/3*(QP1.y - CP3.y);
              CurveTo(CP1, CP2, CP3);
              while (Pos <= Length(PathString)) and (RSIsDigit(PathString[Pos]) or (PathString[Pos] = '-')) do
              begin
                CP0 := CP3;
                QP1 := CP0 + (CP0 - QP1);
                CP3 := CP0+GetPointFromStr(PathString, Pos);
                CP1.x := CP0.x + 2/3*(QP1.x - CP0.x);
                CP1.y := CP0.y + 2/3*(QP1.y - CP0.y);
                CP2.x := CP3.x + 2/3*(QP1.x - CP3.x);
                CP2.y := CP3.y + 2/3*(QP1.y - CP3.y);
                CurveTo(CP1, CP2, CP3);
              end;
            end;
            'A':
            begin
              // arc
              if Count > 0 then
                CP1 := Points[Count-1].Point//FPathData[High(FPathData)].Point
              else
                CP1 := RSPoint(0, 0);
              R := GetPointFromStr(PathString, Pos);
              Angle := StrToFloat(GetNum(PathString, Pos), USFormatSettings);
              floatBool := StrToFloat(GetDigit(PathString,Pos));
//              if (floatBool <> 0) and (floatBool <> 1) then
//                raise Exception.Create('Invalid boolean, not 0 or 1');
              large := floatBool = 1;
              floatBool := StrToFloat(GetDigit(PathString,Pos));
//              if (floatBool <> 0) and (floatBool <> 1) then
//                raise Exception.Create('Invalid boolean, not 0 or 1');
              sweep := floatBool = 1;
              CP2 := GetPointFromStr(PathString, Pos);
              AddArcSvg(CP1, R, Angle, large, sweep, CP2);
              QP1 := LastPoint; // control point should be coincident with the current point
            end;
            'a':
            begin
              // arc rel
              if Count > 0 then
                CP1 := Points[Count-1].Point//FPathData[High(FPathData)].Point
              else
                CP1 := RSPoint(0, 0);
              R := GetPointFromStr(PathString, Pos);
              Angle := StrToFloat(GetNum(PathString, Pos), USFormatSettings);
              floatBool := StrToFloat(GetDigit(PathString,Pos));
//              if (floatBool <> 0) and (floatBool <> 1) then
//                raise Exception.Create('Invalid boolean, not 0 or 1');
              large := floatBool = 1;
              floatBool := StrToFloat(GetDigit(PathString,Pos));
//              if (floatBool <> 0) and (floatBool <> 1) then
//                raise Exception.Create('Invalid boolean, not 0 or 1');
              sweep := floatBool = 1;
              CP2 := GetPointFromStr(PathString, Pos);
              CP2.X := CP1.X + CP2.X;
              CP2.Y := CP1.Y + CP2.Y;
              AddArcSvg(CP1, R, Angle, large, sweep, CP2);
              QP1 := LastPoint; // control point should be coincident with the current point
            end;
          end;
        except
        end;
      end;
    end;
    // FMX seems to have a problem with a path that ends with a move, it
    // fills the space with black... so add a close path
    if (Count > 0) and (Points[Count-1].Kind = TRSPathPointKind.ppMoveTo) then
      ClosePath;
  finally
    FreeAndNil(TokenBuilder);
    FreeAndNil(Builder);
    EndUpdate;
  end;
end;

{$IFDEF FMX}
procedure TRSPathData.AssignTo(Dest: TPersistent);
var
  j: Integer;
begin
  if Dest is TPathData then
  begin
    j := 0;
    while j < Count do
    begin
      case Points[j].Kind of
        TPathPointKind.{$IFDEF DELPHI20PLUS}MoveTo{$ELSE}ppMoveTo{$ENDIF}: TPathData(Dest).MoveTo(Points[j].Point);
        TPathPointKind.{$IFDEF DELPHI20PLUS}LineTo{$ELSE}ppLineTo{$ENDIF}: TPathData(Dest).LineTo(Points[j].Point);
        TPathPointKind.{$IFDEF DELPHI20PLUS}CurveTo{$ELSE}ppCurveTo{$ENDIF}:
        begin
          if (j+2) >= Count then
            raise Exception.Create('Path Copying Error');
          TPathData(Dest).CurveTo(Points[j].Point, Points[j+1].Point, Points[j+2].Point);
          Inc(j, 2);
        end;
        TPathPointKind.{$IFDEF DELPHI20PLUS}Close{$ELSE}ppClose{$ENDIF}: TPathData(Dest).ClosePath;
      end;
      Inc(j);
    end;
  end
  else
    inherited AssignTo(Dest);
end;

procedure TRSPathData.PathChange(Sender: TObject);
begin
  if FUpdateCount > 0 then Exit;
  FLength := 0;
  SetLength(FDistances, 0);
  SetLength(FPolygons, 0);
  if Assigned(FOnChanged) then FOnChanged(Self);
end;
{$ENDIF}

procedure TRSPathData.BeginUpdate;
begin
  Inc(FUpdateCount);
end;

procedure TRSPathData.EndUpdate;
begin
  Dec(FUpdateCount);
  if UpdateCount = 0 then {$IFDEF FMX}PathChange(Self);{$ELSE}Change;{$ENDIF}
end;

procedure TRSPathData.DrawPath(const Canvas: TCanvas; Opacity: Single);
begin
{$IFDEF FMX}
  Canvas.DrawPath(Self, Opacity);
{$ELSE}
  DrawPath(Canvas, False, Opacity);
{$ENDIF}
end;

procedure TRSPathData.FillPath(const Canvas: TCanvas; Opacity: Single);
{$IFDEF FMX}
var
  p: TRSPathData;
begin
  // FMX (at least up to XE5) has a bug where if the last point is not a Close
  // it can just black out the entire canvas
  if (Canvas.Brush.Kind <> TBrushKind.{$IFDEF DELPHI20PLUS}None{$ELSE}bkNone{$ENDIF}) and
    (Count > 0) and
    (Points[Count-1].Kind <> TRSPathPointKind.{$IFDEF DELPHI20PLUS}Close{$ELSE}ppClose{$ENDIF}) then
  begin
    p := TRSPathData.Create;
    try
      p.Assign(Self);
      p.ClosePath;
      Canvas.FillPath(p, Opacity);
    finally
      p.Free;
    end;
  end
  else
    Canvas.FillPath(Self, Opacity);
{$ELSE}
begin
  DrawPath(Canvas, True, Opacity);
{$ENDIF}
end;

procedure TRSPathData.AddRectangleForSegment( p1, p2: TRSPoint; OffSet: Single );
var
  xLineDelta: Single;
  yLineDelta: Single;
  d: Single;
  normX, normY: Single;
  xPerp, yPerp: Single;
  Pt1, Pt2, Pt3, Pt4: TRSPoint;
begin
  // create a rectangle around a line segment (useful for outlining a path)
  yLineDelta := p2.Y - p1.Y;
  xLineDelta := p2.X - p1.X;
  // if points make a point, not a line
  if (xLineDelta = 0) and (yLineDelta = 0) then Exit;

  d := Sqrt(Sqr(xLineDelta) + Sqr(yLineDelta));

  normX := xLineDelta / d;
  normY := yLineDelta / d;

  xPerp := offset * normX;
  yPerp := offset * normY;

  // Create perpendicular points
  Pt1.x := p1.x + yPerp;
  Pt1.y := p1.y - xPerp;
  Pt2.x := p1.x - yPerp;
  Pt2.y := p1.y + xPerp;

  Pt3.x := p2.x - yPerp;
  Pt3.y := p2.y + xPerp;
  Pt4.x := p2.x + yPerp;
  Pt4.y := p2.y - xPerp;
  MoveTo(Pt1);
  LineTo(Pt2);
  LineTo(Pt3);
  LineTo(Pt4);
  ClosePath;
end;

procedure TRSPathData.AddPathOutline( const P: TRSPathData; StrokeWidth: Single );
var
  NewP: TRSPathData;
  aPoint: TRSPathPoint;
  NextPoint: TRSPathPoint;
  i: Integer;
  ClosePoint: TRSPoint;
begin
  StrokeWidth := StrokeWidth / 2;
  if P.Count < 2 then Exit;

  // create a crude outline around the path and add to current path
  // note that this routine ignores joins and miters so it is very crude
  // mostly useful for detection of mouse along outline
  NewP := TRSPathData.Create;
  BeginUpdate;
  try
    NewP.Assign(P);
    NewP.Flatten;

    aPoint := NewP.Points[0];
    ClosePoint := aPoint.Point;
    i := 1;
    while i < NewP.Count do
    begin
      // get line segment
      NextPoint := NewP[i];
      case NextPoint.Kind of
        TRSPathPointKind.ppMoveTo:
          ClosePoint := aPoint.Point;
        TRSPathPointKind.ppLineTo:
          AddRectangleForSegment(aPoint.Point, NextPoint.Point, StrokeWidth);
        TRSPathPointKind.ppCurveTo: ;
        TRSPathPointKind.ppClose:
        begin
          AddRectangleForSegment(aPoint.Point, NextPoint.Point, StrokeWidth);
          AddRectangleForSegment(NextPoint.Point, ClosePoint, StrokeWidth);
          ClosePoint := NextPoint.Point;
        end;
      end;
      aPoint := NextPoint;
      Inc(i);
    end;
  finally
    EndUpdate;
    NewP.Free;
  end;
end;

function TRSPathData.PointInPathOutline(const P: TRSPoint; StrokeWidth: Single): Boolean;
var
  NewP: TRSPathData;
  aPoint: TRSPathPoint;
  NextPoint: TRSPathPoint;
  i: Integer;
  ClosePoint: TRSPoint;
  CheckPath: TRSPathData;

  function IsInOutline( p1, p2: TRSPoint ): Boolean;
  begin
    CheckPath.Clear;
    CheckPath.AddRectangleForSegment(p1, p2, StrokeWidth);
    result := CheckPath.PointInPath(P);
  end;
begin
  result := False;
  StrokeWidth := StrokeWidth / 2;
  if Count < 2 then Exit;

  result := True;
  NewP := TRSPathData.Create;
  CheckPath := TRSPathData.Create;
  try
    NewP.Assign(Self);
    NewP.Flatten;

    aPoint := NewP.Points[0];
    ClosePoint := aPoint.Point;
    i := 1;
    while i < NewP.Count do
    begin
      // get line segment
      NextPoint := NewP[i];
      case NextPoint.Kind of
        TRSPathPointKind.ppMoveTo:
          ClosePoint := aPoint.Point;
        TRSPathPointKind.ppLineTo:
          if IsInOutline(aPoint.Point, NextPoint.Point) then Exit;
        TRSPathPointKind.ppCurveTo: ;
        TRSPathPointKind.ppClose:
        begin
          if IsInOutline(aPoint.Point, NextPoint.Point) then Exit;
          if IsInOutline(NextPoint.Point, ClosePoint) then Exit;
          ClosePoint := NextPoint.Point;
        end;
      end;
      aPoint := NextPoint;
      Inc(i);
    end;
  finally
    CheckPath.Free;
    NewP.Free;
  end;
  Result := False;
end;

function TRSPathData.PointInPathOutline(const X, Y: Single; StrokeWidth: Single): Boolean;
begin
  result := PointInPathOutline(RSPoint(X,Y), StrokeWidth);
end;

function TRSPathData.PointInPath(const P: TRSPoint): Boolean;
begin
  result := PointInPath(P.X, P.Y);
end;

function TRSPathData.PointInPath(const X, Y: Single): Boolean;
var
  i: Integer;
  j: Integer;
  k: Integer;
  PathPoint: TRSPathPoint;
  aPoints: TRSPoints;
  BezierPoints: TRSPoints;
begin
  i := 0;
  while i < Count do
  begin
    PathPoint := Points[i];
    case PathPoint.Kind of
      TRSPathPointKind.ppMoveTo:
      begin
        if (i > 0) and (Points[i - 1].Kind <> TRSPathPointKind.ppClose) then
        begin
          result := {$IFDEF FMX}FMX.RS.Graphics.{$ELSE}RSGraphics.{$ENDIF}PointInPath(aPoints, X, Y);
          if result then Exit;
          SetLength(aPoints, 1);
        end
        else
          SetLength(aPoints, Length(aPoints)+1);
        aPoints[Length(aPoints)-1] := PathPoint.Point;
      end;
      TRSPathPointKind.ppLineTo:
      begin
        SetLength(aPoints, Length(aPoints)+1);
        aPoints[Length(aPoints)-1] := PathPoint.Point;
      end;
      TRSPathPointKind.ppCurveTo:
      begin
        SetLength(BezierPoints, 4);
        BezierPoints[0] := Points[i-1].Point;
        BezierPoints[1] := PathPoint.Point;
        BezierPoints[2] := Points[i+1].Point;
        BezierPoints[3] := Points[i+2].Point;
        // use smoothed for better selection on curves
        BezierPoints := BezierPolygonPoints(BezierPoints, True, False, 1);
        i := i + 2;
        // copy aPoints to current aPoints
        j := Length(aPoints);
        SetLength(aPoints, Length(aPoints)+Length(BezierPoints));
        for k := 0 to Length(BezierPoints)-1 do
        begin
          aPoints[j] := BezierPoints[k];
          Inc(j);
        end;
      end;
      TRSPathPointKind.ppClose:
      begin
        SetLength(aPoints, Length(aPoints)+1);
        aPoints[Length(aPoints)-1] := aPoints[0];
        result := {$IFDEF FMX}FMX.RS.Graphics.{$ELSE}RSGraphics.{$ENDIF}PointInPath(aPoints, X, Y);
        if result then Exit;
        SetLength(aPoints, 0);
      end;
    end;
    i := i + 1;
  end;
  result := {$IFDEF FMX}FMX.RS.Graphics.{$ELSE}RSGraphics.{$ENDIF}PointInPath(aPoints, X, Y);
end;

function TRSPathData.SplitDataByLength( const Len: Integer; const Delimiter: String ): String;
var
  I: Integer;
  RunLen: Integer;
  BeforeLen: Integer;
  Builder: TStringBuilder;
begin
  Builder := TStringBuilder.Create;
  Result := Builder.ToString;
  try
    I := 0;
    RunLen := 0;
    while I < Count do
    begin
      BeforeLen := Builder.Length;
      case Points[I].Kind of
        TRSPathPointKind.ppMoveTo:
          Builder.Append('M')
                 .Append(FormatFloat(FormatStr, Points[I].Point.X, USFormatSettings))
                 .Append(',')
                 .Append(FormatFloat(FormatStr, Points[I].Point.Y, USFormatSettings))
                 .Append(' ');
        TRSPathPointKind.ppLineTo:
          Builder.Append('L')
                 .Append(FormatFloat(FormatStr, Points[I].Point.X, USFormatSettings))
                 .Append(',')
                 .Append(FormatFloat(FormatStr, Points[I].Point.Y, USFormatSettings))
                 .Append(' ');
        TRSPathPointKind.ppCurveTo:
          begin
            Builder.Append('C')
                   .Append(FormatFloat(FormatStr, Points[I].Point.X, USFormatSettings))
                   .Append(',')
                   .Append(FormatFloat(FormatStr, Points[I].Point.Y, USFormatSettings))
                   .Append(' ');

            Builder.Append(FormatFloat(FormatStr, Points[I + 1].Point.X, USFormatSettings))
                   .Append(',')
                   .Append(FormatFloat(FormatStr, Points[I + 1].Point.Y, USFormatSettings))
                   .Append(' ');

            Builder.Append(FormatFloat(FormatStr, Points[I + 2].Point.X, USFormatSettings))
                   .Append(',')
                   .Append(FormatFloat(FormatStr, Points[I + 2].Point.Y, USFormatSettings))
                   .Append(' ');

            Inc(I, 2);
          end;
        TRSPathPointKind.ppClose:
          Builder.Append('Z ');
      end;
      Inc(I);
      RunLen := RunLen + Builder.Length - BeforeLen;
      if RunLen > Len then
      begin
        Builder.Append(Delimiter);
        RunLen := 0;
      end;
    end;
    Result := Builder.ToString;
  finally
    Builder.Free;
  end;
end;

function TRSPathData.CalculateDistances: Double;
var
  i: Integer;
  j: Integer;
  LastPoint: TRSPoint;
begin
  result := 0;
  FLength := 0;
  ToPolygons;
  SetLength(FDistances,0); // clear out memory
  SetLength(FDistances, Length(FPolygons));
  if Length(FDistances) = 0 then Exit;

  for i := 0 to Length(FPolygons) - 1 do
  begin
    SetLength(FDistances[i], Length(FPolygons[i]));
    if Length(FDistances[i]) = 0 then Continue;

    FDistances[i][0] := result;
    LastPoint := FPolygons[i][0];
    for j := 1 to Length(FPolygons[i]) - 1 do
    begin
      result := result + RSDistance( LastPoint, FPolygons[i][j] );
      FDistances[i][j] := result;
      LastPoint := FPolygons[i][j];
    end;
  end;
  FLength := result;
end;

function TRSPathData.GetLength: Double;
begin
  if FLength = 0 then
    FLength := CalculateDistances;
  result := FLength;
end;

function TRSPathData.GetDistances: TDistances;
begin
  if (FLength = 0) and (Count > 0) then
    FLength := CalculateDistances;  // distances are created as part of method
  result := FDistances;
end;

function TRSPathData.ToPolygons: TRSPolygons;
var
  i: Integer;
  j: Integer;
  k: Integer;
  p: TRSPolygon;
  BezierPoints: TRSPoints;
  PathPoint: TRSPathPoint;
begin
  if (FLength > 0) then
  begin
    result := FPolygons;
    Exit;
  end;
  // back to back bezier curves can give bad results  It is recommended that
  // Flatten is called before calling ToPolygons
  SetLength(result, 0);
  SetLength(P, 0);
  i := 0;
  while i < Count do
  begin
    PathPoint := Points[i];
    case PathPoint.Kind of
      TRSPathPointKind.ppMoveTo:
      begin
        if Length(p) > 0 then
        begin
          SetLength(result, Length(result)+1);
          result[Length(result)-1] := p;
        end;
        SetLength(p, 1);
        p[0] := PathPoint.Point;
      end;
      TRSPathPointKind.ppLineTo:
      begin
        SetLength(p, Length(p)+1);
        p[Length(p)-1] := PathPoint.Point;
      end;
      TRSPathPointKind.ppCurveTo:
      begin
        SetLength(BezierPoints, 4);
        BezierPoints[0] := Points[i-1].Point;
        BezierPoints[1] := PathPoint.Point;
        BezierPoints[2] := Points[i+1].Point;
        BezierPoints[3] := Points[i+2].Point;
        BezierPoints := BezierPolygonPoints(BezierPoints);
        i := i + 2;
        // copy bezier points to current points
        j := Length(P);
        SetLength(P, j+Length(BezierPoints));
        for k := 0 to Length(BezierPoints)-1 do
        begin
          P[j] := BezierPoints[k];
          Inc(j);
        end;
      end;
      TRSPathPointKind.ppClose:
      begin
        SetLength(p, Length(p)+1);
        p[Length(p)-1] := P[0];
        SetLength(result, Length(result)+1);
        result[Length(result)-1] := p;
        SetLength(p, 0);
      end;
    end;
    Inc(i);
  end;
  if Length(p) > 0 then
  begin
    SetLength(result, Length(result)+1);
    result[Length(result)-1] := p;
  end;
  FPolygons := result;
end;

procedure DrawArcWithBezier(Path: TRSPathData; CenterX, CenterY, RadiusX, RadiusY, StartAngle, SweepRange: Single;
  UseMoveTo: Boolean);
var
  Coord: array of TRSPoint;
  Pts: array of TRSPoint;
  A, B, C, X, Y: Single;
  ss, cc: Single;
  I: Integer;
begin
  if SweepRange = 0 then
  begin
    if UseMoveTo then
    begin
      if (Path.Count = 0) then
        Path.MoveTo(RSPoint(CenterX + RadiusX * Cos(StartAngle), CenterY - RadiusY * Sin(StartAngle)))
      else
        Path.LineTo(RSPoint(CenterX + RadiusX * Cos(StartAngle), CenterY - RadiusY * Sin(StartAngle)));
    end;
    Path.LineTo(RSPoint(CenterX + RadiusX * Cos(StartAngle), CenterY - RadiusY * Sin(StartAngle)));
    Exit;
  end;
  B := Sin(SweepRange / 2);
  C := Cos(SweepRange / 2);
  A := 1 - C;
  X := A * 4.0 / 3.0;
  Y := B - X * C / B;
  ss := Sin(StartAngle + SweepRange / 2);
  cc := Cos(StartAngle + SweepRange / 2);
  SetLength(Coord, 4);
  Coord[0] := RSPoint(C, -B);
  Coord[1] := RSPoint(C + X, -Y);
  Coord[2] := RSPoint(C + X, Y);
  Coord[3] := RSPoint(C, B);
  SetLength(Pts, 4);
  for I := 0 to 3 do
  begin
    Pts[I] := RSPoint(CenterX + RadiusX * (Coord[I].X * cc - Coord[I].Y * ss),
      CenterY + RadiusY * (Coord[I].X * ss + Coord[I].Y * cc));
  end;
  if UseMoveTo then
  begin
    if (Path.Count = 0) then
      Path.MoveTo(Pts[0])
    else
      Path.LineTo(Pts[0]);
  end;
  Path.CurveTo(Pts[1], Pts[2], Pts[3]);
end;

procedure TRSPathData.SmoothCurveTo(const ControlPoint2, EndPoint: TRSPoint);
{$IFNDEF FMX}
var
  ControlPoint1: TRSPoint;
  i: Integer;
  LP: TRSPoint;
begin
  LP := LastPoint;
  if (Count > 2) and (FPathData[High(FPathData)].Kind <> TRSPathPointKind.ppMoveTo) then
  begin
    ControlPoint1.X := LastPoint.X + (LastPoint.X - FPathData[High(FPathData) - 1].Point.X);
    ControlPoint1.Y := LastPoint.Y + (LastPoint.Y - FPathData[High(FPathData) - 1].Point.Y);
  end
  else
    ControlPoint1 := LP;
  i := Length(FPathData);
  SetLength(FPathData, i + 3);
  FPathData[i].Kind := TRSPathPointKind.ppCurveTo;
  FPathData[i].Point := ControlPoint1;
  Inc(i);
  FPathData[i].Kind := TRSPathPointKind.ppCurveTo;
  FPathData[i].Point := ControlPoint2;
  Inc(i);
  FPathData[i].Kind := TRSPathPointKind.ppCurveTo;
  FPathData[i].Point := EndPoint;
  Change;
{$ELSE}
begin
  if (Count <= 2) or (Points[Count-1].Kind = TRSPathPointKind.ppMoveTo) then
    CurveTo(LastPoint, ControlPoint2, EndPoint)
  else
    inherited SmoothCurveTo(ControlPoint2, EndPoint);
{$ENDIF}
end;

procedure TRSPathData.SmoothCurveToRel(const ControlPoint2, EndPoint: TRSPoint);
{$IFNDEF FMX}
var
  LP, ControlPoint1: TRSPoint;
  i: Integer;
begin
  LP := LastPoint;
  if (Count > 2) and (FPathData[High(FPathData)].Kind <> TRSPathPointKind.ppMoveTo) then
  begin
    ControlPoint1.X := LastPoint.X + (LastPoint.X - FPathData[High(FPathData) - 1].Point.X);
    ControlPoint1.Y := LastPoint.Y + (LastPoint.Y - FPathData[High(FPathData) - 1].Point.Y);
  end
  else
    ControlPoint1 := LP;
  i := Length(FPathData);
  SetLength(FPathData, i + 3);
  FPathData[i].Kind := TRSPathPointKind.ppCurveTo;
  FPathData[i].Point := RSPoint(ControlPoint1.X, ControlPoint1.Y);
  Inc(i);
  FPathData[i].Kind := TRSPathPointKind.ppCurveTo;
  FPathData[i].Point := RSPoint(LP.X + ControlPoint2.X, LP.Y + ControlPoint2.Y);
  Inc(i);
  FPathData[i].Kind := TRSPathPointKind.ppCurveTo;
  FPathData[i].Point := RSPoint(LP.X + EndPoint.X, LP.Y + EndPoint.Y);
  Change;
{$ELSE}
begin
  if (Count <= 2) or (Points[Count-1].Kind = TRSPathPointKind.ppMoveTo) then
    CurveToRel(RSPoint(0,0), ControlPoint2, EndPoint)
  else
    inherited SmoothCurveToRel(ControlPoint2, EndPoint);
{$ENDIF}
end;

{$IFNDEF FMX}
procedure TRSPathData.Assign(Source: TPersistent);
begin
  if Source is TRSPathData then
  begin
      SetLength(FPathData, TRSPathData(Source).Count);
      System.Move(TRSPathData(Source).FPathData[0], FPathData[0], SizeOf(TRSPathPoint) * Count);
      FFillMode := TRSPathData(Source).FillMode;
      FFormatStr := TRSPathData(Source).FormatStr;
      Change;
      FLength := TRSPathData(Source).FLength;
      FDistances := Copy(TRSPathData(Source).FDistances);
      FPolygons := Copy(TRSPathData(Source).FPolygons);
  end
  else
    inherited Assign(Source);
end;

function TRSPathData.GetCount: Integer;
begin
  Result := Length(FPathData);
end;

function TRSPathData.GetPoint(AIndex: Integer): TRSPathPoint;
begin
  Result := FPathData[AIndex];
end;

procedure TRSPathData.DefineProperties(Filer: TFiler);
begin
  inherited;
  Filer.DefineBinaryProperty('Path', ReadPath, WritePath, Count > 0);
end;

procedure TRSPathData.DrawPath(const Canvas: TCanvas; Filled: Boolean; Opacity: Single);
var
  i: Integer;
  PathPoint: TRSPathPoint;
  BezierPoints: TPoints;
  DC: HDC;
begin
  if (Opacity = 0) or (Length(FPathData) < 2) then Exit;
  // assume FitToRect has been called
  // and Pen/Brush have been set
  DC := Canvas.Handle;
  BeginPath(DC);
  SetLength(BezierPoints, 3);
  i := 0;
  while i < Length(FPathData) do
  begin
    PathPoint := FPathData[i];
    case PathPoint.Kind of
      TRSPathPointKind.ppMoveTo: MoveToEx(DC, Round(PathPoint.Point.X), Round(PathPoint.Point.Y), nil);
      TRSPathPointKind.ppLineTo: Windows.LineTo(DC, Round(PathPoint.Point.X), Round(PathPoint.Point.Y));
      TRSPathPointKind.ppCurveTo:
      begin
        BezierPoints[0] := RSPointToPoint(PathPoint.Point);
        BezierPoints[1] := RSPointToPoint(Points[i+1].Point);
        BezierPoints[2] := RSPointToPoint(Points[i+2].Point);
        i := i + 2;
        PolyBezierTo(DC, BezierPoints[0], 3);
      end;
      TRSPathPointKind.ppClose: CloseFigure(DC);
    end;
    Inc(i);
  end;
  EndPath(DC);
  if Filled then
    StrokeAndFillPath(DC)
  else
    StrokePath(DC);
end;

procedure TRSPathData.ReadPath(Stream: TStream);
var
  S: Longint;
  I: Integer;
  k: Byte;
  P: TRSPoint;
begin
  Stream.Read(S, SizeOf(S));
{$IFDEF FPC_BIG_ENDIAN}
  ReverseBytes(@S, 4);
{$ENDIF}
  SetLength(FPathData, S);
  if S > 0 then
  begin
    if (Stream.Size - 4) div S = 9 then
    begin
      for I := 0 to S - 1 do
      begin
        Stream.Read(k, 1);
        Stream.Read(P, SizeOf(P));
{$IFDEF FPC_BIG_ENDIAN}
        ReverseBytes(@P.X, 4);
        ReverseBytes(@P.Y, 4);
{$ENDIF}
        FPathData[I].Kind := TRSPathPointKind(k);
        FPathData[I].Point := P;
      end;
    end
    else
    begin
      Stream.Read(FPathData[0], S * SizeOf(TRSPathPoint));
{$IFDEF FPC_BIG_ENDIAN}
      for I := 0 to S * 3 - 1 do
        ReverseBytes(@PColorArray(PathData)[I], 4);
{$ENDIF}
    end;
  end;
  Change;
end;

procedure TRSPathData.WritePath(Stream: TStream);
var
  S: Longint;
begin
  S := Count;
  Stream.Write(S, SizeOf(S));
  if S > 0 then
    Stream.Write(FPathData[0], S * SizeOf(TRSPathPoint));
end;

function TRSPathData.LastPoint: TRSPoint;
begin
  if Count > 0 then
    Result := FPathData[High(FPathData)].Point
  else
    Result := RSPoint(0, 0);
end;

procedure TRSPathData.MoveTo(const P: TRSPoint);
begin
  SetLength(FPathData, Length(FPathData) + 1);
  FPathData[High(FPathData)].Kind := TRSPathPointKind.ppMoveTo;
  FPathData[High(FPathData)].Point := P;
  FStartPoint := FPathData[High(FPathData)].Point;
  Change;
end;

procedure TRSPathData.MoveToRel(const P: TRSPoint);
var
  LP: TRSPoint;
begin
  LP := LastPoint;
  SetLength(FPathData, Length(FPathData) + 1);
  FPathData[High(FPathData)].Kind := TRSPathPointKind.ppMoveTo;
  FPathData[High(FPathData)].Point := RSPoint(LP.X + P.X, LP.Y + P.Y);
  FStartPoint := FPathData[High(FPathData)].Point;
  Change;
end;

procedure TRSPathData.LineTo(const P: TRSPoint);
begin
  SetLength(FPathData, Length(FPathData) + 1);
  FPathData[High(FPathData)].Kind := TRSPathPointKind.ppLineTo;
  FPathData[High(FPathData)].Point := P;
  Change;
end;

procedure TRSPathData.LineToRel(const P: TRSPoint);
var
  LP: TRSPoint;
begin
  LP := LastPoint;
  SetLength(FPathData, Length(FPathData) + 1);
  FPathData[High(FPathData)].Kind := TRSPathPointKind.ppLineTo;
  FPathData[High(FPathData)].Point := RSPoint(LP.X + P.X, LP.Y + P.Y);
  Change;
end;

procedure TRSPathData.HLineTo(const X: Single);
begin
  SetLength(FPathData, Length(FPathData) + 1);
  FPathData[High(FPathData)].Kind := TRSPathPointKind.ppLineTo;
  FPathData[High(FPathData)].Point := RSPoint(X, FPathData[High(FPathData) - 1].Point.Y);
  Change;
end;

procedure TRSPathData.HLineToRel(const X: Single);
var
  LP: TRSPoint;
begin
  LP := LastPoint;
  SetLength(FPathData, Length(FPathData) + 1);
  FPathData[High(FPathData)].Kind := TRSPathPointKind.ppLineTo;
  FPathData[High(FPathData)].Point := RSPoint(LP.X + X, LP.Y);
  Change;
end;

procedure TRSPathData.VLineTo(const Y: Single);
begin
  SetLength(FPathData, Length(FPathData) + 1);
  FPathData[High(FPathData)].Kind := TRSPathPointKind.ppLineTo;
  FPathData[High(FPathData)].Point := RSPoint(FPathData[High(FPathData) - 1].Point.X, Y);
  Change;
end;

procedure TRSPathData.VLineToRel(const Y: Single);
var
  LP: TRSPoint;
begin
  LP := LastPoint;
  SetLength(FPathData, Length(FPathData) + 1);
  FPathData[High(FPathData)].Kind := TRSPathPointKind.ppLineTo;
  FPathData[High(FPathData)].Point := RSPoint(LP.X, LP.Y + Y);
  Change;
end;

procedure TRSPathData.QuadCurveTo(const ControlPoint, EndPoint: TRSPoint);
var
  LP, CP1, CP2: TRSPoint;
begin
  LP := LastPoint;
  CP1.X := (1/3 * LP.X) + (2/3 * ControlPoint.X);
  CP1.Y := (1/3 * LP.Y) + (2/3 * ControlPoint.Y);
  CP2.X := (2/3 * ControlPoint.X) + (1/3 * EndPoint.X);
  CP2.Y := (2/3 * ControlPoint.Y) + (1/3 * EndPoint.Y);
  CurveTo(CP1, CP2, EndPoint);
end;

procedure TRSPathData.CurveTo(const ControlPoint1, ControlPoint2, EndPoint: TRSPoint);
var
  i: Integer;
begin
  i := Length(FPathData);
  SetLength(FPathData, i + 3);
  FPathData[i].Kind := TRSPathPointKind.ppCurveTo;
  FPathData[i].Point := ControlPoint1;
  Inc(i);
  FPathData[i].Kind := TRSPathPointKind.ppCurveTo;
  FPathData[i].Point := ControlPoint2;
  Inc(i);
  FPathData[i].Kind := TRSPathPointKind.ppCurveTo;
  FPathData[i].Point := EndPoint;
  Change;
end;

procedure TRSPathData.CurveToRel(const ControlPoint1, ControlPoint2, EndPoint: TRSPoint);
var
  i: Integer;
  LP: TRSPoint;
begin
  i := Length(FPathData);
  LP := LastPoint;
  SetLength(FPathData, i + 3);
  FPathData[i].Kind := TRSPathPointKind.ppCurveTo;
  FPathData[i].Point := RSPoint(LP.X + ControlPoint1.X, LP.Y + ControlPoint1.Y);
  Inc(i);
  FPathData[i].Kind := TRSPathPointKind.ppCurveTo;
  FPathData[i].Point := RSPoint(LP.X + ControlPoint2.X, LP.Y + ControlPoint2.Y);;
  Inc(i);
  FPathData[i].Kind := TRSPathPointKind.ppCurveTo;
  FPathData[i].Point := RSPoint(LP.X + EndPoint.X, LP.Y + EndPoint.Y);;
  Change;
end;

procedure TRSPathData.ClosePath;
begin
  SetLength(FPathData, Length(FPathData) + 1);
  FPathData[High(FPathData)].Kind := TRSPathPointKind.ppClose;
  FPathData[High(FPathData)].Point := FStartPoint;
  Change;
end;

procedure TRSPathData.Change;
begin
  if UpdateCount > 0 then Exit;
  FRecalcBounds := True;
  FLength := 0;
  SetLength(FDistances, 0);
  SetLength(FPolygons, 0);
  if Assigned(FOnChanged) then FOnChanged(Self);
end;

procedure TRSPathData.Clear;
begin
  SetLength(FPathData, 0);
  Change;
end;

function TRSPathData.GetBounds: TRSRect;
var
  I: Integer;
begin
  if Length(FPathData) = 0 then
  begin
    Result := RSRect(0, 0, 0, 0);
    Exit;
  end;
  if FRecalcBounds then
  begin
    Result := RSRect($FFFF, $FFFF, -$FFFF, -$FFFF);
    for I := 0 to High(FPathData) do
    begin
      if FPathData[I].Kind = TRSPathPointKind.ppClose then
        Continue;

      if FPathData[I].Point.X < Result.Left then
        Result.Left := FPathData[I].Point.X;
      if FPathData[I].Point.X > Result.Right then
        Result.Right := FPathData[I].Point.X;
      if FPathData[I].Point.Y < Result.Top then
        Result.Top := FPathData[I].Point.Y;
      if FPathData[I].Point.Y > Result.Bottom then
        Result.Bottom := FPathData[I].Point.Y;
    end;
    // add small amount
    if RSRectWidth(Result) = 0 then
      Result.Right := Result.Left + 0.001;
    if RSRectHeight(Result) = 0 then
      Result.Bottom := Result.Top + 0.001;
    FBounds := Result;
    FRecalcBounds := False;
  end
  else
    Result := FBounds;
end;

procedure TRSPathData.Scale(const scaleX, scaleY: Single);
var
  I: Integer;
begin
  if Length(FPathData) > 0 then
  begin
    for I := 0 to High(FPathData) do
      case FPathData[I].Kind of
        TRSPathPointKind.ppMoveTo, TRSPathPointKind.ppLineTo, TRSPathPointKind.ppCurveTo:
          begin
            FPathData[I].Point.X := FPathData[I].Point.X * scaleX;
            FPathData[I].Point.Y := FPathData[I].Point.Y * scaleY;
          end;
        TRSPathPointKind.ppClose:
          begin
          end;
      end;
    Change;
  end;
end;

procedure TRSPathData.Translate(const dX, dY: Single);
var
  I: Integer;
begin
  if Count > 0 then
  begin
    for I := 0 to High(FPathData) do
      case FPathData[I].Kind of
        TRSPathPointKind.ppMoveTo, TRSPathPointKind.ppLineTo, TRSPathPointKind.ppCurveTo:
          begin
            FPathData[I].Point.X := FPathData[I].Point.X + dX;
            FPathData[I].Point.Y := FPathData[I].Point.Y + dY;
          end;
        TRSPathPointKind.ppClose:
          begin
          end;
      end;
    Change;
  end;
end;

procedure TRSPathData.FitToRect(const ARect: TRSRect);
var
  B: TRSRect;
  M: TRSMatrix;
begin
  B := GetBounds;
  if (RSRectWidth(B) = 0) or (RSRectHeight(B) = 0) then Exit;

  M := CreateTranslateRSMatrix(-B.Left, -B.Top);
  M := RSMatrixMultiply(M, CreateScaleRSMatrix(RSRectWidth(ARect) / RSRectWidth(B), RSRectHeight(ARect) / RSRectHeight(B)));
  M := RSMatrixMultiply(M, CreateTranslateRSMatrix(ARect.Left, ARect.Top));
  ApplyMatrix(M);
end;

procedure TRSPathData.ApplyMatrix(const M: TRSMatrix);
var
  I: Integer;
begin
  if Length(FPathData) > 0 then
  begin
    for I := 0 to High(FPathData) do
      case FPathData[I].Kind of
        TRSPathPointKind.ppMoveTo, TRSPathPointKind.ppLineTo, TRSPathPointKind.ppCurveTo:
          begin
            FPathData[I].Point := RSPoint(RSMatrixTransform(M, RSVector(FPathData[I].Point)));
          end;
        TRSPathPointKind.ppClose:
          begin
          end;
      end;
    Change;
  end;
end;

procedure TRSPathData.Flatten(const Flatness: Single = 0.25);

  procedure CalculateBezierCoefficients(const Bezier: TRSCubicBezier; out ax, bx, cx, ay, by, cy: Single);
  begin
    cx := 3.0 * (Bezier[1].X - Bezier[0].X);
    cy := 3.0 * (Bezier[1].Y - Bezier[0].Y);
    bx := 3.0 * (Bezier[2].X - Bezier[1].X) - cx;
    by := 3.0 * (Bezier[2].Y - Bezier[1].Y) - cy;
    ax := Bezier[3].X - Bezier[0].X - cx - bx;
    ay := Bezier[3].Y - Bezier[0].Y - cy - by;
  end;

  function PointOnBezier(const StartPoint: TRSPoint; const ax, bx, cx, ay, by, cy, t: Single): TRSPoint;
  var
    tSqr: Single;
    tCube: Single;
  begin
    tSqr := t * t;
    tCube := tSqr * t;
    Result.X := (ax * tCube) + (bx * tSqr) + (cx * t) + StartPoint.X;
    Result.Y := (ay * tCube) + (by * tSqr) + (cy * t) + StartPoint.Y;
  end;

  function CreateBezier(const Bezier: TRSCubicBezier; const PointCount: Integer): TRSPolygon;
  var
    ax: Single;
    bx: Single;
    cx: Single;
    ay: Single;
    by: Single;
    cy: Single;
    dT: Single;
    t: Single;
    I: Integer;
  begin
    if PointCount = 0 then
      Exit;
    dT := 1.0 / (1.0 * PointCount - 1.0);
    t := 0.0;
    SetLength(Result, PointCount);
    CalculateBezierCoefficients(Bezier, ax, bx, cx, ay, by, cy);
    for I := 0 to PointCount - 1 do
    begin
      Result[I] := PointOnBezier(Bezier[0], ax, bx, cx, ay, by, cy, t);
      t := t + dT;
    end;
  end;

var
  I, j: Integer;
  BPts: TRSPolygon;
  B: TRSCubicBezier;
  F, Len: Single;
  SegCount: Integer;
  OldPathData: array of TRSPathPoint;
  CurPoint: TRSPoint;
begin
  { scale }
  if Length(FPathData) > 0 then
  begin
    F := Flatness;
    if F < 0.05 then
      F := 0.05;

    { copy data }
    SetLength(OldPathData, Count);
    System.Move(FPathData[0], OldPathData[0], Count * SizeOf(FPathData[0]));
    SetLength(FPathData, 0);

    I := 0;
    while I < Length(OldPathData) do
    begin
      case OldPathData[I].Kind of
        TRSPathPointKind.ppMoveTo:
          begin
            MoveTo(OldPathData[I].Point);
            CurPoint := OldPathData[I].Point;
          end;
        TRSPathPointKind.ppLineTo:
          begin
            LineTo(OldPathData[I].Point);
            CurPoint := OldPathData[I].Point;
          end;
        TRSPathPointKind.ppCurveTo:
          begin
            B[0] := CurPoint;
            B[1] := OldPathData[I].Point;
            inc(I);
            B[2] := OldPathData[I].Point;
            inc(I);
            B[3] := OldPathData[I].Point;
            // calc length
            BPts := CreateBezier(B, 6);
            Len := 0;
            for j := 0 to High(BPts) - 1 do
              Len := Len + RSVectorLength(RSVectorSubtract(RSVector(BPts[j]), RSVector(BPts[j + 1])));
            // flatten
            SegCount := Round(Len / F);
            if SegCount < 2 then
              LineTo(B[3])
            else
            begin
              BPts := CreateBezier(B, SegCount);
              for j := 0 to High(BPts) do
                LineTo(BPts[j]);
              CurPoint := OldPathData[I].Point;
            end;
          end;
        TRSPathPointKind.ppClose:
          begin
            ClosePath;
          end;
      end;
      inc(I);
    end;
    Change;
  end;
end;

function TRSPathData.FlattenToPolygon(var Polygon: TRSPolygon; const Flatness: Single = 0.25): TRSPoint;

  procedure CalculateBezierCoefficients(const Bezier: TRSCubicBezier; out ax, bx, cx, ay, by, cy: Single);
  begin
    cx := 3.0 * (Bezier[1].X - Bezier[0].X);
    cy := 3.0 * (Bezier[1].Y - Bezier[0].Y);
    bx := 3.0 * (Bezier[2].X - Bezier[1].X) - cx;
    by := 3.0 * (Bezier[2].Y - Bezier[1].Y) - cy;
    ax := Bezier[3].X - Bezier[0].X - cx - bx;
    ay := Bezier[3].Y - Bezier[0].Y - cy - by;
  end;

  function PointOnBezier(const StartPoint: TRSPoint; const ax, bx, cx, ay, by, cy, t: Single): TRSPoint;
  var
    tSqr: Single;
    tCube: Single;
  begin
    tSqr := t * t;
    tCube := tSqr * t;
    Result.X := (ax * tCube) + (bx * tSqr) + (cx * t) + StartPoint.X;
    Result.Y := (ay * tCube) + (by * tSqr) + (cy * t) + StartPoint.Y;
  end;

  function CreateBezier(const Bezier: TRSCubicBezier; const PointCount: Integer): TRSPolygon;
  var
    ax: Single;
    bx: Single;
    cx: Single;
    ay: Single;
    by: Single;
    cy: Single;
    dT: Single;
    t: Single;
    I: Integer;
  begin
    if PointCount = 0 then
      Exit;
    dT := 1.0 / (1.0 * PointCount - 1.0);
    t := 0.0;
    SetLength(Result, PointCount);
    CalculateBezierCoefficients(Bezier, ax, bx, cx, ay, by, cy);
    for I := 0 to PointCount - 1 do
    begin
      Result[I] := PointOnBezier(Bezier[0], ax, bx, cx, ay, by, cy, t);
      t := t + dT;
    end;
  end;

  procedure AddPoint(const P: TRSPoint);
  begin
    if (Length(Polygon) > 0) and (SameValue(P.X, Polygon[High(Polygon)].X, Epsilon) and SameValue(P.Y, Polygon[High(Polygon)].Y, Epsilon)) then
      Exit;
    SetLength(Polygon, Length(Polygon) + 1);
    Polygon[High(Polygon)] := P;
  end;

var
  I, j: Integer;
  BPts: TRSPolygon;
  B: TRSCubicBezier;
  SP, CurPoint: TRSPoint;
  F, Len: Single;
  SegCount: Integer;
begin
  Result := RSPoint(0, 0);
  SetLength(Polygon, 0);
  if Length(FPathData) > 0 then
  begin
    F := Flatness;
    if F < 0.05 then
      F := 0.05;

    I := 0;
    while I < Count do
    begin
      case FPathData[I].Kind of
        TRSPathPointKind.ppMoveTo:
          begin
            if Length(Polygon) > 0 then
              AddPoint(RSClosePolygon);
            AddPoint(FPathData[I].Point);
            CurPoint := FPathData[I].Point;
            SP := CurPoint;
          end;
        TRSPathPointKind.ppLineTo:
          begin
            AddPoint(FPathData[I].Point);
            CurPoint := FPathData[I].Point;
          end;
        TRSPathPointKind.ppCurveTo:
          begin
            B[0] := CurPoint;
            B[1] := FPathData[I].Point;
            inc(I);
            B[2] := FPathData[I].Point;
            inc(I);
            B[3] := FPathData[I].Point;
            // calc length
            BPts := CreateBezier(B, 6);
            Len := 0;
            for j := 0 to High(BPts) - 1 do
              Len := Len + RSVectorLength(RSVectorSubtract(RSVector(BPts[j]), RSVector(BPts[j + 1])));
            // flatten
            SegCount := Round(Len / F);
            if SegCount < 2 then
            begin
              AddPoint(B[0]);
              AddPoint(B[3]);
            end
            else
            begin
              BPts := CreateBezier(B, SegCount);
              for j := 0 to High(BPts) do
                AddPoint(BPts[j]);
            end;
            CurPoint := FPathData[I].Point;
          end;
        TRSPathPointKind.ppClose:
          begin
            AddPoint(SP);
            AddPoint(RSClosePolygon);
          end;
      end;
      inc(I);
    end;
    Result := RSPoint(Abs(RSRectWidth(GetBounds)), Abs(RSRectHeight(GetBounds)));
  end;
end;

procedure TRSPathData.AddEllipse(const ARect: TRSRect);
var
  cx, cy: Single;
  px, py: Single;
begin
  cx := (ARect.Left + ARect.Right) / 2;
  cy := (ARect.Top + ARect.Bottom) / 2;
  px := CurveKappa * (RSRectWidth(ARect) / 2);
  py := CurveKappa * (RSRectHeight(ARect) / 2);
  MoveTo(RSPoint(ARect.Left, cy));
  CurveTo(RSPoint(ARect.Left, cy - py), RSPoint(cx - px, ARect.Top), RSPoint(cx, ARect.Top));
  CurveTo(RSPoint(cx + px, ARect.Top), RSPoint(ARect.Right, cy - py), RSPoint(ARect.Right, cy));
  CurveTo(RSPoint(ARect.Right, cy + py), RSPoint(cx + px, ARect.Bottom), RSPoint(cx, ARect.Bottom));
  CurveTo(RSPoint(cx - px, ARect.Bottom), RSPoint(ARect.Left, cy + py), RSPoint(ARect.Left, cy));
end;

procedure TRSPathData.AddRectangle(const ARect: TRSRect; const XRadius, YRadius: Single; const ACorners: TRSCorners;
  const ACornerType: TRSCornerType = TRSCornerType.ctRound);
var
  R: TRSRect;
  X1, X2, Y1, Y2: Single;
begin
  R := ARect;
  X1 := XRadius;
  if RSRectWidth(R) - (X1 * 2) < 0 then
    X1 := (XRadius * (RSRectWidth(R) / (X1 * 2)));
  X2 := X1 / 2;
  Y1 := YRadius;
  if RSRectHeight(R) - (Y1 * 2) < 0 then
    Y1 := (YRadius * (RSRectHeight(R) / (Y1 * 2)));
  Y2 := Y1 / 2;

  MoveTo(RSPoint(R.Left, R.Top + Y1));
  if TRSCorner.crTopLeft in ACorners then
  begin
    case ACornerType of
      // ctRound - default
      TRSCornerType.ctBevel:
        LineTo(RSPoint(R.Left + X1, R.Top));
      TRSCornerType.ctInnerRound:
        CurveTo(RSPoint(R.Left + X2, R.Top + Y1), RSPoint(R.Left + X1, R.Top + Y2), RSPoint(R.Left + X1, R.Top));
      TRSCornerType.ctInnerLine:
        begin
          LineTo(RSPoint(R.Left + X2, R.Top + Y1));
          LineTo(RSPoint(R.Left + X1, R.Top + Y2));
          LineTo(RSPoint(R.Left + X1, R.Top));
        end;
    else
      CurveTo(RSPoint(R.Left, R.Top + (Y2)), RSPoint(R.Left + X2, R.Top), RSPoint(R.Left + X1, R.Top))
    end;
  end
  else
  begin
    LineTo(RSPoint(R.Left, R.Top));
    LineTo(RSPoint(R.Left + X1, R.Top));
  end;
  LineTo(RSPoint(R.Right - X1, R.Top));
  if TRSCorner.crTopRight in ACorners then
  begin
    case ACornerType of
      // ctRound - default
      TRSCornerType.ctBevel:
        LineTo(RSPoint(R.Right, R.Top + Y1));
      TRSCornerType.ctInnerRound:
        CurveTo(RSPoint(R.Right - X1, R.Top + Y2), RSPoint(R.Right - X2, R.Top + Y1), RSPoint(R.Right, R.Top + Y1));
      TRSCornerType.ctInnerLine:
        begin
          LineTo(RSPoint(R.Right - X1, R.Top + Y2));
          LineTo(RSPoint(R.Right - X2, R.Top + Y1));
          LineTo(RSPoint(R.Right, R.Top + Y1));
        end;
    else
      CurveTo(RSPoint(R.Right - X2, R.Top), RSPoint(R.Right, R.Top + (Y2)), RSPoint(R.Right, R.Top + Y1))
    end;
  end
  else
  begin
    LineTo(RSPoint(R.Right, R.Top));
    LineTo(RSPoint(R.Right, R.Top + Y1));
  end;
  LineTo(RSPoint(R.Right, R.Bottom - Y1));
  if TRSCorner.crBottomRight in ACorners then
  begin
    case ACornerType of
      // ctRound - default
      TRSCornerType.ctBevel:
        LineTo(RSPoint(R.Right - X1, R.Bottom));
      TRSCornerType.ctInnerRound:
        CurveTo(RSPoint(R.Right - X2, R.Bottom - Y1), RSPoint(R.Right - X1, R.Bottom - Y2), RSPoint(R.Right - X1, R.Bottom));
      TRSCornerType.ctInnerLine:
        begin
          LineTo(RSPoint(R.Right - X2, R.Bottom - Y1));
          LineTo(RSPoint(R.Right - X1, R.Bottom - Y2));
          LineTo(RSPoint(R.Right - X1, R.Bottom));
        end;
    else
      CurveTo(RSPoint(R.Right, R.Bottom - (Y2)), RSPoint(R.Right - X2, R.Bottom), RSPoint(R.Right - X1, R.Bottom))
    end;
  end
  else
  begin
    LineTo(RSPoint(R.Right, R.Bottom));
    LineTo(RSPoint(R.Right - X1, R.Bottom));
  end;
  LineTo(RSPoint(R.Left + X1, R.Bottom));
  if TRSCorner.crBottomLeft in ACorners then
  begin
    case ACornerType of
      // ctRound - default
      TRSCornerType.ctBevel:
        LineTo(RSPoint(R.Left, R.Bottom - Y1));
      TRSCornerType.ctInnerRound:
        CurveTo(RSPoint(R.Left + X1, R.Bottom - Y2), RSPoint(R.Left + X2, R.Bottom - Y1), RSPoint(R.Left, R.Bottom - Y1));
      TRSCornerType.ctInnerLine:
        begin
          LineTo(RSPoint(R.Left + X1, R.Bottom - Y2));
          LineTo(RSPoint(R.Left + X2, R.Bottom - Y1));
          LineTo(RSPoint(R.Left, R.Bottom - Y1));
        end;
    else
      CurveTo(RSPoint(R.Left + X2, R.Bottom), RSPoint(R.Left, R.Bottom - (Y2)), RSPoint(R.Left, R.Bottom - Y1))
    end;
  end
  else
  begin
    LineTo(RSPoint(R.Left, R.Bottom));
    LineTo(RSPoint(R.Left, R.Bottom - Y1));
  end;
  ClosePath;
end;

procedure TRSPathData.AddArc(const Center, Radius: TRSPoint; StartAngle, SweepAngle: Single);
const
  bezier_arc_angle_epsilon = 0.01;
var
  UseMoveTo: Boolean;
  I: Integer;
  F: Single;
  total_sweep, local_sweep, prev_sweep: Single;
  done: Boolean;
begin
  StartAngle := DegToRad(StartAngle);
  SweepAngle := DegToRad(SweepAngle);

  I := trunc(StartAngle / (2.0 * cPI));
  F := StartAngle - (I * 2.0 * cPI);

  StartAngle := F;

  if SweepAngle >= 2.0 * cPI then
    SweepAngle := 2.0 * cPI;
  if SweepAngle <= -2.0 * cPI then
    SweepAngle := -2.0 * cPI;

  if Abs(SweepAngle) < 1E-10 then
  begin
    Exit;
  end;

  total_sweep := 0.0;

  done := False;
  UseMoveTo := True;
  repeat
    if SweepAngle < 0.0 then
    begin
      prev_sweep := total_sweep;
      local_sweep := -cPI * 0.5;
      total_sweep := total_sweep - (cPI * 0.5);
      if total_sweep <= SweepAngle + bezier_arc_angle_epsilon then
      begin
        local_sweep := SweepAngle - prev_sweep;
        done := True;
      end;
    end
    else
    begin
      prev_sweep := total_sweep;
      local_sweep := cPI * 0.5;
      total_sweep := total_sweep + (Pi * 0.5);
      if total_sweep >= SweepAngle - bezier_arc_angle_epsilon then
      begin
        local_sweep := SweepAngle - prev_sweep;
        done := True;
      end;
    end;
    DrawArcWithBezier(Self, Center.X, Center.Y, Radius.X, Radius.Y, StartAngle, local_sweep, UseMoveTo);
    UseMoveTo := False;
    StartAngle := StartAngle + local_sweep;
  until done;
end;

function TRSPathData.IsEmpty: Boolean;
begin
  Result := (Length(FPathData) = 0) or (RSRectWidth(GetBounds) * RSRectHeight(GetBounds) = 0);
end;

{$ENDIF}

procedure TRSPathData.AddArcSvgPart(const Center, Radius: TRSPoint; StartAngle, SweepAngle: Single);
const
  bezier_arc_angle_epsilon = 0.01;
var
  UseMoveTo: Boolean;
  I: Integer;
  F: Single;
  total_sweep, local_sweep, prev_sweep: Single;
  done: Boolean;
begin
  StartAngle := DegToRad(StartAngle);
  SweepAngle := DegToRad(SweepAngle);

  I := trunc(StartAngle / (2.0 * cPI));
  F := StartAngle - (I * 2.0 * cPI);

  StartAngle := F;

  if SweepAngle >= 2.0 * cPI then
    SweepAngle := 2.0 * cPI;
  if SweepAngle <= -2.0 * cPI then
    SweepAngle := -2.0 * cPI;

  if Abs(SweepAngle) < 1E-10 then
  begin
    Exit;
  end;

  total_sweep := 0.0;

  done := False;
  UseMoveTo := False;
  repeat
    if SweepAngle < 0.0 then
    begin
      prev_sweep := total_sweep;
      local_sweep := -cPI * 0.5;
      total_sweep := total_sweep - (cPI * 0.5);
      if total_sweep <= SweepAngle + bezier_arc_angle_epsilon then
      begin
        local_sweep := SweepAngle - prev_sweep;
        done := True;
      end;
    end
    else
    begin
      prev_sweep := total_sweep;
      local_sweep := cPI * 0.5;
      total_sweep := total_sweep + (Pi * 0.5);
      if total_sweep >= SweepAngle - bezier_arc_angle_epsilon then
      begin
        local_sweep := SweepAngle - prev_sweep;
        done := True;
      end;
    end;
    DrawArcWithBezier(Self, Center.X, Center.Y, Radius.X, Radius.Y, StartAngle, local_sweep, UseMoveTo);
    UseMoveTo := False;
    StartAngle := StartAngle + local_sweep;
  until done;
end;

procedure TRSPathData.AddArcSvg(const P1, Radius: TRSPoint; Angle: Single; const LargeFlag, SweepFlag: Boolean;
  const P2: TRSPoint);
var
  I: Integer;
  m_radii_ok: Boolean;
  V, P, N, sq, rx, ry, x0, y0, X1, Y1, X2, Y2, cx, cy, ux, uy, vx, vy,

    dx2, dy2, prx, pry, px1, py1, cx1, cy1, sx2, sy2,

    sign, coef,

    radii_check, start_angle, sweep_angle,

    cos_a, sin_a: Single;
  tm: TRSMatrix;
  len: Integer;
{$IFDEF FMX}
  tempPath: TRSPathData;
{$ENDIF}
begin
  // Trivial case: Arc transforms to point or ellipse
  if (P1 = P2) then
  begin
    AddEllipse(RSRect(P1.X-Radius.X, P1.Y-2*Radius.Y, P1.X+Radius.X, P1.Y));
    Exit;
  end;

  rx := Radius.X;
  ry := Radius.Y;
  x0 := P1.X;
  y0 := P1.Y;
  X2 := P2.X;
  Y2 := P2.Y;
  Angle := DegToRad(Angle);

  m_radii_ok := True;

  if rx < 0.0 then
    rx := -rx;

  if ry < 0.0 then
    ry := -rx;

  // Calculate the middle point between
  // the current and the final points
  dx2 := (x0 - X2) / 2.0;
  dy2 := (y0 - Y2) / 2.0;

  // Convert angle from degrees to radians
  cos_a := Cos(Angle);
  sin_a := Sin(Angle);

  // Calculate (x1, y1)
  X1 := cos_a * dx2 + sin_a * dy2;
  Y1 := -sin_a * dx2 + cos_a * dy2;

  // Ensure radii are large enough
  prx := rx * rx;
  pry := ry * ry;
  px1 := X1 * X1;
  py1 := Y1 * Y1;

  // Check that radii are large enough
  radii_check := px1 / prx + py1 / pry;

  if radii_check > 1.0 then
  begin
    rx := Sqrt(radii_check) * rx;
    ry := Sqrt(radii_check) * ry;
    prx := rx * rx;
    pry := ry * ry;

    if radii_check > 10.0 then
      m_radii_ok := False;

  end;

  // Calculate (cx1, cy1)
  if LargeFlag = SweepFlag then
    sign := -1.0
  else
    sign := 1.0;

  sq := (prx * pry - prx * py1 - pry * px1) / (prx * py1 + pry * px1);

  if sq < 0 then
    coef := sign * Sqrt(0)
  else
    coef := sign * Sqrt(sq);

  cx1 := coef * ((rx * Y1) / ry);
  cy1 := coef * -((ry * X1) / rx);

  // Calculate (cx, cy) from (cx1, cy1)
  sx2 := (x0 + X2) / 2.0;
  sy2 := (y0 + Y2) / 2.0;
  cx := sx2 + (cos_a * cx1 - sin_a * cy1);
  cy := sy2 + (sin_a * cx1 + cos_a * cy1);

  // Calculate the start_angle (angle1) and the sweep_angle (dangle)
  ux := (X1 - cx1) / rx;
  uy := (Y1 - cy1) / ry;
  vx := (-X1 - cx1) / rx;
  vy := (-Y1 - cy1) / ry;

  // Calculate the angle start
  N := Sqrt(ux * ux + uy * uy);
  P := ux; // (1 * ux ) + (0 * uy )

  if uy < 0 then
    sign := -1.0
  else
    sign := 1.0;

  V := P / N;

  if V < -1.0 then
    V := -1.0;

  if V > 1.0 then
    V := 1.0;

  start_angle := sign * ArcCos(V);

  // Calculate the sweep angle
  N := Sqrt((ux * ux + uy * uy) * (vx * vx + vy * vy));
  P := ux * vx + uy * vy;

  if ux * vy - uy * vx < 0 then
    sign := -1.0
  else
    sign := 1.0;

  V := P / N;

  if V < -1.0 then
    V := -1.0;

  if V > 1.0 then
    V := 1.0;

  sweep_angle := sign * ArcCos(V);

  if (not SweepFlag) and (sweep_angle > 0) then
    sweep_angle := sweep_angle - Pi * 2.0
  else if SweepFlag and (sweep_angle < 0) then
    sweep_angle := sweep_angle + Pi * 2.0;

  len := Count;
  {$IFDEF FMX}
  tempPath := TRSPathData.Create;
  try
    tempPath.AddArcSvgPart(RSPoint(0, 0), RSPoint(rx, ry), RadToDeg(start_angle), RadToDeg(sweep_angle));

    tm := RSIdentityMatrix;
    tm.m31 := cx;
    tm.m32 := cy;
    tm := RSMatrixMultiply(CreateRotationRSMatrix(Angle), tm);

    i := 0;
    while i < tempPath.Count do
    begin
      case tempPath.Points[i].Kind of
        TRSPathPointKind.{$IFDEF DELPHI20PLUS}MoveTo{$ELSE}ppMoveTo{$ENDIF}:
          MoveTo( RSPoint(RSMatrixTransform(tm, RSVector(tempPath.Points[I].Point))));
        TRSPathPointKind.{$IFDEF DELPHI20PLUS}LineTo{$ELSE}ppLineTo{$ENDIF}:
          LineTo( RSPoint(RSMatrixTransform(tm, RSVector(tempPath.Points[I].Point))));
        TRSPathPointKind.{$IFDEF DELPHI20PLUS}CurveTo{$ELSE}ppCurveTo{$ENDIF}:
        begin
          CurveTo( RSPoint(RSMatrixTransform(tm, RSVector(tempPath.Points[I].Point))),
                   RSPoint(RSMatrixTransform(tm, RSVector(tempPath.Points[I+1].Point))),
                   RSPoint(RSMatrixTransform(tm, RSVector(tempPath.Points[I+2].Point))) );
          Inc(i, 2);
        end;
        TRSPathPointKind.{$IFDEF DELPHI20PLUS}Close{$ELSE}ppClose{$ENDIF}:
          ClosePath;
      end;
      Inc(i);
    end;
  finally
    tempPath.Free;
  end;
  {$ELSE}
  AddArcSvgPart(RSPoint(0, 0), RSPoint(rx, ry), RadToDeg(start_angle), RadToDeg(sweep_angle));

  tm := RSIdentityMatrix;
  tm.m31 := cx;
  tm.m32 := cy;
  tm := RSMatrixMultiply(CreateRotationRSMatrix(Angle), tm);

  I := len;
  while I < Count do
  begin
    FPathData[I].Point := RSPoint(RSMatrixTransform(tm, RSVector(FPathData[I].Point)));
    Inc(I);
  end;
  {$ENDIF}
end;

function TRSPathData.GetTok(const S: string; var Pos: Integer): string;
var
  StringBuilder: TStringBuilder;
begin
  StringBuilder := TStringBuilder.Create;
  try
    Result := StringBuilder.ToString;
    if Pos > Length(S) then
      Exit;
    while (Pos <= Length(S)) and (S[Pos] = ' ') do
      Inc(Pos);

    while Pos <= Length(S) do
    begin
      if not ContainsStr( 'zmlchvsqtaZMLCHVSQTA', S[Pos] ) then
        Break;
      StringBuilder.Append(S[Pos]);
      Inc(Pos);
    end;
    Result := StringBuilder.ToString;
  finally
    StringBuilder.Free;
  end;
end;

function TRSPathData.GetDigit(const S: string; var Pos: Integer): string;
var
  StringBuilder: TStringBuilder;
  FoundComma: Boolean;
  FoundDecimal: Boolean;
begin
  StringBuilder := TStringBuilder.Create;
  try
    Result := '';
    if Pos > Length(S) then
      Exit;
    FoundComma := False;
    while (Pos <= Length(S)) and ((S[Pos] = ' ') or ((not FoundComma) and(S[Pos] = ','))) do
    begin
      if S[Pos] = ',' then
        FoundComma := True;
      Inc(Pos);
    end;
    FoundDecimal := False;
    while Pos <= Length(S) do
    begin
      if (StringBuilder.Length > 0) and (S[Pos] = '-') then
        Break;
      if (not FoundDecimal) and (not ContainsStr('0123456789.', S[Pos]) and not (S[Pos] = '-')) then
        Break;
      StringBuilder.Append(S[Pos]);
      Inc(Pos);
      if ContainsStr('0123456789', S[Pos-1]) then
        Break;
    end;
    FoundComma := False;
    while (Pos <= Length(S)) and ((S[Pos] = ' ') or ((not FoundComma) and (S[Pos] = ','))) do
    begin
      if S[Pos] = ',' then
        FoundComma := True;
      Inc(Pos);
    end;
    Result := StringBuilder.ToString;
  finally
    StringBuilder.Free;
  end;
end;

function TRSPathData.GetNum(const S: string; var Pos: Integer): string;
var
  StringBuilder: TStringBuilder;
  FoundComma: Boolean;
  FoundDecimal: Boolean;
begin
  StringBuilder := TStringBuilder.Create;
  try
    Result := '';
    if Pos > Length(S) then
      Exit;
    FoundComma := False;
    while (Pos <= Length(S)) and ((S[Pos] = ' ') or ((not FoundComma) and(S[Pos] = ','))) do
    begin
      if S[Pos] = ',' then
        FoundComma := True;
      Inc(Pos);
    end;
    FoundDecimal := False;
    while Pos <= Length(S) do
    begin
      if SameText(S[Pos], 'e') then
      begin
        StringBuilder.Append(S[Pos]);
        Inc(Pos);
        Continue;
      end;
      if (S[Pos] = '-') and (StringBuilder.Length > 0) and SameText(StringBuilder[StringBuilder.Length - 1], 'e') then
      begin
        StringBuilder.Append(S[Pos]);
        Inc(Pos);
        Continue;
      end;
      if (StringBuilder.Length > 0) and (S[Pos] = '-') then
        Break;
      if (not FoundDecimal) and (not ContainsStr('0123456789.', S[Pos]) and not (S[Pos] = '-')) then
        Break;
      if (FoundDecimal) and (not ContainsStr('0123456789', S[Pos]) and not (S[Pos] = '-')) then
        Break;
      if S[Pos] = '.' then
        FoundDecimal := True;
      StringBuilder.Append(S[Pos]);
      Inc(Pos);
    end;
    FoundComma := False;
    while (Pos <= Length(S)) and ((S[Pos] = ' ') or ((not FoundComma) and (S[Pos] = ','))) do
    begin
      if S[Pos] = ',' then
        FoundComma := True;
      Inc(Pos);
    end;
    Result := StringBuilder.ToString;
  finally
    StringBuilder.Free;
  end;
end;

function TRSPathData.GetPointFromStr(const S: string; var Pos: Integer): TRSPoint;
var
  X, Y: string;
begin
  if Pos > Length(S) then
  begin
    Result := RSPoint(0, 0);
    Exit;
  end;
  while (Pos <= Length(S)) and ((S[Pos] = ',') or (S[Pos] = ' ')) do
    Inc(Pos);
  X := GetNum(S, Pos);
  while (Pos <= Length(S)) and ((S[Pos] = ',') or (S[Pos] = ' ')) do
    Inc(Pos);
  Y := GetNum(S, Pos);
  while (Pos <= Length(S)) and ((S[Pos] = ',') or (S[Pos] = ' ')) do
    Inc(Pos);
  Result := RSPoint(StrToFloat(X, USFormatSettings), StrToFloat(Y, USFormatSettings));
end;

function TRSPathData.IsFormatStrStored: Boolean;
begin
  result := FormatStr <> LongFormatStr;
end;

{$IF not (Defined(DELPHI16PLUS) and Defined(FMX))}

{ TRSPosition }

constructor TRSPosition.Create(const ADefaultValue: TRSPoint);
begin
  inherited Create;
  FDefaultValue := ADefaultValue;
  FX := FDefaultValue.X;
  FY := FDefaultValue.Y;
end;

procedure TRSPosition.Assign(Source: TPersistent);
begin
  if Source is TRSPosition then
  begin
    Point := TRSPosition(Source).Point;
  end
  else
    inherited
end;

procedure TRSPosition.Change;
begin
  if Assigned(OnChange) then OnChange(Self);
end;

procedure TRSPosition.SetPointNoChange(const P: TRSPoint);
begin
  FX := P.X;
  FY := P.Y;
end;

function TRSPosition.Empty: Boolean;
begin
  Result := (FX = 0) and (FY = 0);
end;

procedure TRSPosition.DefineProperties(Filer: TFiler);
begin
  inherited;
  Filer.DefineProperty('Point', ReadPoint, WritePoint, False);
end;

procedure TRSPosition.ReadPoint(Reader: TReader);
begin
  Point := StringToRSPoint(Reader.ReadString);
end;

procedure TRSPosition.WritePoint(Writer: TWriter);
begin
  Writer.WriteString(String(RSPointToString(Point)));
end;

function TRSPosition.IsXStored: Boolean;
begin
  Result := not SameValue(FDefaultValue.X, FX, Epsilon);
end;

function TRSPosition.IsYStored: Boolean;
begin
  Result := not SameValue(FDefaultValue.Y, FY, Epsilon);
end;

function TRSPosition.GetVector: TRSVector;
begin
  Result := RSVector(FX, FY);
end;

procedure TRSPosition.SetVector(const Value: TRSVector);
begin
  Point := RSPoint(Value.X, Value.Y);
end;

procedure TRSPosition.Reflect(const Normal: TRSVector);
begin
  Vector := RSVectorReflect(Vector,Normal);
end;

function TRSPosition.GetPoint: TRSPoint;
begin
  Result := RSPoint(FX, FY);
end;

procedure TRSPosition.SetPoint(const Value: TRSPoint);
begin
  FX := Value.X;
  FY := Value.Y;
  Change;
end;

procedure TRSPosition.SetX(const Value: Single);
begin
  if FX <> Value then
  begin
    FX := Value;
    Change;
  end;
end;

procedure TRSPosition.SetY(const Value: Single);
begin
  if FY <> Value then
  begin
    FY := Value;
    Change;
  end;
end;

{ TRSTransform }

constructor TRSTransform.Create;
begin
  inherited;
  FMatrix := RSIdentityMatrix;
  FPosition := TRSPosition.Create(RSPoint(0, 0));
  FPosition.OnChange := MatrixChanged;
  FScale := TRSPosition.Create(RSPoint(1, 1));
  FScale.OnChange := MatrixChanged;
  FSkew := TRSPosition.Create(RSPoint(0, 0));
  FSkew.OnChange := MatrixChanged;
  FRotationCenter := TRSPosition.Create(RSPoint(0.5, 0.5));
  FRotationCenter.OnChange := MatrixChanged;
end;

destructor TRSTransform.Destroy;
begin
  FRotationCenter.Free;
  FScale.Free;
  FSkew.Free;
  FPosition.Free;
  inherited;
end;

procedure TRSTransform.Assign(Source: TPersistent);
begin
  if Source is TRSTransform then
  begin
    FPosition.FX := TRSTransform(Source).Position.FX;
    FPosition.FY := TRSTransform(Source).Position.FY;
    FScale.FX := TRSTransform(Source).Scale.FX;
    FScale.FY := TRSTransform(Source).Scale.FY;
    FSkew.FX := TRSTransform(Source).Skew.FX;
    FSkew.FY := TRSTransform(Source).Skew.FY;
    FRotationCenter.FX := TRSTransform(Source).RotationCenter.FX;
    FRotationCenter.FY := TRSTransform(Source).RotationCenter.FY;
    MatrixChanged(Self);
  end
  else
    inherited Assign(Source);
end;

procedure TRSTransform.Change;
begin
  if Assigned(FOnChanged) then FOnChanged(Self);
end;

procedure TRSTransform.MatrixChanged(Sender: TObject);
var
  aRect: TRSRect;
begin
  FMatrix := RSIdentityMatrix;
  FMatrix.m31 := FPosition.X;
  FMatrix.m32 := FPosition.Y;
  FMatrix.m13 := FSkew.X;
  FMatrix.m23 := FSkew.Y;
  FMatrix.m11 := FScale.X;
  FMatrix.m22 := FScale.Y;
  if FRotationAngle <> 0 then
  begin
    if Assigned(FOnGetShapeRect) then
    begin
      FOnGetShapeRect(Self, aRect);
      FMatrix := RSMatrixMultiply(FMatrix, CreateRotationRSMatrix(RSCenterPoint(aRect), DegToRad(FRotationAngle)));
    end
    else
      FMatrix := RSMatrixMultiply(FMatrix, CreateRotationRSMatrix(DegToRad(FRotationAngle)));
  end;
  Change;
end;

procedure TRSTransform.SetPosition(const Value: TRSPosition);
begin
  FPosition.Assign(Value);
end;

procedure TRSTransform.SetRotationAngle(const Value: Single);
begin
  if FRotationAngle <> Value then
  begin
    FRotationAngle := Value;
    MatrixChanged(Self);
  end;
end;

procedure TRSTransform.SetScale(const Value: TRSPosition);
begin
  FScale.Assign(Value);
end;

procedure TRSTransform.SetSkew(const Value: TRSPosition);
begin
  FSkew.Assign(Value);
end;

{$IFEND}

initialization
{$IFNDEF DELPHI15PLUS}
  GetLocaleFormatSettings(1033, USFormatSettings);
{$ELSE}
  USFormatSettings := TFormatSettings.Create('en-us');
{$ENDIF}
end.